#include	"compiler.h"

#include	"dosio.h"
#include	"milstr.h"

#define	PROF_MAXLINE	1024

#define	PROF_SIZE		2048
#define	PROF_LINESIZE	1024
#define	PROF_KEYSIZE	64

typedef struct {
	FILEH	fh;
	DWORD	remain;
	DWORD	pos;
	DWORD	headpos;
	int		bufp;
	int		eof;
	int		lines;
	int		appfound;
	char	buf[PROF_SIZE];
	char	linebuf[PROF_LINESIZE];
	short	access;
} PROF_TBL;


#ifdef PROFILE_CRLF
#define		MES_CR		"\r\n"
#define		MES_CRLEN	2
#endif

#ifdef PROFILE_CR
#define		MES_CR		"\r"
#define		MES_CRLEN	1
#endif

#ifdef PROFILE_LF
#define		MES_CR		"\n"
#define		MES_CRLEN	1
#endif


static char *search_head(char *str) {

	while((*str) && ((BYTE)(*str) <= 0x20)) {
		str++;
	}
	return(str);
}

static int search_size(char *str) {

	int		size = 0;
	int		pos = 0;

	while(*str) {
		pos++;
		if ((BYTE)(*str++) > 0x20) {
			size = pos;
		}
	}
	return(size);
}

static BOOL memcmpex(char *str, char *cmp, int size) {

	BYTE	s, c;

	if (!size) {
		return(1);
	}
	while(size--) {
		s = *str++;
		if ((s >= 'a') && (s <= 'z')) {
			s -= 0x20;
		}
		c = *cmp++;
		if ((c >= 'a') && (c <= 'z')) {
			c -= 0x20;
		}
		if (s != c) {
			return(1);
		}
	}
	return(0);
}

static BOOL a2iex(char *p, int len, int *ret) {

	int		value;
	int		negflag = 0;
	char	c;

	if (!len) {
		return(FAILURE);
	}
	if (*p == '+') {
		p++;
		len--;
	}
	else if (*p == '-') {
		p++;
		len--;
		negflag = 1;
	}
	while((len) && (*p) && ((BYTE)*p <= 0x20)) {
		p++;
		len--;
	}
	if (!len) {
		return(FAILURE);
	}
	value = 0;
	while(len--) {
		c = *p++;
		if ((c < '0') && (c > '9')) {
			return(FAILURE);
		}
		value *= 10;
		value += (c - '0');
	}
	if (negflag) {
		value *= -1;
	}
	*ret = value;
	return(SUCCESS);
}




// MS-CȤ C++Ϻʤ(


static void _prf_construct(PROF_TBL *f, char *filename, int write) {

	ZeroMemory(f, sizeof(PROF_TBL));
	f->fh = (FILEH)-1;
	f->access = file_attr(filename);
	if (f->access != -1) {
		f->fh = file_open(filename);
	}
	else if (write) {
		f->fh = file_create(filename, FTYPE_TEXT);
		if (f->fh != (FILEH)-1) {
			f->access = 0x20;
		}
	}
}

static void _prf_destruct(PROF_TBL *f) {

	if ((f->fh) != (FILEH)-1) {
		file_close(f->fh);
	}
}








static BOOL lineread(PROF_TBL *f) {

	int		i;
	char	*p;
	char	c = 0;

	(f->headpos) = (f->pos);
	if (((f->fh) == (FILEH)-1) || (f->eof)) {
		return(FAILURE);
	}
	(f->lines)++;
	p = f->linebuf;
	for (i=0; i<(PROF_LINESIZE-1); i++) {
		if (!(f->remain)) {
			f->bufp = 0;
			f->remain = file_read(f->fh, f->buf, PROF_SIZE);
			if ((f->remain) == -1) {
				(f->remain) = 0;
			}
			if ((f->remain) == 0) {
				if (!i) {
					f->eof = 1;
					return(FAILURE);
				}
				if (c == 0x1a) {
					(f->pos)--;
				}
				f->eof = -1;
				break;
			}
		}
		c = f->buf[f->bufp];
		(f->bufp)++;
		(f->pos)++;
		(f->remain)--;
		if ((c == 0x0a) || (c == 0x0d)) {
			break;
		}
		*p++ = c;
	}
	*p++ = '\0';

	if ((c == 0x0d) & (!(f->eof))) {
		if (!(f->remain)) {
			f->bufp = 0;
			f->remain = file_read(f->fh, f->buf, PROF_SIZE);
			if ((f->remain) == -1) {
				(f->remain) = 0;
			}
		}
		if (f->remain) {
			if ((f->buf[f->bufp]) == 0x0a) {
				(f->bufp)++;
				(f->pos)++;
				(f->remain)--;
			}
		}
		else {
			f->eof = 1;
		}
	}
	return(SUCCESS);
}

static char *keycheck(PROF_TBL *f, char *app, char *key) {

	char	*p;
	int		appsize;
	int		keysize;
	int		cursize;
	DWORD	posbak;

	app = search_head(app);
	appsize = search_size(app);
	if (!appsize) {
		return(NULL);
	}
	key = search_head(key);
	keysize = search_size(key);
	if (!keysize) {
		return(NULL);
	}
	while(1) {
		if ((lineread(f)) || (f->lines > PROF_MAXLINE)) {
			return(NULL);
		}
		p = search_head(f->linebuf);
		cursize = search_size(p);
		if (((cursize - 2) == appsize) &&
			(p[0] == '[') && (p[appsize + 1] == ']') &&
			(!memcmpex(p+1, app, appsize))) {
			break;
		}
	}
	f->appfound = 1;
	posbak = f->pos;

	while(1) {
		if ((lineread(f)) || (f->lines > PROF_MAXLINE)) {
			(f->headpos) = posbak;
			(f->pos) = posbak;
			return(NULL);
		}
		p = search_head(f->linebuf);
		cursize = search_size(p);
		if (!cursize) {
			continue;
		}
		if ((p[0] == '[') && (p[cursize - 1] == ']')) {
			(f->headpos) = posbak;
			(f->pos) = posbak;
			return(NULL);
		}
		posbak = f->pos;
		if (memcmpex(p, key, keysize)) {
			continue;
		}
		p += keysize;
		while((*p) && ((BYTE)(*p) <= 0x20)) {
			p++;
		}
		if (*p == '=') {
			break;
		}
	}
	return(p + 1);
}


static BOOL fileinsert(PROF_TBL *f, DWORD baklen, DWORD wrlen) {

	DWORD	p1, p2;
	DWORD	remain;
	int		size;
	int		rsize;
	int		wsize;

	f->remain = 0;
	f->pos = file_seek(f->fh, 0, FSEEK_END);
	if (baklen < wrlen) {
		p1 = f->pos;
		p2 = (f->pos) + wrlen - baklen;
		file_zeroclr(f->fh, wrlen - baklen);			// ver0.28
		remain = (f->pos) - (f->headpos);
		while(remain) {
			if (remain >= PROF_SIZE) {
				size = PROF_SIZE;
			}
			else {
				size = (int)remain;
			}
			p1 -= size;
			p2 -= size;
			remain -= size;
			f->pos = file_seek(f->fh, p1, FSEEK_SET);
			rsize = file_read(f->fh, f->buf, size);
			if (rsize != size) {
				return(FAILURE);
			}
			f->pos = file_seek(f->fh, p2, FSEEK_SET);
			wsize = file_write(f->fh, f->buf, rsize);
			if (wsize != rsize) {
				return(FAILURE);
			}
		}
	}
	else if (baklen > wrlen) {
		p1 = (f->headpos) + baklen;
		p2 = (f->headpos) + wrlen;
		remain = (f->pos) - p1;
		while(remain) {
			if (remain >= PROF_SIZE) {
				size = PROF_SIZE;
			}
			else {
				size = (int)remain;
			}
			remain -= size;
			f->pos = file_seek(f->fh, p1, FSEEK_SET);
			rsize = file_read(f->fh, f->buf, size);
			if (rsize != size) {
				return(FAILURE);
			}
			f->pos = file_seek(f->fh, p2, FSEEK_SET);
			wsize = file_write(f->fh, f->buf, rsize);
			if (wsize != rsize) {
				return(FAILURE);
			}
			p1 += size;
			p2 += size;
		}
		file_write(f->fh, f->buf, 0);
	}
	f->pos = file_seek(f->fh, f->headpos, FSEEK_SET);
	return(SUCCESS);
}

static void filew1(PROF_TBL *f, char c) {

	file_write(f->fh, &c, 1);
}


static BOOL insertitem(PROF_TBL *f, char *app, char *key, 
													char *str, BOOL change) {

	int		appsize;
	int		keysize;
	int		strsize;
	int		qout = 0;
	DWORD	baklen = 0;
	DWORD	wrlen = 0;

	app = search_head(app);
	appsize = search_size(app);
	if (!appsize) {
		return(FAILURE);
	}
	key = search_head(key);
	keysize = search_size(key);
	if (!keysize) {
		return(FAILURE);
	}
	if (str) {
		strsize = strlen(str);
	}
	else {
		str = "";
		strsize = 0;
	}

	wrlen = keysize + 1 + strsize + MES_CRLEN;
	if (strsize) {
		if (((BYTE)str[0] <= 0x20) ||
			((BYTE)str[strsize - 1] <= 0x20) ||
			(((BYTE)str[0] == '\"') && ((BYTE)str[strsize - 1] == '\"'))) {
			qout = 1;
			wrlen += 2;
		}
	}
	if (change) {
		baklen = (f->pos) - (f->headpos);
		if (fileinsert(f, baklen, wrlen)) {
			return(FAILURE);
		}
	}
	else if (!(f->appfound)) {
		wrlen += appsize + 2 + MES_CRLEN;
		if ((f->eof) < 0) {
			wrlen += MES_CRLEN;
		}
		if (f->headpos) {
			wrlen += MES_CRLEN;
		}
		if (fileinsert(f, 0, wrlen)) {
			return(FAILURE);
		}
		if ((f->eof) < 0) {
			file_write(f->fh, MES_CR, MES_CRLEN);
		}
		if (f->headpos) {
			file_write(f->fh, MES_CR, MES_CRLEN);
		}
		filew1(f, '[');
		file_write(f->fh, app, appsize);
		filew1(f, ']');
		file_write(f->fh, MES_CR, MES_CRLEN);
	}
	else {
		if ((f->eof) < 0) {
			wrlen += MES_CRLEN;
		}
		if (fileinsert(f, 0, wrlen)) {
			return(FAILURE);
		}
		if ((f->eof) < 0) {
			file_write(f->fh, MES_CR, MES_CRLEN);
		}
	}
	file_write(f->fh, key, keysize);
	filew1(f, '=');
	if (qout) {
		filew1(f, '\"');
	}
	file_write(f->fh, str, strsize);
	if (qout) {
		filew1(f, '\"');
	}
	file_write(f->fh, MES_CR, MES_CRLEN);
	return(SUCCESS);
}

// --------------------------------------------------------------------------

BOOL profile_getstr(char *appname, char *key, char *def, char *out,
												int size, char *fname) {

	BOOL		ret;
	PROF_TBL	f;
	char		*p;
	int			len;

	ret = FALSE;
	if ((!appname) || (!key) || (!fname)) {
		return(FALSE);
	}
	_prf_construct(&f, fname, 0);
	p = keycheck(&f, appname, key);
	if (p) {
		ret = TRUE;
		p = search_head(p);
		len = search_size(p);
		if ((len >= 2) && (p[0] == '\"') && (p[len - 1] == '\"')) {
			p++;
			len -= 2;
		}
	}
	else {
		p = def;
		if (!def) {
			p = "";
		}
		len = strlen(def);
	}
	if ((out) && (size)) {
		size--;
		if (len > size) {
			len = size;
		}
		if (len) {
			CopyMemory(out, p, len);
		}
		out[len] = '\0';
	}
	_prf_destruct(&f);
	return(ret);
}

int profile_getint(char *appname, char *key, int def, char *fname) {

	int			ret;
	PROF_TBL	f;
	char		*p;
	int			len;

	ret = def;
	if ((!appname) || (!key) || (!fname)) {
		return(FALSE);
	}
	_prf_construct(&f, fname, 0);
	p = keycheck(&f, appname, key);
	if (p) {
		p = search_head(p);
		len = search_size(p);
		if ((len >= 2) && (p[0] == '\"') && (p[len - 1] == '\"')) {
			p++;
			len -= 2;
		}
		if (a2iex(p, len, &ret)) {
			ret = def;
		}
	}
	_prf_destruct(&f);
	return(ret);
}


BOOL profile_putstr(char *appname, char *key, char *str, char *fname) {

	BOOL		ret;
	PROF_TBL	f;
	char		*p;

	ret = FALSE;
	if ((!appname) || (!key) || (!fname)) {
		return(FALSE);
	}
	_prf_construct(&f, fname, 1);
	if (!(f.access & 1)) {
		p = keycheck(&f, appname, key);
		ret = insertitem(&f, appname, key, str, (p != NULL));
	}
	_prf_destruct(&f);
	return(ret);
}
