/*	$Id: ctrl_trans.c,v 1.2 2003/07/22 15:09:57 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "cpu.h"
#include "ia32.mcr"

#include "ctrl_trans.h"


/*
 * JMP
 */
void
JMP_Jb(void)
{
	DWORD ip;

	CPU_CLOCK(7);
	GET_PCBYTESD(ip);
	ADD_EIP(ip);
}

void
JMP_Jw(void)
{
	WORD ip;

	CPU_CLOCK(7);
	GET_PCWORDS(ip);
	ADD_EIP(ip);
}

void
JMP_Jd(void)
{
	DWORD ip;

	CPU_CLOCK(7);
	GET_PCDWORD(ip);
	ADD_EIP(ip);
}

void
JMP_Ew(DWORD op)
{
	DWORD madr;
	WORD new_ip;

	if (op >= 0xc0) {
		CPU_CLOCK(7);
		new_ip = *(reg16_b20[op]);
		SET_EIP(new_ip);
	} else {
		CPU_CLOCK(11);
		madr = calc_ea_dst(op);
		new_ip = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		SET_EIP(new_ip);
	}
}

void
JMP_Ed(DWORD op)
{
	DWORD madr;
	DWORD new_ip;

	if (op >= 0xc0) {
		CPU_CLOCK(7);
		new_ip = *(reg32_b20[op]);
		SET_EIP(new_ip);
	} else {
		CPU_CLOCK(11);
		madr = calc_ea_dst(op);
		new_ip = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		SET_EIP(new_ip);
	}
}

void
JMP16_Ap(void)
{
	WORD new_ip;
	WORD cs;

	CPU_CLOCK(11);
	GET_PCWORD(new_ip);
	GET_PCWORD(cs);
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* Real mode or VM86 mode */
		CPU_SET_SEGREG(CPU_CS_INDEX, cs);
		SET_EIP(new_ip);
	} else {
		/* Protected mode */
		/* XXX */
	}
}

void
JMP32_Ap(void)
{
	DWORD new_ip;
	WORD cs;

	CPU_CLOCK(11);
	GET_PCDWORD(new_ip);
	GET_PCWORD(cs);
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* Real mode or VM86 mode */
		CPU_SET_SEGREG(CPU_CS_INDEX, cs);
		SET_EIP(new_ip);
	} else {
		/* Protected mode */
		/* XXX */
	}
}

void
JMP16_Ep(DWORD op)
{
	DWORD madr;
	WORD new_ip;
	WORD new_cs;

	CPU_CLOCK(11);
	if (op < 0xc0) {
		madr = get_ea(op);
		new_ip = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		new_cs = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr + 2);
		if (!CPU_INST_PM || CPU_INST_VM86) {
			/* Real mode or VM86 mode */
			CPU_SET_SEGREG(CPU_CS_INDEX, new_cs);
			SET_EIP(new_ip);
		} else {
			/* Protected mode */
			/* XXX */
		}
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
JMP32_Ep(DWORD op)
{
	DWORD madr;
	DWORD new_ip;
	WORD new_cs;

	CPU_CLOCK(11);
	if (op < 0xc0) {
		madr = get_ea(op);
		new_ip = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		new_cs = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr + 4);
		if (!CPU_INST_PM || CPU_INST_VM86) {
			/* Real mode or VM86 mode */
			CPU_SET_SEGREG(CPU_CS_INDEX, new_cs);
			SET_EIP(new_ip);
		} else {
			/* Protected mode */
			/* XXX */
		}
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

/*
 * Jcc
 */
/* jo */
void
JO_Jb(void)
{

	if (!CPU_OV) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JO_Jw(void)
{

	if (!CPU_OV) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JO_Jd(void)
{

	if (!CPU_OV) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jno */
void
JNO_Jb(void)
{

	if (CPU_OV) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JNO_Jw(void)
{

	if (CPU_OV) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JNO_Jd(void)
{

	if (CPU_OV) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jc */
void
JC_Jb(void)
{

	if (!(CPU_FLAGL & C_FLAG)) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JC_Jw(void)
{

	if (!(CPU_FLAGL & C_FLAG)) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JC_Jd(void)
{

	if (!(CPU_FLAGL & C_FLAG)) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jnc */
void
JNC_Jb(void)
{

	if (CPU_FLAGL & C_FLAG) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}
void
JNC_Jw(void)
{

	if (CPU_FLAGL & C_FLAG) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}
void
JNC_Jd(void)
{

	if (CPU_FLAGL & C_FLAG) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jz */
void
JZ_Jb(void)
{

	if (!(CPU_FLAGL & Z_FLAG)) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JZ_Jw(void)
{

	if (!(CPU_FLAGL & Z_FLAG)) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JZ_Jd(void)
{

	if (!(CPU_FLAGL & Z_FLAG)) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jnz */
void
JNZ_Jb(void)
{

	if (CPU_FLAGL & Z_FLAG) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JNZ_Jw(void)
{

	if (CPU_FLAGL & Z_FLAG) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JNZ_Jd(void)
{

	if (CPU_FLAGL & Z_FLAG) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jna */
void
JNA_Jb(void)
{

	if (!(CPU_FLAGL & (Z_FLAG | C_FLAG))) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JNA_Jw(void)
{

	if (!(CPU_FLAGL & (Z_FLAG | C_FLAG))) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JNA_Jd(void)
{

	if (!(CPU_FLAGL & (Z_FLAG | C_FLAG))) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* ja */
void
JA_Jb(void)
{

	if (CPU_FLAGL & (Z_FLAG | C_FLAG)) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JA_Jw(void)
{

	if (CPU_FLAGL & (Z_FLAG | C_FLAG)) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JA_Jd(void)
{

	if (CPU_FLAGL & (Z_FLAG | C_FLAG)) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* js */
void
JS_Jb(void)
{

	if (!(CPU_FLAGL & S_FLAG)) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JS_Jw(void)
{

	if (!(CPU_FLAGL & S_FLAG)) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JS_Jd(void)
{

	if (!(CPU_FLAGL & S_FLAG)) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jns */
void
JNS_Jb(void)
{

	if (CPU_FLAGL & S_FLAG) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JNS_Jw(void)
{

	if (CPU_FLAGL & S_FLAG) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JNS_Jd(void)
{

	if (CPU_FLAGL & S_FLAG) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jp */
void
JP_Jb(void)
{

	if (!(CPU_FLAGL & P_FLAG)) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JP_Jw(void)
{

	if (!(CPU_FLAGL & P_FLAG)) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JP_Jd(void)
{

	if (!(CPU_FLAGL & P_FLAG)) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jnp */
void
JNP_Jb(void)
{

	if (CPU_FLAGL & P_FLAG) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JNP_Jw(void)
{

	if (CPU_FLAGL & P_FLAG) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JNP_Jd(void)
{

	if (CPU_FLAGL & P_FLAG) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jl */
void
JL_Jb(void)
{

	if ((!(CPU_FLAGL & S_FLAG)) == (!CPU_OV)) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JL_Jw(void)
{

	if ((!(CPU_FLAGL & S_FLAG)) == (!CPU_OV)) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JL_Jd(void)
{

	if ((!(CPU_FLAGL & S_FLAG)) == (!CPU_OV)) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jnl */
void
JNL_Jb(void)
{

	if ((!(CPU_FLAGL & S_FLAG)) != (!CPU_OV)) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JNL_Jw(void)
{

	if ((!(CPU_FLAGL & S_FLAG)) != (!CPU_OV)) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JNL_Jd(void)
{

	if ((!(CPU_FLAGL & S_FLAG)) != (!CPU_OV)) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jle */
void
JLE_Jb(void)
{

	if ((!(CPU_FLAGL & Z_FLAG)) && ((!(CPU_FLAGL & S_FLAG)) == (!CPU_OV))) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JLE_Jw(void)
{

	if ((!(CPU_FLAGL & Z_FLAG)) && ((!(CPU_FLAGL & S_FLAG)) == (!CPU_OV))) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JLE_Jd(void)
{

	if ((!(CPU_FLAGL & Z_FLAG)) && ((!(CPU_FLAGL & S_FLAG)) == (!CPU_OV))) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jnle */
void
JNLE_Jb(void)
{

	if ((CPU_FLAGL & Z_FLAG) || ((!(CPU_FLAGL & S_FLAG)) != (!CPU_OV))) {
		JMPNOP(2, 1);
	} else {
		JMPSHORT(7);
	}
}

void
JNLE_Jw(void)
{

	if ((CPU_FLAGL & Z_FLAG) || ((!(CPU_FLAGL & S_FLAG)) != (!CPU_OV))) {
		JMPNOP(2, 2);
	} else {
		JMPNEAR(7);
	}
}

void
JNLE_Jd(void)
{

	if ((CPU_FLAGL & Z_FLAG) || ((!(CPU_FLAGL & S_FLAG)) != (!CPU_OV))) {
		JMPNOP(2, 4);
	} else {
		JMPNEAR_4(7);
	}
}

/* jcxz */
void
JeCXZ_Jb(void)
{

	if (!CPU_INST_AS32) {
		if (CPU_CX) {
			JMPNOP(4, 1);
		} else {
			JMPSHORT(8);
		}
	} else {
		if (CPU_ECX) {
			JMPNOP(4, 1);
		} else {
			JMPSHORT(8);
		}
	}
}

/*
 * LOOPcc
 */
/* loopne */
void
LOOPNE_Jb(void)
{

	if (!CPU_INST_AS32) {
		CPU_CX--;
		if (CPU_CX == 0 || (CPU_FLAGL & Z_FLAG)) {
			JMPNOP(4, 1);
		} else {
			JMPSHORT(8);
		}
	} else {
		CPU_ECX--;
		if (CPU_ECX == 0 || (CPU_FLAGL & Z_FLAG)) {
			JMPNOP(4, 1);
		} else {
			JMPSHORT(8);
		}
	}
}

/* loope */
void
LOOPE_Jb(void)
{

	if (!CPU_INST_AS32) {
		CPU_CX--;
		if (CPU_CX == 0 || !(CPU_FLAGL & Z_FLAG)) {
			JMPNOP(4, 1);
		} else {
			JMPSHORT(8);
		}
	} else {
		CPU_ECX--;
		if (CPU_ECX == 0 || !(CPU_FLAGL & Z_FLAG)) {
			JMPNOP(4, 1);
		} else {
			JMPSHORT(8);
		}
	}
}

/* loop */
void
LOOP_Jb(void)
{

	if (!CPU_INST_AS32) {
		CPU_CX--;
		if (CPU_CX == 0) {
			JMPNOP(4, 1);
		} else {
			JMPSHORT(8);
		}
	} else {
		CPU_ECX--;
		if (CPU_ECX == 0) {
			JMPNOP(4, 1);
		} else {
			JMPSHORT(8);
		}
	}
}

/*
 * CALL
 */
void
CALL_Aw(void)
{
	DWORD ip;

	CPU_CLOCK(7);
	GET_PCWORDS(ip);
	PUSH0_16(CPU_IP);
	ADD_EIP(ip);
}

void
CALL_Ad(void)
{
	DWORD ip;

	CPU_CLOCK(7);
	GET_PCDWORD(ip);
	PUSH0_32(CPU_EIP);
	ADD_EIP(ip);
}

void
CALL_Ew(DWORD op)
{
	DWORD madr;
	WORD new_ip;

	if (op >= 0xc0) {
		CPU_CLOCK(7);
		new_ip = *(reg16_b20[op]);
		PUSH0_16(CPU_IP);
		SET_EIP(new_ip);
	} else {
		CPU_CLOCK(11);
		madr = calc_ea_dst(op);
		new_ip = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		PUSH0_16(CPU_IP);
		SET_EIP(new_ip);
	}
}

void
CALL_Ed(DWORD op)
{
	DWORD madr;
	DWORD new_ip;

	if (op >= 0xc0) {
		CPU_CLOCK(7);
		new_ip = *(reg32_b20[op]);
		PUSH0_32(CPU_EIP);
		SET_EIP(new_ip);
	} else {
		CPU_CLOCK(11);
		madr = calc_ea_dst(op);
		new_ip = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr);
		PUSH0_32(CPU_EIP);
		SET_EIP(new_ip);
	}
}

void
CALL16_Ap(void)
{
	WORD new_ip;
	WORD cs;

	CPU_CLOCK(13);
	GET_PCWORD(new_ip);
	GET_PCWORD(cs);
	PUSH0_16(CPU_CS);
	PUSH0_16(CPU_IP);
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* Real mode or VM86 mode */
		CPU_SET_SEGREG(CPU_CS_INDEX, cs);
		SET_EIP(new_ip);
	} else {
		/* Protected mode */
		/* XXX */
	}
}

void
CALL32_Ap(void)
{
	DWORD new_ip;
	WORD cs;

	CPU_CLOCK(13);
	GET_PCDWORD(new_ip);
	GET_PCWORD(cs);
	PUSH0_32(CPU_CS);
	PUSH0_32(CPU_EIP);
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* Real mode or VM86 mode */
		CPU_SET_SEGREG(CPU_CS_INDEX, cs);
		SET_EIP(new_ip);
	} else {
		/* Protected mode */
		/* XXX */
	}
}

void
CALL16_Ep(DWORD op)
{
	DWORD ad;
	WORD new_ip;
	WORD cs;

	CPU_CLOCK(16);
	if (op < 0xc0) {
		ad = get_ea(op);
		PUSH0_16(CPU_CS);
		PUSH0_16(CPU_IP);
		new_ip = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, ad);
		cs = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, ad + 2);
		if (!CPU_INST_PM || CPU_INST_VM86) {
			/* Real mode or VM86 mode */
			CPU_SET_SEGREG(CPU_CS_INDEX, cs);
			SET_EIP(new_ip);
		} else {
			/* Protected mode */
			/* XXX */
		}
		return;
	} 
	EXCEPTION(UD_EXCEPTION, 0);
}

void
CALL32_Ep(DWORD op)
{
	DWORD ad;
	DWORD new_ip;
	WORD cs;

	CPU_CLOCK(16);
	if (op < 0xc0) {
		ad = get_ea(op);
		PUSH0_32(CPU_CS);
		PUSH0_32(CPU_EIP);
		new_ip = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, ad);
		cs = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, ad + 4);
		if (!CPU_INST_PM || CPU_INST_VM86) {
			/* Real mode or VM86 mode */
			CPU_SET_SEGREG(CPU_CS_INDEX, cs);
			SET_EIP(new_ip);
		} else {
			/* Protected mode */
			/* XXX */
		}
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

/*
 * RET
 */
void
RETnear16(void)
{
	WORD new_ip;

	CPU_CLOCK(11);
	POP0_16(new_ip);
	SET_EIP(new_ip);
}

void
RETnear32(void)
{
	DWORD new_ip;

	CPU_CLOCK(11);
	POP0_32(new_ip);
	SET_EIP(new_ip);
}

void
RETnear16_Iw(void)
{
	WORD new_ip;
	WORD ad;

	CPU_CLOCK(11);
	GET_PCWORD(ad);
	POP0_16(new_ip);
	SET_EIP(new_ip);
	CPU_SP += ad;
}

void
RETnear32_Iw(void)
{
	DWORD new_ip;
	DWORD ad;

	CPU_CLOCK(11);
	GET_PCDWORD(ad);
	POP0_32(new_ip);
	SET_EIP(new_ip);
	CPU_ESP += ad;
}

void
RETfar16(void)
{
	WORD ip;
	WORD cs;

	CPU_CLOCK(15);
	POP0_16(ip);
	POP0_16(cs);
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* Real mode or VM86 mode */
		CPU_SET_SEGREG(CPU_CS_INDEX, cs);
		SET_EIP(ip);
	} else {
		/* Protected mode */
		/* XXX */
	}
}

void
RETfar32(void)
{
	DWORD new_ip;
	WORD cs;

	CPU_CLOCK(15);
	POP0_32(new_ip);
	POP0_32(cs);
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* Real mode or VM86 mode */
		CPU_SET_SEGREG(CPU_CS_INDEX, cs);
		SET_EIP(new_ip);
	} else {
		/* Protected mode */
		/* XXX */
	}
}

void
RETfar16_Iw(void)
{
	WORD new_ip;
	WORD ad;
	WORD cs;

	CPU_CLOCK(15);
	GET_PCWORD(ad);
	POP0_16(new_ip);
	POP0_16(cs);
	CPU_SP += ad;
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* Real mode or VM86 mode */
		CPU_SET_SEGREG(CPU_CS_INDEX, cs);
		SET_EIP(new_ip);
	} else {
		/* Protected mode */
		/* XXX */
	}
}

void
RETfar32_Iw(void)
{
	DWORD new_ip;
	DWORD ad;
	WORD cs;

	CPU_CLOCK(15);
	GET_PCDWORD(ad);
	POP0_32(new_ip);
	POP0_32(cs);
	CPU_ESP += ad;
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* Real mode or VM86 mode */
		CPU_SET_SEGREG(CPU_CS_INDEX, cs);
		SET_EIP(new_ip);
	} else {
		/* Protected mode */
		/* XXX */
	}
}

void
IRET(void)
{
	WORD new_ip;
	WORD flag;
	WORD cs;

	CPU_CLOCK(31);
	POP0_16(new_ip);
	POP0_16(cs);
	POP0_16(flag);
	CPU_FLAG = (flag & 0x7fd5) | 0x2; 
	CPU_OV = CPU_FLAG & O_FLAG;
	CPU_TRAP = (CPU_FLAG & (I_FLAG|T_FLAG)) == (I_FLAG|T_FLAG);
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* Real mode or VM86 mode */
		CPU_SET_SEGREG(CPU_CS_INDEX, cs);
		SET_EIP(new_ip);
		IRQCHECKTERM();
	} else {
		/* Protected mode */
		/* XXX */
		IRQCHECKTERM();
	}
}

void
IRETD(void)
{
	DWORD new_ip;
	DWORD flag;
	WORD cs;

	CPU_CLOCK(31);
	POP0_32(new_ip);
	POP0_32(cs);
	POP0_32(flag);
	CPU_EFLAG = (flag & 0x00277fd5) | 0x2; 
	CPU_OV = CPU_FLAG & O_FLAG;
	CPU_TRAP = (CPU_FLAG & (I_FLAG|T_FLAG)) == (I_FLAG|T_FLAG);
	if (!CPU_INST_PM || CPU_INST_VM86) {
		/* Real mode or VM86 mode */
		CPU_SET_SEGREG(CPU_CS_INDEX, cs);
		SET_EIP(new_ip);
		IRQCHECKTERM();
	} else {
		/* Protected mode */
		/* XXX */
		IRQCHECKTERM();
	}
}

/*
 * INT
 */
void
INT1(void)
{

	CPU_CLOCK(23);
	INTERRUPT(1, TRUE, FALSE, 0);
}

void
INT3(void)
{

	CPU_CLOCK(23);
	INTERRUPT(3, TRUE, FALSE, 0);
}

void
INTO(void)
{

	if (!CPU_OV) {
		CPU_CLOCK(4);
		return;
	}
	CPU_CLOCK(24);
	INTERRUPT(4, TRUE, FALSE, 0);
}

void
INT_Ib(void)
{
	BYTE vect;

	CPU_CLOCK(23);
	if (!CPU_INST_PM || !CPU_INST_VM86 || CPU_INST_IOPL == CPU_IOPL3) {
		GET_PCBYTE(vect);
		INTERRUPT(vect, TRUE, FALSE, 0);
		return;
	}
	EXCEPTION(GP_EXCEPTION, 0);
}

void
BOUND_GwMa(void)
{
	DWORD op, madr;
	WORD reg;
	int vect;

	CPU_CLOCK(13);
	GET_PCBYTE(op);
	if (op < 0xc0) {
		reg = *reg16_b53[op];
		madr = calc_ea_dst(op);
		if (reg >= cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr) &&
		    reg <= cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr + 2)) {
				return;
		}
		vect = BR_EXCEPTION;
	} else {
		vect = UD_EXCEPTION;
	}
	EXCEPTION(vect, 0);
}

void
BOUND_GdMa(void)
{
	DWORD op, madr;
	DWORD reg;
	int vect;

	CPU_CLOCK(13);
	GET_PCBYTE(op);
	if (op < 0xc0) {
		reg = *reg32_b53[op];
		madr = calc_ea_dst(op);
		if (reg >= cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr) &&
		    reg <= cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr + 4)) {
				return;
		}
		vect = BR_EXCEPTION;
	} else {
		vect = UD_EXCEPTION;
	}
	EXCEPTION(vect, 0);
}

/*
 * STACK
 */
void
ENTER_IwIb(void)
{
	WORD dimsize;
	BYTE level;

	GET_PCWORD(dimsize);
	GET_PCBYTE(level);
	level &= 0x1f;

	XPUSH0(CPU_EBP);
	if (level == 0) {			/* enter level=0 */
		CPU_CLOCK(11);
		if (!CPU_INST_OP32)
			CPU_BP = CPU_SP;
		else
			CPU_EBP = CPU_ESP;
		if (!CPU_INST_SS32)
			CPU_SP -= dimsize;
		else
			CPU_ESP -= dimsize;
	} else {
		--level;
		if (level == 0) {		/* enter level=1 */
			CPU_CLOCK(15);
			if (!CPU_INST_OP32) {
				WORD tmp = CPU_SP;
				if (!CPU_INST_SS32) {
					REGPUSH0(tmp);
					CPU_BP = tmp;
					CPU_SP -= dimsize;
				} else {
					REGPUSH0_16_32(tmp);
					CPU_EBP = tmp;
					CPU_ESP -= dimsize;
				}
			} else {
				DWORD tmp;
				if (!CPU_INST_SS32) {
					tmp = CPU_SP;
					REGPUSH0_32_16(tmp);
					CPU_BP = (WORD)tmp;
					CPU_SP -= dimsize;
				} else {
					tmp = CPU_ESP;
					REGPUSH0_32(tmp);
					CPU_EBP = tmp;
					CPU_ESP -= dimsize;
				}
			}
		} else {			/* enter level=2-31 */
			CPU_CLOCK(12 + level * 4);
			if (!CPU_INST_OP32) {
				WORD val;

				if (!CPU_INST_SS32) {
					WORD bp = CPU_BP;
					CPU_BP = CPU_SP;
					while (level--) {
						bp -= 2;
						val = cpu_vmemoryread_w(CPU_SS_INDEX, bp);
						REGPUSH0(val);
					}
					REGPUSH0(CPU_BP);
					CPU_SP -= dimsize;
				} else {
					DWORD bp = CPU_EBP;
					CPU_EBP = CPU_ESP;
					while (level--) {
						bp -= 2;
						val = cpu_vmemoryread_w(CPU_SS_INDEX, bp);
						REGPUSH0_16_32(val);
					}
					REGPUSH0_16_32(CPU_EBP);
					CPU_ESP -= dimsize;
				}
			} else {
				DWORD val;

				if (!CPU_INST_SS32) {
					WORD bp = CPU_BP;
					CPU_BP = CPU_SP;
					while (level--) {
						bp -= 4;
						val = cpu_vmemoryread_d(CPU_SS_INDEX, bp);
						REGPUSH0_32_16(val);
					}
					REGPUSH0_32_16(CPU_BP);
					CPU_SP -= dimsize;
				} else {
					DWORD bp = CPU_EBP;
					CPU_EBP = CPU_ESP;
					while (level--) {
						bp -= 4;
						val = cpu_vmemoryread_d(CPU_SS_INDEX, bp);
						REGPUSH0_32(val);
					}
					REGPUSH0_32(CPU_EBP);
					CPU_ESP -= dimsize;
				}
			}
		}
	}
}

void
LEAVE16(void)
{

	CPU_CLOCK(5);

	/* XXX sp check */

	if (!CPU_INST_SS32) {
		CPU_SP = CPU_BP;
	} else {
		CPU_ESP = CPU_EBP;
	}
	REGPOP0(CPU_BP);
}

void
LEAVE32(void)
{

	CPU_CLOCK(5);

	/* XXX sp check */

	if (!CPU_INST_SS32) {
		CPU_SP = CPU_BP;
	} else {
		CPU_ESP = CPU_EBP;
	}
	REGPOP0_32(CPU_EBP);
}
