/*	$Id: flag_ctrl.c,v 1.2 2003/07/22 15:09:57 nonaka Exp $	*/

/*
 * Copyright (c) 2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "cpu.h"
#include "ia32.mcr"

#include "flag_ctrl.h"


/*
 * flag contol instructions
 */
void
STC(void)
{

	CPU_CLOCK(2);
	CPU_FLAGL |= C_FLAG;
}

void
CLC(void)
{

	CPU_CLOCK(2);
	CPU_FLAGL &= ~C_FLAG;
}

void
CMC(void)
{

	CPU_CLOCK(2);
	CPU_FLAGL ^= C_FLAG;
}

void
CLD(void)
{

	CPU_CLOCK(2);
	CPU_FLAG &= ~D_FLAG;
}

void
STD(void)
{

	CPU_CLOCK(2);
	CPU_FLAG |= D_FLAG;
}

void
LAHF(void)
{

	CPU_CLOCK(2);
	CPU_AH = (CPU_FLAGL & 0xd5) | 0x2;	/* SZ0A0P1C */
}

void
SAHF(void)
{

	CPU_CLOCK(2);
	CPU_FLAGL = (CPU_AH & 0xd5) | 0x2;	/* SZ0A0P1C */
}

/*
 * PUSHF/POPF
 */
void
PUSHF_Fw(void)
{

	CPU_CLOCK(3);
	if (!CPU_INST_PM || !CPU_INST_VM86 || (CPU_INST_CPL == 3)) {
		WORD flags = REAL_FLAGREG;
		flags &= 0x7fd5;
		flags |= 0x0002;
		PUSH0_16(flags);
		return;
	}
	/* VM86 && IOPL != 3 */
	EXCEPTION(GP_EXCEPTION, 0);
}

void
PUSHFD_Fd(void)
{

	CPU_CLOCK(3);
	if (!CPU_INST_PM || !CPU_INST_VM86 || (CPU_INST_CPL == 3)) {
		DWORD flags = REAL_EFLAGREG;
		flags &= 0x00fcffff;
		flags &= 0x003f7fd5;
		flags |= 0x00000002;
		PUSH0_32(flags);
		return;
	}
	/* VM86 && IOPL != 3 */
	EXCEPTION(GP_EXCEPTION, 0);
}

void
POPF_Fw(void)
{
	WORD flags;

	CPU_CLOCK(3);
	if (CPU_INST_PM && CPU_INST_VM86 && (CPU_INST_IOPL != CPU_IOPL3)) {
		EXCEPTION(GP_EXCEPTION, 0);
		return;
	}

	POP0_16(flags);
	if (!CPU_INST_PM || CPU_INST_VM86 || (CPU_INST_CPL != 0)) {
		/* real-mode || (VM86 && IOPL == 3) || (!VM86 && IOPL != 0) */
		flags &= ~IOPL_FLAG;
		flags |= (CPU_FLAG & IOPL_FLAG);
	}
	CPU_FLAG = (flags & 0x7fd5) | 0x2;
	CPU_OV = CPU_FLAG & O_FLAG;
	CPU_TRAP = (CPU_FLAG & (I_FLAG|T_FLAG)) == (I_FLAG|T_FLAG);
	IRQCHECKTERM();
}

void
POPFD_Fd(void)
{
	DWORD flags;

	CPU_CLOCK(3);
	if (CPU_INST_PM) {
		if (!CPU_INST_VM86) {
			POP0_32(flags);
			if (CPU_INST_CPL == 0) {
				flags &= ~(VM_FLAG|VIP_FLAG|VIF_FLAG);
				flags |= CPU_EFLAG & VM_FLAG;
			} else {
				flags &= ~(VM_FLAG|IOPL_FLAG|VIP_FLAG|VIF_FLAG);
				flags |= CPU_EFLAG & (VM_FLAG|IOPL_FLAG);
			}
		} else {
			/* VM86 */
			if (CPU_INST_IOPL == CPU_IOPL3) {
				POP0_32(flags);
				flags &= ~(VM_FLAG|RF_FLAG|IOPL_FLAG|VIP_FLAG|VIF_FLAG);
				flags |= CPU_EFLAG & (VM_FLAG|RF_FLAG|IOPL_FLAG|VIP_FLAG|VIF_FLAG);
			} else {
				EXCEPTION(GP_EXCEPTION, 0);
				/* NOTREACHED*/
				return;
			}
		}
	} else {
		/* real-mode */
		POP0_32(flags);
		flags &= ~(VIF_FLAG|IOPL_FLAG|VIP_FLAG|VIF_FLAG);
		flags |= CPU_EFLAG & (VM_FLAG|IOPL_FLAG);
	}
	CPU_EFLAG = (flags & 0x3f7fd5) | 0x2;
	CPU_OV = CPU_FLAG & O_FLAG;
	CPU_TRAP = (CPU_FLAG & (I_FLAG|T_FLAG)) == (I_FLAG|T_FLAG);
	if (CPU_EFLAG & VM_FLAG) {
		change_vm();
	}
	IRQCHECKTERM();
}

void
STI(void)
{

	CPU_CLOCK(2);
	CPU_FLAG |= I_FLAG;
	CPU_TRAP = (CPU_FLAG & (I_FLAG|T_FLAG)) == (I_FLAG|T_FLAG);
	exec_1step();
	IRQCHECKTERM();
}

void
CLI(void)
{

	CPU_CLOCK(2);
	CPU_FLAG &= ~I_FLAG;
	CPU_TRAP = 0;
}
