/*	$Id: system_inst.c,v 1.2 2003/07/22 15:09:57 nonaka Exp $	*/

/*
 * Copyright (c) 2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "cpu.h"
#include "ia32.mcr"

#include "system_inst.h"


void
LGDT16_Ms(DWORD op)
{
	DWORD madr;
	DWORD base;
	WORD limit;

	GET_PCBYTE(op);
	if (op < 0xc0) {
		CPU_CLOCK(11);
		madr = get_ea(op);
		limit = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		base = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr + 2);
		base &= 0x00ffffff;
		CPU_GDTR_BASE = base;
		CPU_GDTR_LIMIT = limit;
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
LGDT32_Ms(DWORD op)
{
	DWORD madr;
	DWORD base;
	WORD limit;

	GET_PCBYTE(op);
	if (op < 0xc0) {
		CPU_CLOCK(11);
		madr = get_ea(op);
		limit = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		base = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr + 2);
		CPU_GDTR_BASE = base;
		CPU_GDTR_LIMIT = limit;
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
SGDT16_Ms(DWORD op)
{
	DWORD madr;
	DWORD base;
	WORD limit;

	GET_PCBYTE(op);
	if (op < 0xc0) {
		CPU_CLOCK(11);
		base = CPU_GDTR_BASE & 0x00ffffff;
		limit = CPU_GDTR_LIMIT;
		madr = get_ea(op);
		cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, madr, limit);
		cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, madr + 2, base);
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
SGDT32_Ms(DWORD op)
{
	DWORD madr;
	DWORD base;
	WORD limit;

	if (op < 0xc0) {
		CPU_CLOCK(11);
		base = CPU_GDTR_BASE;
		limit = CPU_GDTR_LIMIT;
		madr = get_ea(op);
		cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, madr, limit);
		cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, madr + 2, base);
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
LLDT_Ew(DWORD op)
{
	DWORD src, madr;

	if (CPU_INST_PM && !CPU_INST_VM86) {
		if (CPU_INST_CPL != 0) {
			EXCEPTION(GP_EXCEPTION, 0);
			/*NOTREACHED*/
			return;
		}

		if (op >= 0xc0) {
			CPU_CLOCK(5);
			src = *reg16_b20[op];
		} else {
			CPU_CLOCK(11);
			madr = calc_ea_dst(op);
			src = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		}
		load_ldtr(src);
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
SLDT_Ew(DWORD op)
{
	DWORD madr;

	if (op >= 0xc0) {
		CPU_CLOCK(5);
		*reg16_b20[op] = CPU_LDTR;
	} else {
		CPU_CLOCK(11);
		madr = calc_ea_dst(op);
		cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, madr, CPU_LDTR);
	}
}

void
SLDT_Ed(DWORD op)
{
	DWORD madr;

	if (op >= 0xc0) {
		CPU_CLOCK(5);
		*reg32_b20[op] = CPU_LDTR;
	} else {
		CPU_CLOCK(11);
		madr = calc_ea_dst(op);
		cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, madr, CPU_LDTR);
	}
}

void
LTR_Ew(DWORD op)
{
	DWORD src, madr;

	if (CPU_INST_PM && !CPU_INST_VM86) {
		if (CPU_INST_CPL != 0) {
			EXCEPTION(GP_EXCEPTION, 0);
			/*NOTREACHED*/
			return;
		}

		if (op >= 0xc0) {
			CPU_CLOCK(5);
			src = *reg16_b20[op];
		} else {
			CPU_CLOCK(11);
			madr = calc_ea_dst(op);
			src = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		}
		load_tr(src);
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
STR_Ew(DWORD op)
{
	DWORD madr;

	if (op >= 0xc0) {
		CPU_CLOCK(5);
		*reg16_b20[op] = CPU_TR;
	} else {
		CPU_CLOCK(11);
		madr = calc_ea_dst(op);
		cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, madr, CPU_TR);
	}
}

void
STR_Ed(DWORD op)
{
	DWORD madr;

	if (op >= 0xc0) {
		CPU_CLOCK(5);
		*reg32_b20[op] = CPU_TR;
	} else {
		CPU_CLOCK(11);
		madr = calc_ea_dst(op);
		cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, madr, CPU_TR);
	}
}

void
LIDT16_Ms(DWORD op)
{
	DWORD madr;
	DWORD base;
	WORD limit;

	GET_PCBYTE(op);
	if (op < 0xc0) {
		CPU_CLOCK(11);
		madr = get_ea(op);
		limit = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		base = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr + 2);
		base &= 0x00ffffff;
		CPU_IDTR_BASE = base;
		CPU_IDTR_LIMIT = limit;
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
LIDT32_Ms(DWORD op)
{
	DWORD madr;
	DWORD base;
	WORD limit;

	GET_PCBYTE(op);
	if (op < 0xc0) {
		CPU_CLOCK(11);
		madr = get_ea(op);
		limit = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		base = cpu_vmemoryread_d(CPU_INST_SEGREG_INDEX, madr + 2);
		CPU_IDTR_BASE = base;
		CPU_IDTR_LIMIT = limit;
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
SIDT16_Ms(DWORD op)
{
	DWORD madr;
	DWORD base;
	WORD limit;

	GET_PCBYTE(op);
	if (op < 0xc0) {
		CPU_CLOCK(11);
		base = CPU_IDTR_BASE & 0x00ffffff;
		limit = CPU_IDTR_LIMIT;
		madr = get_ea(op);
		cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, madr, limit);
		cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, madr + 2, base);
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
SIDT32_Ms(DWORD op)
{
	DWORD madr;
	DWORD base;
	WORD limit;

	if (op < 0xc0) {
		CPU_CLOCK(11);
		base = CPU_IDTR_BASE;
		limit = CPU_IDTR_LIMIT;
		madr = get_ea(op);
		cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, madr, limit);
		cpu_vmemorywrite_d(CPU_INST_SEGREG_INDEX, madr + 2, base);
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
MOV_CdRd(void)
{
	DWORD op, src;
	DWORD reg;
	int idx;

	CPU_CLOCK(11);
	GET_PCBYTE(op);
	if (op >= 0xc0) {
		if (CPU_INST_PM && (CPU_INST_VM86 || CPU_INST_CPL != 0)) {
			EXCEPTION(GP_EXCEPTION, 0);
			/*NOTREACHED*/
			return;
		}

		src = *reg32_b20[op];
		idx = (op >> 3) & 7;
		switch (idx) {
		case 0: /* CR0 */
			/*
			 * 0 = PE (protect enable)
			 * 1 = MP (monitor coprocesser)
			 * 2 = EM (emulation)
			 * 3 = TS (task switch)
			 * 4 = ET (extend type, FPU present = 1)
			 * 5 = NE (numeric error)
			 * 16 = WP (write protect)
			 * 18 = AM (alignment mask)
			 * 29 = NW (not write-through)
			 * 30 = CD (cache diable)
			 * 31 = PG (pageing)
			 */
			reg = CPU_CR0;
			src &= 0xe005003f;
#ifndef USE_FPU
			src &= ~CPU_CR0_ET;	/* FPU not present */
#else
			src |= CPU_CR0_ET;	/* FPU present */
#endif
			CPU_CR0 = src;
			if ((reg ^ CPU_CR0) & CPU_CR0_PE) {
				change_pm();
			}
			if ((reg ^ CPU_CR0) & CPU_CR0_PG) {
				change_pg();
			}
			break;

		case 2: /* CR2 */
			CPU_CR2 = src;	/* page fault linear address */
			break;

		case 3: /* CR3 */
			/*
			 * 31-12 = page directory base
			 * 4 = PCD (page level cache diable)
			 * 3 = PWT (page level write throgh)
			 */
			CPU_CR3 = src & 0xfffff018;
			break;

		case 4: /* CR4 */
			/*
			 * 10 = OSXMMEXCPT (support non masking exception by OS)
			 * 9 = OSFXSR (support FXSAVE, FXRSTOR by OS)
			 * 8 = PCE (performance monitoring counter enable)
			 * 7 = PGE (page global enable)
			 * 6 = MCE (machine check enable)
			 * 5 = PAE (physical address extention)
			 * 4 = PSE (page size extention)
			 * 3 = DE (debug extention)
			 * 2 = TSD (time stamp diable)
			 * 1 = PVI (protected mode virtual interrupt)
			 * 0 = VME (VM8086 mode extention)
			 */
			reg = 0;	/* allow */
			if (src & ~reg) {
				if (src & 0xfffffc00) {
					EXCEPTION(GP_EXCEPTION, 0);
					/*NOTREACHED*/
					return;
				}
				ia32_warning("MOV_CdRd: CR4 <- 0x%08x", src);
			}
			CPU_CR4 = src;
			break;

		default:
			ia32_panic("MOV_CdRd: CR reg index (%d)", idx);
			/*NOTREACHED*/
			break;
		}
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
MOV_RdCd(void)
{
	DWORD *out;
	DWORD op;

	CPU_CLOCK(11);
	PREPART_EA_REG32P(op, out);
	if (op >= 0xc0) {
		if (CPU_INST_PM && (CPU_INST_VM86 || CPU_INST_CPL != 0)) {
			EXCEPTION(GP_EXCEPTION, 0);
			/*NOTREACHED*/
			return;
		}

		switch (op & 7) {
		case 0:
			*out = CPU_CR0;
			break;

		case 2:
			*out = CPU_CR2;
			break;

		case 3:
			*out = CPU_CR3;
			break;

		case 4:
			*out = CPU_CR4;
			break;

		default:
			ia32_panic("MOV_RdCd: CR reg index (%d)", op & 7);
			/*NOTREACHED*/
			break;
		}
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
LMSW_Ew(DWORD op)
{
	DWORD src, madr;

	if (CPU_INST_PM && CPU_INST_CPL != 0) {
		EXCEPTION(GP_EXCEPTION, 0);
		/*NOTREACHED*/
		return;
	}

	if (op >= 0xc0) {
		CPU_CLOCK(2);
		src = *reg16_b20[op];
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		src = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
	}

	CPU_CR0 &= ~0xe;	/* can't switch back from protected mode */
	CPU_CR0 |= (src & 0xf);	/* TS, EM, MP, PE */
	if ((src ^ CPU_CR0) & CPU_CR0_PE) {
		change_pm();
	}
}

void
SMSW_Ew(DWORD op)
{
	DWORD madr;

	if (op >= 0xc0) {
		CPU_CLOCK(2);
		*reg16_b20[op] = (WORD)CPU_CR0;
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, madr, (WORD)CPU_CR0);
	}
}

void
SMSW_Ed(DWORD op)
{
	DWORD madr;

	if (op >= 0xc0) {
		CPU_CLOCK(2);
		*reg32_b20[op] = CPU_CR0;
	} else {
		CPU_CLOCK(3);
		madr = calc_ea_dst(op);
		cpu_vmemorywrite_w(CPU_INST_SEGREG_INDEX, madr, (WORD)CPU_CR0);
	}
}

void
CLTS(void)
{

	CPU_CLOCK(2);
	if (CPU_INST_PM && CPU_INST_CPL != 0) {
		EXCEPTION(GP_EXCEPTION, 0);
		/*NOTREACHED*/
		return;
	}
	CPU_CR0 &= ~CPU_CR0_TS;
}

void
ARPL_EwGw(void)
{
	DWORD op, src, dst, madr;

	if (CPU_INST_PM && !CPU_INST_VM86) {
		PREPART_EA_REG16(op, src);
		if (op >= 0xc0) {
			CPU_CLOCK(2);
			dst = *reg16_b20[op];
		} else {
			CPU_CLOCK(3);
			madr = calc_ea_dst(op);
			dst = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		}

		if ((dst & 3) < (src & 3)) {
			dst &= ~3;
			dst |= (src & 3);
			CPU_FLAGL |= Z_FLAG;
		} else {
			CPU_FLAGL &= ~Z_FLAG;
		}
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

/*
 * DPL
 */
void
LAR_GwEw(void)
{
	WORD *out;
	DWORD op;
	WORD selector;

	if (CPU_INST_PM && !CPU_INST_VM86) {
		SEGDESC sd;		/* segment register cache */
		DWORD sd_idx;		/* segment descriptor index */
		DWORD dt_base;		/* descriptor table base addr */
		WORD dt_limit;		/* descriptor table size limit */
		int sd_rpl;		/* segment descriptor RPL */

		PREPART_REG16_EA(op, selector, out, 5, 11);

		sd_idx = selector & ~7;
		sd_rpl = selector & 3;

		/* null segment */
		if (selector < 4) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		/* descriptor table */
		if (selector & CPU_SEGMENT_TABLE_IND) {
			/* LDT */
			if (!CPU_LDTR_DESC.valid) {
				EXCEPTION(GP_EXCEPTION, selector);
				/*NOTREACHED*/
				return;
			}
			dt_base = CPU_LDTR_BASE;
			dt_limit = CPU_LDTR_LIMIT;
		} else {
			/* GDT */
			dt_base = CPU_GDTR_BASE;
			dt_limit = CPU_GDTR_LIMIT;
		}
		if (sd_idx >= dt_limit) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}
		CPU_SET_SEGDESC(sd, dt_base + sd_idx);

		if (!sd.valid) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		if (sd.s) {
			if (!(sd.dc && !(sd.type & CPU_SEGDESC_TYPE_EC)) &&
			    ((sd.dpl < CPU_INST_CPL) || (sd.dpl < sd_rpl))) {
				CPU_FLAGL &= ~Z_FLAG;
				return;
			}
		} else {
			switch (sd.type) {
			case 1: case 2: case 3: case 4: case 5:
			case 9: case 11: case 12:
				break;

			default:
				CPU_FLAGL &= ~Z_FLAG;
				return;
			}
		}

		*out = sd.h & 0xff00;
		CPU_FLAGL |= Z_FLAG;
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
LAR_GdEw(void)
{
	DWORD *out;
	DWORD op;
	WORD selector;

	if (CPU_INST_PM && !CPU_INST_VM86) {
		SEGDESC sd;		/* segment register cache */
		DWORD sd_idx;		/* segment descriptor index */
		DWORD dt_base;		/* descriptor table base addr */
		WORD dt_limit;		/* descriptor table size limit */
		int sd_rpl;		/* segment descriptor RPL */

		PREPART_REG32_EA(op, selector, out, 5, 11);

		sd_idx = selector & ~7;
		sd_rpl = selector & 3;

		/* null segment */
		if (selector < 4) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		/* descriptor table */
		if (selector & CPU_SEGMENT_TABLE_IND) {
			/* LDT */
			if (!CPU_LDTR_DESC.valid) {
				EXCEPTION(GP_EXCEPTION, selector);
				/*NOTREACHED*/
				return;
			}
			dt_base = CPU_LDTR_BASE;
			dt_limit = CPU_LDTR_LIMIT;
		} else {
			/* GDT */
			dt_base = CPU_GDTR_BASE;
			dt_limit = CPU_GDTR_LIMIT;
		}
		if (sd_idx >= dt_limit) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}
		CPU_SET_SEGDESC(sd, dt_base + sd_idx);

		if (!sd.valid) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		if (sd.s) {
			if (!(sd.dc && !(sd.type & CPU_SEGDESC_TYPE_EC)) &&
			    ((sd.dpl < CPU_INST_CPL) || (sd.dpl < sd_rpl))) {
				CPU_FLAGL &= ~Z_FLAG;
				return;
			}
		} else {
			switch (sd.type) {
			case 1: case 2: case 3: case 4: case 5:
			case 9: case 11: case 12:
				break;

			default:
				CPU_FLAGL &= ~Z_FLAG;
				return;
			}
		}

		*out = sd.h & 0x00ffff00;	/* 0x00fxff00, x? */
		CPU_FLAGL |= Z_FLAG;
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
LSL_GwEw(void)
{
	WORD *out;
	DWORD op;
	WORD selector;

	if (CPU_INST_PM && !CPU_INST_VM86) {
		SEGDESC sd;		/* segment register cache */
		DWORD sd_idx;		/* segment descriptor index */
		DWORD dt_base;		/* descriptor table base addr */
		WORD dt_limit;		/* descriptor table size limit */
		int sd_rpl;		/* segment descriptor RPL */

		PREPART_REG16_EA(op, selector, out, 5, 11);

		sd_idx = selector & ~7;
		sd_rpl = selector & 3;

		/* null segment */
		if (selector < 4) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		/* descriptor table */
		if (selector & CPU_SEGMENT_TABLE_IND) {
			/* LDT */
			if (!CPU_LDTR_DESC.valid) {
				EXCEPTION(GP_EXCEPTION, selector);
				/*NOTREACHED*/
				return;
			}
			dt_base = CPU_LDTR_BASE;
			dt_limit = CPU_LDTR_LIMIT;
		} else {
			/* GDT */
			dt_base = CPU_GDTR_BASE;
			dt_limit = CPU_GDTR_LIMIT;
		}
		if (sd_idx >= dt_limit) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}
		CPU_SET_SEGDESC(sd, dt_base + sd_idx);

		if (!sd.valid) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		if (sd.s) {
			if (!(sd.dc && !(sd.type & CPU_SEGDESC_TYPE_EC)) &&
			    ((sd.dpl < CPU_INST_CPL) || (sd.dpl < sd_rpl))) {
				CPU_FLAGL &= ~Z_FLAG;
				return;
			}
		} else {
			switch (sd.type) {
			case 1: case 2: case 3: case 9: case 11:
				break;

			default:
				CPU_FLAGL &= ~Z_FLAG;
				return;
			}
		}

		*out = (WORD)sd.limit;
		CPU_FLAGL |= Z_FLAG;
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
LSL_GdEw(void)
{
	DWORD *out;
	DWORD op;
	WORD selector;

	if (CPU_INST_PM && !CPU_INST_VM86) {
		SEGDESC sd;		/* segment register cache */
		DWORD sd_idx;		/* segment descriptor index */
		DWORD dt_base;		/* descriptor table base addr */
		WORD dt_limit;		/* descriptor table size limit */
		int sd_rpl;		/* segment descriptor RPL */

		PREPART_REG32_EA(op, selector, out, 5, 11);

		sd_idx = selector & ~7;
		sd_rpl = selector & 3;

		/* null segment */
		if (selector < 4) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		/* descriptor table */
		if (selector & CPU_SEGMENT_TABLE_IND) {
			/* LDT */
			if (!CPU_LDTR_DESC.valid) {
				EXCEPTION(GP_EXCEPTION, selector);
				/*NOTREACHED*/
				return;
			}
			dt_base = CPU_LDTR_BASE;
			dt_limit = CPU_LDTR_LIMIT;
		} else {
			/* GDT */
			dt_base = CPU_GDTR_BASE;
			dt_limit = CPU_GDTR_LIMIT;
		}
		if (sd_idx >= dt_limit) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}
		CPU_SET_SEGDESC(sd, dt_base + sd_idx);

		if (!sd.valid) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		if (sd.s) {
			if (!(sd.dc && !(sd.type & CPU_SEGDESC_TYPE_EC)) &&
			    ((sd.dpl < CPU_INST_CPL) || (sd.dpl < sd_rpl))) {
				CPU_FLAGL &= ~Z_FLAG;
				return;
			}
		} else {
			switch (sd.type) {
			case 1: case 2: case 3: case 9: case 11:
				break;

			default:
				CPU_FLAGL &= ~Z_FLAG;
				return;
			}
		}

		*out = sd.limit;
		CPU_FLAGL |= Z_FLAG;
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
VERR_Ew(DWORD op)
{
	WORD *out;
	DWORD madr;
	WORD selector;

	if (CPU_INST_PM && !CPU_INST_VM86) {
		SEGDESC sd;		/* segment register cache */
		DWORD sd_idx;		/* segment descriptor index */
		DWORD dt_base;		/* descriptor table base addr */
		WORD dt_limit;		/* descriptor table size limit */
		int sd_rpl;		/* segment descriptor RPL */

		if (op >= 0xc0) {
			CPU_CLOCK(5);
			selector = *(reg16_b20[op]);
		} else {
			CPU_CLOCK(11);
			madr = calc_ea_dst(op);
			selector = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		}
		out = reg16_b53[op];

		sd_idx = selector & ~7;
		sd_rpl = selector & 3;

		/* null segment */
		if (selector < 4) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		/* descriptor table */
		if (selector & CPU_SEGMENT_TABLE_IND) {
			/* LDT */
			if (!CPU_LDTR_DESC.valid) {
				EXCEPTION(GP_EXCEPTION, selector);
				/*NOTREACHED*/
				return;
			}
			dt_base = CPU_LDTR_BASE;
			dt_limit = CPU_LDTR_LIMIT;
		} else {
			/* GDT */
			dt_base = CPU_GDTR_BASE;
			dt_limit = CPU_GDTR_LIMIT;
		}
		if (sd_idx >= dt_limit) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}
		CPU_SET_SEGDESC(sd, dt_base + sd_idx);

		if (!sd.valid || !sd.s) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}
		if (!(sd.dc && !(sd.type & CPU_SEGDESC_TYPE_EC))) {
			if ((sd.dpl < CPU_INST_CPL) || (sd.dpl < sd_rpl)) {
				CPU_FLAGL &= ~Z_FLAG;
				return;
			}
		}
		if (sd.dc && !sd.wr) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		CPU_FLAGL |= Z_FLAG;
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
VERW_Ew(DWORD op)
{
	WORD *out;
	DWORD madr;
	WORD selector;

	if (CPU_INST_PM && !CPU_INST_VM86) {
		SEGDESC sd;		/* segment register cache */
		DWORD sd_idx;		/* segment descriptor index */
		DWORD dt_base;		/* descriptor table base addr */
		WORD dt_limit;		/* descriptor table size limit */
		int sd_rpl;		/* segment descriptor RPL */

		if (op >= 0xc0) {
			CPU_CLOCK(5);
			selector = *(reg16_b20[op]);
		} else {
			CPU_CLOCK(11);
			madr = calc_ea_dst(op);
			selector = cpu_vmemoryread_w(CPU_INST_SEGREG_INDEX, madr);
		}
		out = reg16_b53[op];

		sd_idx = selector & ~7;
		sd_rpl = selector & 3;

		/* null segment */
		if (selector < 4) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		/* descriptor table */
		if (selector & CPU_SEGMENT_TABLE_IND) {
			/* LDT */
			if (!CPU_LDTR_DESC.valid) {
				EXCEPTION(GP_EXCEPTION, selector);
				/*NOTREACHED*/
				return;
			}
			dt_base = CPU_LDTR_BASE;
			dt_limit = CPU_LDTR_LIMIT;
		} else {
			/* GDT */
			dt_base = CPU_GDTR_BASE;
			dt_limit = CPU_GDTR_LIMIT;
		}
		if (sd_idx >= dt_limit) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}
		CPU_SET_SEGDESC(sd, dt_base + sd_idx);

		if (!sd.valid || !sd.s || sd.dc ||
		    !(sd.type & CPU_SEGDESC_TYPE_WR) ||
		    (sd.dpl < CPU_INST_CPL) || (sd.dpl < sd_rpl)) {
			CPU_FLAGL &= ~Z_FLAG;
			return;
		}

		CPU_FLAGL |= Z_FLAG;
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
MOV_DdRd(void)
{

	ia32_panic("MOV_DdRd: not implemented yet!");
}

void
MOV_RdDd(void)
{

	ia32_panic("MOV_DdRd: not implemented yet!");
}

void
INVD(void)
{

	CPU_CLOCK(11);
	if (CPU_INST_PM && (CPU_INST_VM86 || CPU_INST_CPL != 0)) {
		EXCEPTION(GP_EXCEPTION, 0);
		/*NOTREACHED*/
		return;
	}
}

void
WBINVD(void)
{

	CPU_CLOCK(11);
	if (CPU_INST_PM && (CPU_INST_VM86 || CPU_INST_CPL != 0)) {
		EXCEPTION(GP_EXCEPTION, 0);
		/*NOTREACHED*/
		return;
	}
}

void
INVLPG(DWORD op)
{
	DWORD madr;

	if (CPU_INST_PM && (CPU_INST_VM86 || CPU_INST_CPL != 0)) {
		EXCEPTION(GP_EXCEPTION, 0);
		/*NOTREACHED*/
		return;
	}

	if (op < 0xc0) {
		CPU_CLOCK(11);
		madr = get_ea(op);
		/* Nothing to do */
		return;
	}
	EXCEPTION(UD_EXCEPTION, 0);
}

void
_LOCK(void)
{

	/* Nothing to do */
}

void
HLT(void)
{

	if (CPU_INST_PM && CPU_INST_CPL != 0) {
		EXCEPTION(GP_EXCEPTION, 0);
		/*NOTREACHED*/
		return;
	}

	CPU_HALT();
	if (!CPU_INST_OP32) {
		CPU_IP--;
	} else {
		CPU_EIP--;
	}
}

void
RSM(void)
{

	ia32_panic("RSM: not implemented yet!");
}

void
RDMSR(void)
{

#if 1
	EXCEPTION(UD_EXCEPTION, 0);
#else
	if (CPU_INST_PM && (CPU_INST_VM86 || CPU_INST_CPL != 0)) {
		EXCEPTION(GP_EXCEPTION, 0);
		/*NOTREACHED*/
		return;
	}
#endif
}

void
WRMSR(void)
{

#if 1
	EXCEPTION(UD_EXCEPTION, 0);
#else
	if (CPU_INST_PM && (CPU_INST_VM86 || CPU_INST_CPL != 0)) {
		EXCEPTION(GP_EXCEPTION, 0);
		/*NOTREACHED*/
		return;
	}
#endif
}

void
RDTSC(void)
{

	ia32_panic("RDTSC: not implemented yet!");
}

void
MOV_TdRd(void)
{

	ia32_panic("MOV_TdRd: not implemented yet!");
}

void
MOV_RdTd(void)
{

	ia32_panic("MOV_RdTd: not implemented yet!");
}
