/*	$Id: segments.h,v 1.1 2003/07/22 15:09:44 nonaka Exp $	*/

/*
 * Copyright (c) 2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef	IA32_CPU_SEGMENTS_H__
#define	IA32_CPU_SEGMENTS_H__

#ifdef __cplusplus
extern "C" {
#endif

#define	CPU_SEGMENT_TABLE_IND		(1 << 2)	/* 0 = GDT, 1 = LDT */

typedef struct {
	DWORD	segbase;
	DWORD	limit;

	DWORD	sdaddr;	/* segment descriptor addr */
	DWORD	h;	/* high dword cache */
	DWORD	l;	/* low dword cache */

	int	valid;	/* segment descriptor valid flag */

	BYTE	p;	/* avail */

	BYTE	type;
	BYTE	dpl;

	BYTE	s;	/* 0 = system, 1 = code/data */
	BYTE	dc;	/* 0 = data, 1 = code */
	BYTE	g;	/* 4k base */

	BYTE	bd;	/* 32bit */
	BYTE	wr;	/* readable/writable */
} SEGDESC;

#define	CPU_SEGDESC_H_TYPE	(0xf <<  8)
#define	CPU_SEGDESC_H_A		(  1 <<  8)
#define	CPU_SEGDESC_H_D_C	(  1 << 11)	/* 0 = data, 1 = code */
#define	CPU_SEGDESC_H_S		(  1 << 12)	/* 0 = system, 1 = code/data */
#define	CPU_SEGDESC_H_DPL	(  3 << 13)
#define	CPU_SEGDESC_H_P		(  1 << 15)	/* exist */
#define	CPU_SEGDESC_H_B_D	(  1 << 22)
#define	CPU_SEGDESC_H_G		(  1 << 23)

#define	CPU_SEGDESC_TYPE_A	(  1 << 0)
#define	CPU_SEGDESC_TYPE_WR	(  1 << 1)
#define	CPU_SEGDESC_TYPE_EC	(  1 << 2)

#define	CPU_SET_SEGDESC(d, addr) \
do { \
	(d).sdaddr = (addr); \
	(d).l = cpu_memoryread_d((d).sdaddr); \
	(d).h = cpu_memoryread_d((d).sdaddr + 4); \
	(d).valid = 1; \
	(d).s = ((d).h & CPU_SEGDESC_H_S) == CPU_SEGDESC_H_S; \
	(d).type = ((d).h & CPU_SEGDESC_H_TYPE) >> 8; \
	(d).dc = ((d).h & CPU_SEGDESC_H_D_C) == CPU_SEGDESC_H_D_C; \
	(d).dpl = ((d).h & CPU_SEGDESC_H_DPL) >> 13; \
	(d).p = ((d).h & CPU_SEGDESC_H_P) == CPU_SEGDESC_H_P; \
	(d).bd = ((d).h & CPU_SEGDESC_H_B_D) == CPU_SEGDESC_H_B_D; \
	(d).g = ((d).h & CPU_SEGDESC_H_G) == CPU_SEGDESC_H_G; \
	(d).segbase = ((d).h & 0xff000000) | (((d).h & 0xff) << 16) | \
	                                                       ((d).l >> 16); \
	(d).limit = ((d).h & 0xf0000) | ((d).l & 0xffff); \
	if ((d).g) { \
		(d).limit <<= 12; \
		(d).limit |= 0xfff; \
	} \
	if (!(d).s) { \
		(d).valid = 0; \
		switch ((d).type) { \
		case 1: /* 286 TSS */ \
		case 3: /* 286 TSS Busy */ \
			(d).valid = 1; \
			break; \
		case 2: /* LDT descriptor */ \
			(d).valid = 1; \
			break; \
		case 4: /* 286 call gate */ \
		case 6: /* 286 interrupt gate */ \
		case 7: /* 286 trap gate */ \
			(d).valid = 1; \
			break; \
		case 9: /* 386 TSS */ \
		case 11: /* 386 TSS Busy */ \
			(d).valid = 1; \
			break; \
		case 12: /* 386 call gate */ \
		case 14: /* 386 interrupt gate */ \
		case 15: /* 386 trap gate */ \
			(d).valid = 1; \
			break; \
		case 0: case 8: case 10: case 13: /* reserved */ \
		default: \
			ia32_panic("bad segment descriptor (%d)", (d).type); \
			break; \
		} \
	} \
} while (/*CONSTCOND*/ 0)

#define	CPU_SET_SEGDESC_POSTPART(d) \
do { \
	if ((d).s) { \
		if (!((d).h & CPU_SEGDESC_H_A)) { \
			(d).h |= CPU_SEGDESC_H_A; \
			cpu_memorywrite_d((d).sdaddr + 4, (d).h); \
		} \
	} \
} while (/*CONSTCOND*/ 0)

#define	CPU_SET_SEGREG(idx, selector)	load_segreg(idx, selector)
void load_segreg(int idx, WORD selector);
void load_ldtr(WORD selector);

#ifdef __cplusplus
}
#endif

#endif	/* !IA32_CPU_SEGMENTS_H__ */
