/*	$Id: task.c,v 1.1 2003/07/22 15:09:44 nonaka Exp $	*/

/*
 * Copyright (c) 2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "cpu.h"
#include "ia32.mcr"


void
load_tr(WORD selector)
{
	SEGDESC sd;		/* segment descriptor cache */
	DWORD sd_idx;		/* segment descriptor index */
	int sd_rpl;		/* segment descriptor RPL */

	sd_idx = selector & ~7;
	sd_rpl = selector & 3;

	/* null segment */
	if (selector < 4) {
		ia32_panic("load_tr: null selector (%x)", selector);
		/*NOTREACHED*/
		return;
	}

	/* descriptor table */
	if (sd_idx >= CPU_GDTR_LIMIT) {
		EXCEPTION(GP_EXCEPTION, selector);
		/*NOTREACHED*/
		return;
	}
	/* gdt */
	if (selector & CPU_SEGMENT_TABLE_IND) {
		EXCEPTION(GP_EXCEPTION, selector);
		/*NOTREACHED*/
		return;
	}
	CPU_SET_SEGDESC(sd, CPU_GDTR_BASE + sd_idx);

	if (!sd.valid || sd.s || ((sd.type != 1) && (sd.type != 9))) {
		EXCEPTION(GP_EXCEPTION, selector);
		/*NOTREACHED*/
		return;
	}
	if (sd.type == 1 && sd.limit < 0x2b) {
		EXCEPTION(TS_EXCEPTION, selector);	/* panic? XXX */
		/*NOTREACHED*/
		return;
	}
	else if (sd.type == 9 && sd.limit < 0x67) {
		EXCEPTION(TS_EXCEPTION, selector);	/* panic? XXX */
		/*NOTREACHED*/
		return;
	}
	if (!sd.p) {
		EXCEPTION(NP_EXCEPTION, selector);
		/*NOTREACHED*/
		return;
	}
	CPU_SET_TASK_BUSY(sd);
	CPU_LDTR = selector;
	CPU_LDTR_DESC = sd;
}
