#include	"compiler.h"

#include	"dosio.h"
#include	"memory.h"
#include	"font.h"
#include	"milstr.h"

#ifndef FONTMEMORYBIND
#ifdef __cplusplus
extern "C" {
#endif
	BYTE	__font[0x84000];
#ifdef __cplusplus
}
#endif
#endif

#ifndef	SYSFONTPATH
#define	SYSFONTPATH	"/path/to/your/sys/font/path"
#endif

	char	pc88ankname[]	= "pc88.fnt";
	char	pc88knj1name[]	= "kanji1.rom";
	char	pc88knj2name[]	= "kanji2.rom";
	char	pc98fontname[]	= "font.bmp";
	char	v98fontname[]	= "font.rom";
	char	fm7ankname[]	= "subsys_c.rom";
	char	fm7knjname[]	= "kanji.rom";
	char	x1ank1name[]	= "fnt0808.x1";
	char	x1ank2name[]	= "fnt0816.x1";
	char	x1knjname[]	= "fnt1616.x1";
	char	x68kfontname[]	= "cgrom.dat";
	char	fonttmpname[]	= "fonttmp.bmp";


void font_init(void) {

	BYTE	*p, *q;
	int		i, j;
	DWORD	dbit;

	ZeroMemory(font, sizeof(font));

	p = font + 0x81000;
	q = font + 0x82800;
	for (i=0; i<256; i++) {
		for (j=0; j<4; j++) {
			dbit = 0;
			if (i & (0x01 << j)) {
				dbit |= 0xf0f0f0f0;
			}
			if (i & (0x10 << j)) {
				dbit |= 0x0f0f0f0f;
			}
			*(DWORD *)p = dbit;
			p += 4;
			*(WORD *)q = (WORD)dbit;
			q += 2;
		}
	}
}

static BYTE fonttypecheck(char *fname) {

	char *p;

	p = getExtName(fname);
	if (!milstr_cmpA(p, "BMP")) {
		return(1);
	}
	p = getFileName(fname);
	if (!milstr_cmpA(p, v98fontname)) {
		return(2);
	}
	if ((!milstr_cmpA(p, pc88ankname)) ||
		(!milstr_cmpA(p, pc88knj1name)) ||
		(!milstr_cmpA(p, pc88knj2name))) {
		return(3);
	}
	if ((!milstr_cmpA(p, fm7ankname)) ||
		(!milstr_cmpA(p, fm7knjname))) {
		return(4);
	}
	if ((!milstr_cmpA(p, x1ank1name)) ||
		(!milstr_cmpA(p, x1ank2name)) ||
		(!milstr_cmpA(p, x1knjname))) {
		return(5);
	}
	if (!milstr_cmpA(p, x68kfontname)) {
		return(6);
	}
	return(0);
}

BYTE font_load(char *filename, BOOL force) {

	BYTE	*p, *q;
	int		i, j;
	char	fname[MAX_PATH + 16];
	BYTE	type;
	BYTE	loading;

	if (filename) {
		strcpy(fname, filename);
	}
	else {
		fname[0] = '\0';
	}
	type = fonttypecheck(fname);
	if ((!type) && (!force)) {
		return(0);
	}

	// : font[??560-??57f], font[??d60-??d7f] Ϻʤ褦ˡ
	for (i=0; i<0x80; i++) {
		p = font + (i << 12);
		ZeroMemory(p+0x000, 0x0560-0x000);
		ZeroMemory(p+0x580, 0x0d60-0x580);
		ZeroMemory(p+0xd80, 0x1000-0xd80);
	}

	CopyMemory(font + 0x82000, deffnt8x8, 0x800);

	p = font + 0x80000;
	q = deffnt8x8;
	for (i=0; i<256; i++) {
		for (j=0; j<8; j++) {
			*p++ = *q;
			*p++ = *q++;
		}
	}

	loading = 0xff;
	switch(type) {
		case 1:
			loading = pc98fontread(fname, loading);
			break;
		case 2:
			loading = v98fontread(fname, loading);
			break;
		case 3:
			loading = pc88fontread(fname, loading);
			break;
		case 4:
			loading = fm7fontread(fname, loading);
			break;
		case 5:
			loading = x1fontread(fname, loading);
			break;
		case 6:
			loading = x68kfontread(fname, loading);
			break;
	}
	loading = pc98fontread(file_getcd(pc98fontname), loading);
	loading = v98fontread(file_getcd(v98fontname), loading);
	loading = pc88fontread(file_getcd(pc88ankname), loading);
	if (loading & (LOADFONT_16)) {
		loading = pc98fontread(SYSFONTPATH "/font.bmp", loading);
		if (loading & (LOADFONT_16)) {
			if (file_attr_c(fonttmpname) == -1) {
				np2os_makefontbmp(fonttmpname);
			}
			loading = pc98fontread(file_getcd(fonttmpname),loading);
		}
	}
	return(type);
}




void patchank16a(void) {

	CopyMemory(font + 0x80000, fontank16, 0x200);
}

void patchank16b(void) {

	CopyMemory(font + 0x80800, &fontank16[0x100], 0x200);
	CopyMemory(font + 0x80e00, &fontank16[0x200], 0x200);
}
