/*	$Id: gtk_sound.c,v 1.5 2003/06/05 18:11:23 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compiler.h"

#include "np2.h"
#include "pc_core.h"
#include "ini.h"
#include "audio.h"
#include "midi.h"
#include "xnp2.h"
#include "xmenu.h"

/*
 * PC-9801-86
 */

#define	SND86_SHIFT_IOPORT	0
#define	SND86_SHIFT_BIOSROM	1
#define	SND86_SHIFT_INTR	2
#define	SND86_SHIFT_INTERRUPT	4
#define	SND86_SHIFT_SOUNDID	5

#define	SND86_MASK_IOPORT	0x01
#define	SND86_MASK_BIOSROM	0x02
#define	SND86_MASK_INTR		0x0c
#define	SND86_MASK_INTERRUPT	0x10
#define	SND86_MASK_SOUNDID	0xe0

#define	SND86_GET_IOPORT() \
	((np2cfg.snd86opt & SND86_MASK_IOPORT) >> SND86_SHIFT_IOPORT)
#define	SND86_GET_BIOSROM() \
	((np2cfg.snd86opt & SND86_MASK_BIOSROM) >> SND86_SHIFT_BIOSROM)
#define	SND86_GET_INTR() \
	((np2cfg.snd86opt & SND86_MASK_INTR) >> SND86_SHIFT_INTR)
#define	SND86_GET_INTERRUPT() \
	((np2cfg.snd86opt & SND86_MASK_INTERRUPT) >> SND86_SHIFT_INTERRUPT)
#define	SND86_GET_SOUNDID() \
	((np2cfg.snd86opt & SND86_MASK_SOUNDID) >> SND86_SHIFT_SOUNDID)

#define	SND86_SET_IOPORT(v) \
	((np2cfg.snd86opt & ~SND86_MASK_IOPORT) \
	    | (((v) << SND86_SHIFT_IOPORT) & SND86_MASK_IOPORT))
#define	SND86_SET_BIOSROM(v) \
	((np2cfg.snd86opt & ~SND86_MASK_BIOSROM) \
	    | (((v) << SND86_SHIFT_BIOSROM) & SND86_MASK_BIOSROM))
#define	SND86_SET_INTR(v) \
	((np2cfg.snd86opt & ~SND86_MASK_INTR) \
	    | (((v) << SND86_SHIFT_INTR) & SND86_MASK_INTR))
#define	SND86_SET_INTERRUPT(v) \
	((np2cfg.snd86opt & ~SND86_MASK_INTERRUPT) \
	    | (((v) << SND86_SHIFT_INTERRUPT) & SND86_MASK_INTERRUPT))
#define	SND86_SET_SOUNDID(v) \
	((np2cfg.snd86opt & ~SND86_MASK_SOUNDID) \
	    | (((v) << SND86_SHIFT_SOUNDID) & SND86_MASK_SOUNDID))

enum {
	SND86_IOPORT_0288,
	SND86_IOPORT_0188
};

enum {
	SND86_INT_0,
	SND86_INT_4,
	SND86_INT_6,
	SND86_INT_5
};

enum {
	SND86_SOUNDID_7x,
	SND86_SOUNDID_6x,
	SND86_SOUNDID_5x,
	SND86_SOUNDID_4x,
	SND86_SOUNDID_3x,
	SND86_SOUNDID_2x,
	SND86_SOUNDID_1x,
	SND86_SOUNDID_0x
};

static const char *snd86_ioport_str[] = {
	"0288", "0188",
};

static const char *snd86_intr_str[] = {
	"INT0", "INT4", "INT6", "INT5",
};

static const char *snd86_soundid_str[] = {
	"7x", "6x", "5x", "4x", "3x", "2x", "1x", "0x",
};

static const char *driver_name[SNDDRV_DRVMAX] = {
	"None",
	"NetBSD",
	"OSS",
	"EsounD",
	"SDL",
};

static GtkObject *FM_adj;
static GtkObject *PSG_adj;
static GtkObject *ADPCM_adj;
static GtkObject *PCM_adj;
static GtkObject *Rhythm_adj;
static GtkWidget *ioport_entry;
static GtkWidget *int_entry;
static GtkWidget *soundid_entry;
static GtkWidget *int_checkbutton;
static GtkWidget *rom_checkbutton;
static GtkWidget *audio_device_entry;
static GtkWidget *midi_device_entry;
static GtkWidget *driver_radiobutton[SNDDRV_DRVMAX];

static int snddrv;

static void
ok_button_clicked(GtkButton *b, gpointer d)
{
	gchar *ioport = gtk_entry_get_text(GTK_ENTRY(ioport_entry));
	gchar *intr = gtk_entry_get_text(GTK_ENTRY(int_entry));
	gchar *soundid = gtk_entry_get_text(GTK_ENTRY(soundid_entry));
	gchar *audiodevp = gtk_entry_get_text(GTK_ENTRY(audio_device_entry));
	gchar *mididevp = gtk_entry_get_text(GTK_ENTRY(midi_device_entry));
	guint vol_fm = (guint)(GTK_ADJUSTMENT(FM_adj)->value);
	guint vol_ssg = (guint)(GTK_ADJUSTMENT(PSG_adj)->value);
	guint vol_adpcm = (guint)(GTK_ADJUSTMENT(ADPCM_adj)->value);
	guint vol_pcm = (guint)(GTK_ADJUSTMENT(PCM_adj)->value);
	guint vol_rhythm = (guint)(GTK_ADJUSTMENT(Rhythm_adj)->value);
	gint interrupt = GTK_TOGGLE_BUTTON(int_checkbutton)->active;
	gint biosrom = GTK_TOGGLE_BUTTON(rom_checkbutton)->active;
	int i;

	UNUSED(b);

	/* Mixer */
	if (np2cfg.vol_fm != vol_fm) {
		INIT_RENEWAL;
		np2cfg.vol_fm = vol_fm;
	}
	if (np2cfg.vol_ssg != vol_ssg) {
		INIT_RENEWAL;
		np2cfg.vol_ssg = vol_ssg;
	}
	if (np2cfg.vol_adpcm != vol_adpcm) {
		INIT_RENEWAL;
		np2cfg.vol_adpcm = vol_adpcm;
	}
	if (np2cfg.vol_pcm != vol_pcm) {
		INIT_RENEWAL;
		np2cfg.vol_pcm = vol_pcm;
	}
	if (np2cfg.vol_rhythm != vol_rhythm) {
		INIT_RENEWAL;
		np2cfg.vol_rhythm = vol_rhythm;
	}

	/* PC-9801-86 */
	for (i = 0; i < NELEMENTS(snd86_ioport_str); i++) {
		if (strcmp(ioport, snd86_ioport_str[i]) == 0) {
			if (SND86_GET_IOPORT() != i) {
				INIT_RENEWAL;
				np2cfg.snd86opt = SND86_SET_IOPORT(i);
			}
			break;
		}
	}

	for (i = 0; i < NELEMENTS(snd86_intr_str); i++) {
		if (strcmp(intr, snd86_intr_str[i]) == 0) {
			if (SND86_GET_INTR() != i) {
				INIT_RENEWAL;
				np2cfg.snd86opt = SND86_SET_INTR(i);
			}
			break;
		}
	}

	for (i = 0; i < NELEMENTS(snd86_soundid_str); i++) {
		if (strcmp(soundid, snd86_soundid_str[i]) == 0) {
			if (SND86_GET_SOUNDID() != i) {
				INIT_RENEWAL;
				np2cfg.snd86opt = SND86_SET_SOUNDID(i);
			}
			break;
		}
	}

	if (SND86_GET_INTERRUPT() != interrupt) {
		INIT_RENEWAL;
		np2cfg.snd86opt = SND86_SET_INTERRUPT(interrupt);
	}

	if (SND86_GET_BIOSROM() != biosrom) {
		INIT_RENEWAL;
		np2cfg.snd86opt = SND86_SET_BIOSROM(biosrom);
	}

	/* Driver */
	if (np2cfg.snddrv != snddrv) {
		INIT_RENEWAL;
		np2cfg.snddrv = snddrv;
	}

	if (strcmp(audiodevp, audiodev) != 0) {
		INIT_RENEWAL;
		g_snprintf(audiodev, sizeof(audiodev), "%s", audiodevp);
	}

	if (strcmp(mididevp, MIDIDEV) != 0) {
		INIT_RENEWAL;
		g_snprintf(MIDIDEV, sizeof(MIDIDEV), "%s", mididevp);
	}

	gtk_widget_destroy((GtkWidget *)d);
}

static void
sound_destroy(GtkWidget *w, GtkWidget **wp)
{

	UNUSED(wp);

	install_idle_process();
	gtk_widget_destroy(w);
}

static void
mixer_default_button_clicked(GtkButton *b, gpointer d)
{

	UNUSED(b);
	UNUSED(d);

	gtk_adjustment_set_value(GTK_ADJUSTMENT(FM_adj), 64.0);
	gtk_adjustment_set_value(GTK_ADJUSTMENT(PSG_adj), 64.0);
	gtk_adjustment_set_value(GTK_ADJUSTMENT(ADPCM_adj), 64.0);
	gtk_adjustment_set_value(GTK_ADJUSTMENT(PCM_adj), 64.0);
	gtk_adjustment_set_value(GTK_ADJUSTMENT(Rhythm_adj), 64.0);
}

static void
snd86_default_button_clicked(GtkButton *b, gpointer d)
{

	UNUSED(b);
	UNUSED(d);

	gtk_entry_set_text(GTK_ENTRY(ioport_entry), "0188");
	gtk_entry_set_text(GTK_ENTRY(int_entry), "INT5");
	gtk_entry_set_text(GTK_ENTRY(soundid_entry), "4x");
	if (!GTK_TOGGLE_BUTTON(int_checkbutton)->active)
		gtk_signal_emit_by_name(GTK_OBJECT(int_checkbutton), "clicked");
	if (!GTK_TOGGLE_BUTTON(rom_checkbutton)->active)
		gtk_signal_emit_by_name(GTK_OBJECT(rom_checkbutton), "clicked");
}

static void
driver_radiobutton_clicked(GtkButton *b, gpointer d)
{

	UNUSED(b);

	snddrv = (int)d;
}

void
create_sound_dialog(void)
{
	GtkWidget *sound_window;
	GtkWidget *vbox1;
	GtkWidget *hbox1;
	GtkWidget *notebook1;
	GtkWidget *vbox2;
	GtkWidget *hbox2;
	GtkWidget *FM_label;
	GtkWidget *FM_hscale;
	GtkWidget *hbox3;
	GtkWidget *PSG_label;
	GtkWidget *PSG_hscale;
	GtkWidget *hbox4;
	GtkWidget *ADMPCM_label;
	GtkWidget *ADPCM_hscale;
	GtkWidget *hbox5;
	GtkWidget *PCM_label;
	GtkWidget *PCM_hscale;
	GtkWidget *hbox6;
	GtkWidget *Rhythm_label;
	GtkWidget *Rhythm_hscale;
	GtkWidget *hbox9;
	GtkWidget *mixer_label;
	GtkWidget *vbox3;
	GtkWidget *hbox7;
	GtkWidget *ioport_label;
	GtkWidget *ioport_combo;
	GList *ioport_combo_items = NULL;
	GtkWidget *int_combo;
	GList *int_combo_items = NULL;
	GtkWidget *hbox8;
	GtkWidget *soundid_label;
	GtkWidget *soundid_combo;
	GList *soundid_combo_items = NULL;
	GtkWidget *hbox10;
	GtkWidget *PCM86_label;
	GtkWidget *driver_frame;
	GtkWidget *driver_label;
	GtkWidget *vbox4;
	GtkWidget *vbox5;
	GSList *vbox5_group = NULL;
	GtkWidget *audio_device_label;
	GtkWidget *midi_device_label;
	GtkWidget *hbox11;
	GtkWidget *hbox12;
	GtkWidget *cancel_button;
	GtkWidget *ok_button;
	GtkWidget *mixer_default_button;
	GtkWidget *snd86_default_button;
	int i;

	uninstall_idle_process();

	sound_window = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_object_set_data(GTK_OBJECT(sound_window), "sound_window", sound_window);
	gtk_widget_set_usize(sound_window, 320, 240);
	gtk_window_set_title(GTK_WINDOW(sound_window), "Sound Option");
	gtk_window_set_position(GTK_WINDOW(sound_window), GTK_WIN_POS_CENTER);
	gtk_window_set_modal(GTK_WINDOW(sound_window), TRUE);
	gtk_window_set_default_size(GTK_WINDOW(sound_window), 320, 240);
	gtk_window_set_policy(GTK_WINDOW(sound_window), FALSE, FALSE, FALSE);

	gtk_signal_connect(GTK_OBJECT(sound_window), "destroy",
	    GTK_SIGNAL_FUNC(sound_destroy), NULL);

	vbox1 = gtk_vbox_new(FALSE, 0);
	gtk_widget_ref(vbox1);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "vbox1", vbox1,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(vbox1);
	gtk_container_add(GTK_CONTAINER(sound_window), vbox1);

	notebook1 = gtk_notebook_new();
	gtk_widget_ref(notebook1);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "notebook1",
	    notebook1, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(notebook1);
	gtk_box_pack_start(GTK_BOX(vbox1), notebook1, TRUE, TRUE, 0);

	vbox2 = gtk_vbox_new(FALSE, 0);
	gtk_widget_ref(vbox2);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "vbox2", vbox2,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(vbox2);
	gtk_container_add(GTK_CONTAINER(notebook1), vbox2);

	hbox2 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox2);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox2", hbox2,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox2);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox2, TRUE, TRUE, 0);

	FM_label = gtk_label_new("FM");
	gtk_widget_ref(FM_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "FM_label", FM_label,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(FM_label);
	gtk_box_pack_start(GTK_BOX(hbox2), FM_label, FALSE, FALSE, 0);
	gtk_widget_set_usize(FM_label, 50, 0);
	gtk_label_set_justify(GTK_LABEL(FM_label), GTK_JUSTIFY_LEFT);

	FM_adj = gtk_adjustment_new(np2cfg.vol_fm, 0.0, 128.0, 1.0, 1.0, 0.0);
	FM_hscale = gtk_hscale_new(GTK_ADJUSTMENT(FM_adj));
	gtk_widget_ref(FM_hscale);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "FM_hscale",
	    FM_hscale, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(FM_hscale);
	gtk_box_pack_start(GTK_BOX(hbox2), FM_hscale, TRUE, TRUE, 0);
	gtk_scale_set_value_pos(GTK_SCALE(FM_hscale), GTK_POS_RIGHT);
	gtk_scale_set_digits(GTK_SCALE(FM_hscale), 0);

	hbox3 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox3);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox3", hbox3,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox3);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox3, TRUE, TRUE, 0);

	PSG_label = gtk_label_new("PSG");
	gtk_widget_ref(PSG_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "PSG_label",
	    PSG_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(PSG_label);
	gtk_box_pack_start(GTK_BOX(hbox3), PSG_label, FALSE, FALSE, 0);
	gtk_widget_set_usize(PSG_label, 50, 0);
	gtk_label_set_justify(GTK_LABEL(PSG_label), GTK_JUSTIFY_LEFT);

	PSG_adj = gtk_adjustment_new(np2cfg.vol_ssg, 0.0, 128.0, 1.0, 1.0, 0.0);
	PSG_hscale = gtk_hscale_new(GTK_ADJUSTMENT(PSG_adj));
	gtk_widget_ref(PSG_hscale);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "PSG_hscale",
	    PSG_hscale, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(PSG_hscale);
	gtk_box_pack_start(GTK_BOX(hbox3), PSG_hscale, TRUE, TRUE, 0);
	gtk_scale_set_value_pos(GTK_SCALE(PSG_hscale), GTK_POS_RIGHT);
	gtk_scale_set_digits(GTK_SCALE(PSG_hscale), 0);

	hbox4 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox4);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox4", hbox4,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox4);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox4, TRUE, TRUE, 0);

	ADMPCM_label = gtk_label_new("ADPCM");
	gtk_widget_ref(ADMPCM_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "ADMPCM_label",
	    ADMPCM_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(ADMPCM_label);
	gtk_box_pack_start(GTK_BOX(hbox4), ADMPCM_label, FALSE, FALSE, 0);
	gtk_widget_set_usize(ADMPCM_label, 50, 0);
	gtk_label_set_justify(GTK_LABEL(ADMPCM_label), GTK_JUSTIFY_LEFT);

	ADPCM_adj = gtk_adjustment_new(np2cfg.vol_adpcm,0.0,128.0,1.0,0.0,0.0);
	ADPCM_hscale = gtk_hscale_new(GTK_ADJUSTMENT(ADPCM_adj));
	gtk_widget_ref(ADPCM_hscale);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "ADPCM_hscale",
	    ADPCM_hscale, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(ADPCM_hscale);
	gtk_box_pack_start(GTK_BOX(hbox4), ADPCM_hscale, TRUE, TRUE, 0);
	gtk_scale_set_value_pos(GTK_SCALE(ADPCM_hscale), GTK_POS_RIGHT);
	gtk_scale_set_digits(GTK_SCALE(ADPCM_hscale), 0);

	hbox5 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox5);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox5", hbox5,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox5);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox5, TRUE, TRUE, 0);

	PCM_label = gtk_label_new("PCM");
	gtk_widget_ref(PCM_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "PCM_label",
	    PCM_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(PCM_label);
	gtk_box_pack_start(GTK_BOX(hbox5), PCM_label, FALSE, FALSE, 0);
	gtk_widget_set_usize(PCM_label, 50, 0);
	gtk_label_set_justify(GTK_LABEL(PCM_label), GTK_JUSTIFY_LEFT);

	PCM_adj = gtk_adjustment_new(np2cfg.vol_pcm,0.0,128.0,1.0,0.0,0.0);
	PCM_hscale = gtk_hscale_new(GTK_ADJUSTMENT(PCM_adj));
	gtk_widget_ref(PCM_hscale);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "PCM_hscale",
	    PCM_hscale, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(PCM_hscale);
	gtk_box_pack_start(GTK_BOX(hbox5), PCM_hscale, TRUE, TRUE, 0);
	gtk_scale_set_value_pos(GTK_SCALE(PCM_hscale), GTK_POS_RIGHT);
	gtk_scale_set_digits(GTK_SCALE(PCM_hscale), 0);

	hbox6 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox6);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox6", hbox6,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox6);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox6, TRUE, TRUE, 0);

	Rhythm_label = gtk_label_new("Rhythm");
	gtk_widget_ref(Rhythm_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "Rhythm_label",
	    Rhythm_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(Rhythm_label);
	gtk_box_pack_start(GTK_BOX(hbox6), Rhythm_label, FALSE, FALSE, 0);
	gtk_widget_set_usize(Rhythm_label, 50, 0);
	gtk_label_set_justify(GTK_LABEL(Rhythm_label), GTK_JUSTIFY_LEFT);

	Rhythm_adj =gtk_adjustment_new(np2cfg.vol_rhythm,0.0,128.0,1.0,0.0,0.0);
	Rhythm_hscale = gtk_hscale_new(GTK_ADJUSTMENT(Rhythm_adj));
	gtk_widget_ref(Rhythm_hscale);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "Rhythm_hscale",
	    Rhythm_hscale, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(Rhythm_hscale);
	gtk_box_pack_start(GTK_BOX(hbox6), Rhythm_hscale, TRUE, TRUE, 0);
	gtk_scale_set_value_pos(GTK_SCALE(Rhythm_hscale), GTK_POS_RIGHT);
	gtk_scale_set_digits(GTK_SCALE(Rhythm_hscale), 0);

	hbox9 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox9);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox9", hbox9,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox9);
	gtk_box_pack_start(GTK_BOX(vbox2), hbox9, TRUE, FALSE, 0);

	mixer_default_button = gtk_button_new_with_label("Default");
	gtk_widget_ref(mixer_default_button);
	gtk_object_set_data_full(GTK_OBJECT(sound_window),
	     "mixer_default_button", mixer_default_button,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(mixer_default_button);
	gtk_box_pack_end(GTK_BOX(hbox9), mixer_default_button, FALSE, FALSE, 5);
	gtk_signal_connect_object(GTK_OBJECT(mixer_default_button), "clicked",
	    GTK_SIGNAL_FUNC(mixer_default_button_clicked), NULL);

	mixer_label = gtk_label_new("Mixer");
	gtk_widget_ref(mixer_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "mixer_label",
	    mixer_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(mixer_label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook1),
	    gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook1), 0), mixer_label);

	vbox3 = gtk_vbox_new(FALSE, 0);
	gtk_widget_ref(vbox3);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "vbox3", vbox3,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(vbox3);
	gtk_container_add(GTK_CONTAINER(notebook1), vbox3);

	hbox7 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox7);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox7", hbox7,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox7);
	gtk_box_pack_start(GTK_BOX(vbox3), hbox7, FALSE, FALSE, 0);

	ioport_label = gtk_label_new("I/O port");
	gtk_widget_ref(ioport_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "ioport_label",
	    ioport_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(ioport_label);
	gtk_box_pack_start(GTK_BOX(hbox7), ioport_label, FALSE, FALSE, 0);
	gtk_widget_set_usize(ioport_label, 60, 0);
	gtk_label_set_justify(GTK_LABEL(ioport_label), GTK_JUSTIFY_LEFT);

	ioport_combo = gtk_combo_new();
	gtk_widget_ref(ioport_combo);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "ioport_combo",
	    ioport_combo, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(ioport_combo);
	gtk_box_pack_start(GTK_BOX(hbox7), ioport_combo, FALSE, FALSE, 0);
	gtk_widget_set_usize(ioport_combo, 90, 0);
	gtk_combo_set_value_in_list(GTK_COMBO(ioport_combo), TRUE, FALSE);
	for (i = NELEMENTS(snd86_ioport_str) - 1; i >= 0; i--) {
		ioport_combo_items = g_list_append(ioport_combo_items,
		    (gpointer)snd86_ioport_str[i]);
	}
	gtk_combo_set_popdown_strings(GTK_COMBO(ioport_combo), ioport_combo_items);
	g_list_free(ioport_combo_items);

	ioport_entry = GTK_COMBO(ioport_combo)->entry;
	gtk_widget_ref(ioport_entry);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "ioport_entry",
	    ioport_entry, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(ioport_entry);
	gtk_entry_set_editable(GTK_ENTRY(ioport_entry), FALSE);
	gtk_entry_set_text(GTK_ENTRY(ioport_entry), snd86_ioport_str[SND86_GET_IOPORT()]);

	int_checkbutton = gtk_check_button_new_with_label("Interrupt");
	gtk_widget_ref(int_checkbutton);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "int_checkbutton",
	    int_checkbutton, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(int_checkbutton);
	gtk_box_pack_start(GTK_BOX(hbox7), int_checkbutton, FALSE, FALSE, 2);
	if (SND86_GET_INTERRUPT())
		gtk_signal_emit_by_name(GTK_OBJECT(int_checkbutton), "clicked");

	int_combo = gtk_combo_new();
	gtk_widget_ref(int_combo);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "int_combo",
	    int_combo, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(int_combo);
	gtk_box_pack_start(GTK_BOX(hbox7), int_combo, FALSE, FALSE, 0);
	gtk_widget_set_usize(int_combo, 80, 0);
	gtk_combo_set_value_in_list(GTK_COMBO(int_combo), TRUE, FALSE);
	int_combo_items = g_list_append(int_combo_items,(gpointer)"INT0");
	int_combo_items = g_list_append(int_combo_items,(gpointer)"INT4");
	int_combo_items = g_list_append(int_combo_items,(gpointer)"INT5");
	int_combo_items = g_list_append(int_combo_items,(gpointer)"INT6");
	gtk_combo_set_popdown_strings(GTK_COMBO(int_combo), int_combo_items);
	g_list_free(int_combo_items);

	int_entry = GTK_COMBO(int_combo)->entry;
	gtk_widget_ref(int_entry);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "int_entry",
	    int_entry, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(int_entry);
	gtk_entry_set_editable(GTK_ENTRY(int_entry), FALSE);
	gtk_entry_set_text(GTK_ENTRY(int_entry),snd86_intr_str[SND86_GET_INTR()]);

	hbox8 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox8);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox8", hbox8,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox8);
	gtk_box_pack_start(GTK_BOX(vbox3), hbox8, FALSE, FALSE, 0);

	soundid_label = gtk_label_new("Sound ID");
	gtk_widget_ref(soundid_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "soundid_label",
	    soundid_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(soundid_label);
	gtk_box_pack_start(GTK_BOX(hbox8), soundid_label, FALSE, FALSE, 0);
	gtk_widget_set_usize(soundid_label, 60, 0);
	gtk_label_set_justify(GTK_LABEL(soundid_label), GTK_JUSTIFY_LEFT);

	soundid_combo = gtk_combo_new();
	gtk_widget_ref(soundid_combo);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "soundid_combo",
	    soundid_combo, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(soundid_combo);
	gtk_box_pack_start(GTK_BOX(hbox8), soundid_combo, FALSE, FALSE, 0);
	gtk_widget_set_usize(soundid_combo, 90, 0);
	gtk_combo_set_value_in_list(GTK_COMBO(soundid_combo), TRUE, FALSE);
	for (i = NELEMENTS(snd86_soundid_str) - 1; i >= 0; i--) {
		soundid_combo_items = g_list_append(soundid_combo_items,
		    (gpointer)snd86_soundid_str[i]);
	}
	gtk_combo_set_popdown_strings(GTK_COMBO(soundid_combo), soundid_combo_items);
	g_list_free(soundid_combo_items);

	soundid_entry = GTK_COMBO(soundid_combo)->entry;
	gtk_widget_ref(soundid_entry);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "soundid_entry",
	    soundid_entry, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(soundid_entry);
	gtk_entry_set_editable(GTK_ENTRY(soundid_entry), FALSE);
	gtk_entry_set_text(GTK_ENTRY(soundid_entry), snd86_soundid_str[SND86_GET_SOUNDID()]);

	rom_checkbutton = gtk_check_button_new_with_label("ROM");
	gtk_widget_ref(rom_checkbutton);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "rom_checkbutton",
	    rom_checkbutton, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(rom_checkbutton);
	gtk_box_pack_start(GTK_BOX(hbox8), rom_checkbutton, FALSE, FALSE, 2);
	if (SND86_GET_BIOSROM())
		gtk_signal_emit_by_name(GTK_OBJECT(rom_checkbutton), "clicked");

	hbox10 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox10);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox10", hbox10,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox10);
	gtk_box_pack_start(GTK_BOX(vbox3), hbox10, FALSE, FALSE, 0);

	snd86_default_button = gtk_button_new_with_label("Default");
	gtk_widget_ref(snd86_default_button);
	gtk_object_set_data_full(GTK_OBJECT(sound_window),
	    "snd86_default_button", snd86_default_button,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(snd86_default_button);
	gtk_box_pack_end(GTK_BOX(hbox10), snd86_default_button, FALSE, FALSE,5);
	gtk_signal_connect_object(GTK_OBJECT(snd86_default_button), "clicked",
	    GTK_SIGNAL_FUNC(snd86_default_button_clicked), NULL);

	PCM86_label = gtk_label_new("PC-9801-86");
	gtk_widget_ref(PCM86_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "PCM86_label",
	    PCM86_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(PCM86_label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook1),
	    gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook1), 1), PCM86_label);

	vbox4 = gtk_vbox_new(FALSE, 0);
	gtk_widget_ref(vbox4);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "vbox4", vbox4,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(vbox4);
	gtk_container_add(GTK_CONTAINER(notebook1), vbox4);

	driver_frame = gtk_frame_new("Sound driver");
	gtk_widget_ref(driver_frame);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "driver_frame",
	    driver_frame, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(driver_frame);
	gtk_container_set_border_width(GTK_CONTAINER(driver_frame), 5);
	gtk_box_pack_start(GTK_BOX(vbox4), driver_frame, TRUE, TRUE, 0);

	vbox5 = gtk_vbox_new(FALSE, 0);
	gtk_widget_ref(vbox5);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "vbox5", vbox5,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(vbox5);
	gtk_container_add(GTK_CONTAINER(driver_frame), vbox5);

	for (i = 0; i < SNDDRV_DRVMAX; i++) {
		driver_radiobutton[i] = gtk_radio_button_new_with_label(vbox5_group, driver_name[i]);
		vbox5_group = gtk_radio_button_group(GTK_RADIO_BUTTON(driver_radiobutton[i]));
		gtk_widget_ref(driver_radiobutton[i]);
		gtk_object_set_data_full(GTK_OBJECT(sound_window),
		    "driver_radiobutton[i]", driver_radiobutton[i],
		    (GtkDestroyNotify)gtk_widget_unref);
		gtk_widget_show(driver_radiobutton[i]);
		gtk_box_pack_start(GTK_BOX(vbox5), driver_radiobutton[i], TRUE, FALSE, 0);
		gtk_signal_connect(GTK_OBJECT(driver_radiobutton[i]), "clicked",
		    GTK_SIGNAL_FUNC(driver_radiobutton_clicked), (gpointer)i);
	}
	gtk_signal_emit_by_name(GTK_OBJECT(driver_radiobutton[np2cfg.snddrv]), "clicked");
#if !defined(USE_NETBSDAUDIO)
	gtk_widget_set_sensitive(driver_radiobutton[SNDDRV_NETBSD], FALSE);
#endif
#if !defined(USE_OSSAUDIO)
	gtk_widget_set_sensitive(driver_radiobutton[SNDDRV_OSS], FALSE);
#endif
#if !defined(USE_ESDAUDIO)
	gtk_widget_set_sensitive(driver_radiobutton[SNDDRV_ESD], FALSE);
#endif
#if !defined(USE_SDLAUDIO)
	gtk_widget_set_sensitive(driver_radiobutton[SNDDRV_SDL], FALSE);
#endif

	hbox11 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox11);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox11", hbox11,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox11);
	gtk_box_pack_start(GTK_BOX(vbox4), hbox11, FALSE, FALSE, 0);

	audio_device_label = gtk_label_new("Sound device");
	gtk_widget_ref(audio_device_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "audio_device_label",
	    audio_device_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(audio_device_label);
	gtk_widget_set_usize(audio_device_label, 120, 0);
	gtk_label_set_justify(GTK_LABEL(audio_device_label), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(hbox11), audio_device_label, FALSE, FALSE,0);

	audio_device_entry = gtk_entry_new();
	gtk_widget_ref(audio_device_entry);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "audio_device_entry",
	    audio_device_entry, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(audio_device_entry);
	gtk_widget_set_usize(audio_device_entry, 160, 0);
	gtk_entry_set_text(GTK_ENTRY(audio_device_entry), audiodev);
	gtk_box_pack_start(GTK_BOX(hbox11), audio_device_entry, FALSE, FALSE,0);

	hbox12 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox12);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox12", hbox12,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox12);
	gtk_box_pack_start(GTK_BOX(vbox4), hbox12, FALSE, FALSE, 0);

	midi_device_label = gtk_label_new("MIDI device");
	gtk_widget_ref(midi_device_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "midi_device_label",
	    midi_device_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(midi_device_label);
	gtk_widget_set_usize(midi_device_label, 120, 0);
	gtk_label_set_justify(GTK_LABEL(midi_device_label), GTK_JUSTIFY_LEFT);
	gtk_box_pack_start(GTK_BOX(hbox12), midi_device_label, FALSE, FALSE, 0);

	midi_device_entry = gtk_entry_new();
	gtk_widget_ref(midi_device_entry);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "midi_device_entry",
	    midi_device_entry, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(midi_device_entry);
	gtk_widget_set_usize(midi_device_entry, 160, 0);
	gtk_entry_set_text(GTK_ENTRY(midi_device_entry), MIDIDEV);
	gtk_box_pack_start(GTK_BOX(hbox12), midi_device_entry, FALSE, FALSE, 0);

	driver_label = gtk_label_new("Driver");
	gtk_widget_ref(driver_label);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "driver_label",
	    driver_label, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(driver_label);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook1),
	    gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook1), 2),driver_label);

	hbox1 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox1);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "hbox1", hbox1,
	    (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(hbox1);
	gtk_box_pack_start(GTK_BOX(vbox1), hbox1, FALSE, FALSE, 0);
	gtk_widget_set_usize(hbox1, 0, 36);

	cancel_button = gtk_button_new_with_label("Cancel");
	gtk_widget_ref(cancel_button);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "cancel_button",
	    cancel_button, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(cancel_button);
	gtk_box_pack_end(GTK_BOX(hbox1), cancel_button, FALSE, FALSE, 0);
	gtk_widget_set_usize(cancel_button, 96, 24);
	gtk_container_set_border_width(GTK_CONTAINER(cancel_button), 3);
	gtk_signal_connect_object(GTK_OBJECT(cancel_button), "clicked",
	    GTK_SIGNAL_FUNC(gtk_widget_destroy), GTK_OBJECT(sound_window));

	ok_button = gtk_button_new_with_label("OK");
	gtk_widget_ref(ok_button);
	gtk_object_set_data_full(GTK_OBJECT(sound_window), "ok_button",
	    ok_button, (GtkDestroyNotify)gtk_widget_unref);
	gtk_widget_show(ok_button);
	gtk_box_pack_end(GTK_BOX(hbox1), ok_button, FALSE, FALSE, 0);
	gtk_widget_set_usize(ok_button, 96, 24);
	gtk_container_set_border_width(GTK_CONTAINER(ok_button), 2);
	GTK_WIDGET_SET_FLAGS(ok_button, GTK_CAN_DEFAULT);
	GTK_WIDGET_SET_FLAGS(ok_button, GTK_HAS_DEFAULT);
	gtk_signal_connect(GTK_OBJECT(ok_button), "clicked",
	    GTK_SIGNAL_FUNC(ok_button_clicked), (gpointer)sound_window);
	gtk_widget_grab_default(ok_button);

	gtk_widget_show_all(sound_window);
}
