/*	$Id: gtkdraw.c,v 1.15 2003/06/05 18:11:23 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compiler.h"

#include "np2.h"
#include "np2pc98.h"
#include "iobridge.h"
#include "pc_core.h"

#include "gdc.h"
#include "memory.h"
#include "parts.h"
#include "palettes.h"
#include "xnp2.h"
#include "xdraw.h"

BYTE setscmd = 0;
BYTE SCREENMODE = 0;

int allflash = 0;
int displaymodel = 0;
int displaymoder = 0;
int palchanged = 0;
int nordisplay = 0;

static rect_t rectout = { 0, 0, SCREEN_WIDTH, 400 };
static rect_t rectddr = { 0, 0, SURFACE_WIDTH, SURFACE_HEIGHT };

NP2RECT realdisp = { 0, 0, 640, 400 };
NP2WIN_T vir = { 640, 400, 0, 640, 0, 0, 400, 0 };
XYSIZE minisize = { 640, 400 };
XYSIZE minisize2 = { 320, 200 };

static int winw = SCREEN_WIDTH;
static int winh = SCREEN_HEIGHT;

static GdkImage *surface;
static GdkPixmap *pixmap;
static unsigned long pixel[24];

void
xdraws_initwindowsize(WORD width, WORD height)
{

	winw = width + np2cfg.paddingx * 2;
	winh = height + np2cfg.paddingy * 2;
}

void
xdraws_clientmovesize(GtkWidget *w)
{
	WORD width, height;

	if ((setscmd & SCMD_SETTURN) == 0) {
		width = vir.sx;
		height = vir.sy;
	} else {
		width = vir.sy;
		height = vir.sx;
	}
	xdraws_initwindowsize(width, height);
	gtk_widget_set_usize(w, winw, winh);
	allflash = 1;
}

static void
palette_init(GtkWidget *w)
{
	GdkColor color;
	GdkColormap *cmap;
	int rv;
	int i;

	UNUSED(w);

	cmap = gdk_colormap_get_system();

	rv = gdk_colors_alloc(cmap, TRUE, NULL, 0, pixel, 24);
	if (rv == 0) {
		g_error("Can't allocate enough color.\n");
		return;
	}

	for (i = 0; i < 8; i++) {
		color.pixel = pixel[NP2PAL_TEXT + i + 1];
		color.red = np2_pal32[NP2PAL_TEXT + i + 1].p.r << 8;
		color.green = np2_pal32[NP2PAL_TEXT + i + 1].p.g << 8;
		color.blue = np2_pal32[NP2PAL_TEXT + i + 1].p.b << 8;
		gdk_colormap_alloc_color(cmap, &color, TRUE, FALSE);
	}
	palchanged = 0;
}

static void
palette_set(GtkWidget *w)
{
	GdkColor color;
	GdkColormap *cmap;
	int i;

	UNUSED(w);

	cmap = gdk_colormap_get_system();
	for (i = 0; i < NP2PALS_GRPH; i++) {
		color.pixel = pixel[NP2PAL_GRPH + i];
		color.red = np2_pal32[NP2PAL_GRPH + i].p.r << 8;
		color.green = np2_pal32[NP2PAL_GRPH + i].p.g << 8;
		color.blue = np2_pal32[NP2PAL_GRPH + i].p.b << 8;
		gdk_colormap_alloc_color(cmap, &color, TRUE, FALSE);
	}
}

int
xdraws_setup(GtkWidget *w)
{
	GdkVisual *visual;
	int nbit;

	visual = gtk_widget_get_visual(w);

	switch (visual->type) {
	case GDK_VISUAL_TRUE_COLOR:
		setscmd |= SCMD_NONPAL;
		break;
	case GDK_VISUAL_PSEUDO_COLOR:
		setscmd |= SCMD_USEPAL;
		break;
	case GDK_VISUAL_DIRECT_COLOR:
		if (visual->depth > 8) {
			setscmd |= SCMD_NONPAL;
			break;
		}
		/* FALLTHROUGH */
	default:
		fprintf(stderr, "No support visual class.\n");
		return 1;
	}

	if (visual->depth == 32) {
		SCREENMODE |= SCMD_COLOR32;
		nbit = 32;
	} else if (visual->depth == 24) {
		if (is_32bpp(window->window)) {
			SCREENMODE |= SCMD_COLOR32;
			nbit = 32;
		} else {
			SCREENMODE |= SCMD_COLOR24;
			nbit = 24;
		}
	} else if (visual->depth == 15 || visual->depth == 16) {
		SCREENMODE |= SCMD_COLOR16;
		nbit = 16;
		np2pal_make16mask(
		    visual->blue_mask, visual->red_mask, visual->green_mask);
	} else if (visual->depth == 8) {
		SCREENMODE |= SCMD_COLOR8;
		nbit = 8;
		palette_init(w);
	} else if (visual->depth < 8) {
		fprintf(stderr, "Too few allocable color.\n");
		return 1;
	} else {
		fprintf(stderr, "No support depth.\n");
		return 1;
	}
	np2pal_setcolorbit(nbit);

	surface = gdk_image_new(GDK_IMAGE_FASTEST, visual, rectddr.right, rectddr.bottom);
	if (surface == NULL) {
		g_message("can't create surface.");
		return 1;
	}

	xdraws_adjustleft();

	pixmap = gdk_pixmap_new(w->window, winw, winh, visual->depth);
	if (pixmap == NULL) {
		g_message("can't create pixmap.");
		return 1;
	}

	gdk_draw_rectangle(pixmap, w->style->black_gc, TRUE, 0, 0, winw, winh);

	return 0;
}

void
xdraws_term(GtkWidget *w)
{
	GdkVisual *visual;

	visual = gtk_widget_get_visual(w);
#if 0
	if (visual->class == PseudoColor)
		XFreeColors(display, cmap, pixel, 24, 0);
#endif
	if (pixmap)
		gdk_pixmap_unref(pixmap);
	if (surface)
		gdk_image_destroy(surface);
}

void
clearblanklines(GtkWidget *w, DWORD posx, DWORD posy)
{
	DWORD t;

	rectout.left = 0;
	rectout.top = 0;

	if ((setscmd & SCMD_SETTURN) == 0) {
		rectout.right = vir.w + displaymoder;
		rectout.bottom = vir.h;

		if (vir.x1) {
			t = posx - displaymodel;
			gdk_draw_rectangle(pixmap, w->style->black_gc, TRUE,
			    t, posy, t + vir.x1, posy + vir.sy);
		}
		if (vir.x2) {
			t = posx + vir.x1 + vir.w;
			gdk_draw_rectangle(pixmap, w->style->black_gc, TRUE,
			    t, posy, t + vir.x2, posy + vir.sy);
		}
	} else {
		rectout.right = vir.h;
		rectout.top = 0;

		if (setscmd & 0x40) {
			/* right turn */
			rectout.bottom = vir.w;

			if (vir.x1) {
				t = posy;
				gdk_draw_rectangle(pixmap, w->style->black_gc,
				    TRUE, posx, t, posx + vir.sy, t + vir.x1);
			}
			if (vir.x2) {
				t = posy + vir.x1 + vir.sx;
				gdk_draw_rectangle(pixmap, w->style->black_gc,
				    TRUE, posx, t, posx + vir.sy, t + vir.x2);
			}
		} else {
			rectout.top = SURFACE_WIDTH - vir.w;
			rectout.bottom = SURFACE_WIDTH;
			if (vir.x2) {
				t = posy;
				gdk_draw_rectangle(pixmap, w->style->black_gc,
				    TRUE, posx, t, posx + vir.sy, t + vir.x2);
			}
			if (vir.x1) {
				t = posy + vir.x2 + vir.w;
				gdk_draw_rectangle(pixmap, w->style->black_gc,
				    TRUE, posx, t, posx + vir.sy, t + vir.x1);
			}
		}
	}
}

void
xdraws_drawall(GtkWidget *w)
{
	GdkDrawable *d = (GdkDrawable *)(w->window);

	if (SCREENMODE & SCMD_USEPAL) {
		if (palchanged) {
			palchanged = 0;
			palette_set(w);
		}
	}

	if (allflash) {
		allflash = 0;
		clearblanklines(w, 0, 0);
	}

	gdk_draw_pixmap(d, w->style->fg_gc[GTK_WIDGET_STATE(w)], pixmap,
	    0, 0, 0, 0, winw, winh);
}

BYTE
xdraws_draws(GtkWidget *w, int redraw)
{
	GdkDrawable *d = (GdkDrawable *)pixmap;
	BYTE delta[SURFACE_HEIGHT+4];
	void (**func)(void);
	int pitch;
	int flag;
	int h, s;

	memcpy(delta, renewal_line, sizeof(delta));

	pitch = rectddr.right;
	switch (SCREENMODE & SCMD_COLORMASK) {
		case SCMD_COLOR32:
			func = layer32tbl;
			pitch *= 4;
			break;

		case SCMD_COLOR24:
			func = layer24tbl;
			pitch *= 3;
			break;

		case SCMD_COLOR16:
			func = layer16tbl;
			pitch *= 2;
			break;

		case SCMD_COLOR8:
			func = layer8tbl;
			break;

		default:
			func = NULL;
			break;
	}

	flag = 0;
	if ((setscmd & SCMD_SETTURN) == 0) {
		if (np2cfg.skipline)
			flag |= LAYER_EXTSKIPLINE;
		if (displaymoder)
			flag |= LAYER_NORDISP;
	} else {
		flag |= LAYER_ROL;
		if (setscmd & 0x40)
			flag |= 1;
	}

	np2layer_outpreparing((BOOL)redraw);
	np2layer_out(func, flag, surface->mem, pitch);
	if ((setscmd & SCMD_SETTURN) == 0) {
		/* normal */
		for (s = h = 0; h < rectddr.bottom; h++) {
			if (renewal_line[h] != delta[h])
				continue;
			if (s == h)
				continue;

			gdk_draw_image(d, w->style->fg_gc[GTK_WIDGET_STATE(w)],
			    surface, 0, s, np2cfg.paddingx, s + np2cfg.paddingy,
			    rectddr.right, h - s);
			s = h + 1;
		}
		if (s != h) {
			gdk_draw_image(d, w->style->fg_gc[GTK_WIDGET_STATE(w)],
			    surface, 0, s, np2cfg.paddingx, s + np2cfg.paddingy,
			    rectddr.right, h - s);
		}
	} else if (setscmd & 0x40) {
		/* rotate right */
		for (s = h = 0; h < rectddr.right; h++) {
			if (renewal_line[h] != delta[h])
				continue;
			if (s == h)
				continue;

			gdk_draw_image(d, w->style->fg_gc[GTK_WIDGET_STATE(w)],
			    surface,
			    rectddr.right - h, 0,
			    rectddr.right - h + np2cfg.paddingx,np2cfg.paddingy,
			    h - s, rectddr.bottom);
			s = h + 1;
		}
		if (s != h) {
			gdk_draw_image(d, w->style->fg_gc[GTK_WIDGET_STATE(w)],
			    surface,
			    rectddr.right - h, 0,
			    rectddr.right - h + np2cfg.paddingx,np2cfg.paddingy,
			    s - h, rectddr.bottom);
		}
	} else {
		/* rotate left */
		for (s = h = 0; h < rectddr.right; h++) {
			if (renewal_line[h] != delta[h])
				continue;
			if (s == h)
				continue;

			gdk_draw_image(d, w->style->fg_gc[GTK_WIDGET_STATE(w)],
			    surface, s, 0, s + np2cfg.paddingx, np2cfg.paddingy,
			    h - s, rectddr.bottom);
			s = h + 1;
		}
		if (s != h) {
			gdk_draw_image(d, w->style->fg_gc[GTK_WIDGET_STATE(w)],
			    surface, s, 0, s + np2cfg.paddingx, np2cfg.paddingy,
			    h - s, rectddr.bottom);
		}
	}

	return(np2layer_outresult());
}

void
xdraws_redraw(GtkWidget *w)
{

	allflash = 1;
	palchanged = 1;
	pal_change(1);
	xdraws_setdisplaymode();
	xdraws_draws(w, 1);
	xdraws_drawall(w);
}

BYTE
np2os_drawing(BYTE drawflag)
{
	BYTE ret = 0;

	if (drawflag) {
		ret = xdraws_draws(drawarea, drawflag & 2);
		xdraws_drawall(drawarea);
	}
	return 0;
}

void
xdraws_windowstats(GtkWidget *w)
{

	if ((setscmd & SCMD_SETTURN) == 0) {
		rectddr.right = SURFACE_WIDTH + 2 /* XXX + 1 */;
		rectddr.bottom = SURFACE_HEIGHT;
	} else {
		rectddr.right = SURFACE_HEIGHT;
		rectddr.bottom = SURFACE_WIDTH;
	}

	xdraws_clientmovesize(w);
}
