/*	$Id: gtkmenu.c,v 1.25 2003/07/24 15:38:54 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "compiler.h"

#include <gtk/gtkmenu.h>
#include <gdk/gdkkeysyms.h>

#include "np2.h"
#include "np2pc98.h"
#include "diskdrv.h"
#include "ini.h"
#include "caption.h"
#include "midi.h"
#include "dosio.h"

#include "xnp2.h"
#include "xdraw.h"
#include "xmouse.h"
#include "xmenu.h"

static void disable_item(char *);
static void select_item(char *);

static void disable_unused_items(void);

static void xmenu_select_framerate(int);
static void xmenu_select_beepvol(int);
static void xmenu_select_soundboard(int);
static void xmenu_select_extmem(int);
static void xmenu_select_mouse_move_ratio(int);

static void exit_from_menu(gpointer, guint, GtkWidget *);
static void reset(gpointer, guint, GtkWidget *);
static void rotate(gpointer, guint, GtkWidget *);

static void fddopen(gpointer, guint, GtkWidget *);
static void fddeject(gpointer, guint, GtkWidget *);
static void sasiopen(gpointer, guint, GtkWidget *);
static void sasiremove(gpointer, guint, GtkWidget *);

static void change_font(gpointer, guint, GtkWidget *);

static void new_disk(gpointer, guint, GtkWidget *);

static void framerate(gpointer, guint, GtkWidget *);
enum {
	FRAME_AUTO,
	FRAME_FULL,
	FRAME_30,
	FRAME_20,
	FRAME_15,
	NUM_FRAMERATE
};

static void keyboard(gpointer, guint, GtkWidget *);
static void f12(gpointer, guint, GtkWidget *);

static void beepvol(gpointer, guint, GtkWidget *);
enum {
	BEEP_OFF,
	BEEP_LOW,
	BEEP_MID,
	BEEP_HIGH,
	NUM_BEEP
};

static void sound_board(gpointer, guint, GtkWidget *);
enum {
	NOSOUNDBOARD = 0,
	PC9801_14 = 1,
	PC9801_26K = 2,
	PC9801_86 = 4,
	PC9801_118 = 8,
	CHIBI_OTO = 0x14,
	SPEAK_BOARD = 0x20,
	SPARK_BOARD = 0x40
};

static void memory(gpointer, guint, GtkWidget *);

static void mouse_ratio(gpointer, guint, GtkWidget *);
static void midi_panic(gpointer, guint, GtkWidget *);

static void toggle(gpointer, guint, GtkWidget *);

#define	_(f)	((GtkItemFactoryCallback)(f))
static GtkItemFactoryEntry menu_items[] = {
{ "/Emulate", 			NULL, NULL, 0, "<Branch>" },
{ "/Emulate/Reset", 		NULL, _(reset), 0, NULL },
{ "/Emulate/sep1",		NULL, NULL, 0, "<Separator>" },
{ "/Emulate/Configure",		NULL, _(create_configure_dialog), 0, NULL },
{ "/Emulate/New Disk",		NULL, NULL, 0, "<Branch>" },
{ "/Emulate/New Disk/FD image",	NULL, _(new_disk), 0, NULL },
{ "/Emulate/New Disk/HD image",	NULL, _(new_disk), 1, NULL },
{ "/Emulate/Font",		NULL, _(change_font), 0, NULL },
{ "/Emulate/sep2",		NULL, NULL, 0, "<Separator>" },
{ "/Emulate/Exit",		NULL, _(exit_from_menu), 0, NULL },
{ "/FDD",			NULL, NULL, 0, "<Branch>" },
{ "/FDD/Drive1",		NULL, NULL, 0, "<Branch>" },
{ "/FDD/Drive1/Open",		NULL, _(fddopen), 0, NULL },
{ "/FDD/Drive1/Eject",		NULL, _(fddeject), 0, NULL },
{ "/FDD/Drive2",		NULL, NULL, 0, "<Branch>" },
{ "/FDD/Drive2/Open",		NULL, _(fddopen), 1, NULL },
{ "/FDD/Drive2/Eject",		NULL, _(fddeject), 1, NULL },
{ "/FDD/sep1",			NULL, NULL, 0, "<Separator>" },
{ "/FDD/Eject All",		NULL, _(fddeject), ~0, NULL },
{ "/SASI",			NULL, NULL, 0, "<Branch>" },
{ "/SASI/SASI1",		NULL, NULL, 0, "<Branch>" },
{ "/SASI/SASI1/Open",		NULL, _(sasiopen), 0, NULL },
{ "/SASI/SASI1/Remove",		NULL, _(sasiremove), 0, NULL },
{ "/SASI/SASI2",		NULL, NULL, 0, "<Branch>" },
{ "/SASI/SASI2/Open",		NULL, _(sasiopen), 1, NULL },
{ "/SASI/SASI2/Remove",		NULL, _(sasiremove), 1, NULL },
{ "/SASI/sep1",			NULL, NULL, 0, "<Separator>" },
{ "/SASI/Remove All",		NULL, _(sasiremove), ~0, NULL },
{ "/Screen",			NULL, NULL, 0, "<Branch>" },
{ "/Screen/Window",		NULL, NULL, 0, "<RadioItem>" },
{ "/Screen/Fullscreen",		NULL, NULL, 0, "/Screen/Window" },
{ "/Screen/sep1",		NULL, NULL, 0, "<Separator>" },
{ "/Screen/Normal",		NULL, _(rotate), SCMD_SETNORMAL, "<RadioItem>"},
{ "/Screen/Left Rotated",	NULL, _(rotate), SCMD_SETLEFT,"/Screen/Normal"},
{ "/Screen/Right Rotated",	NULL, _(rotate), SCMD_SETRIGHT, "/Screen/Left Rotated" },
{ "/Screen/sep2",		NULL, NULL, 0, "<Separator>" },
{ "/Screen/Disp Vsync",		NULL, _(toggle), DISP_VSYNC, "<ToggleItem>"},
{ "/Screen/Real Palettes",	NULL, _(toggle), REAL_PALETTES, "<ToggleItem>"},
{ "/Screen/No Wait",		NULL, _(toggle), NO_WAIT, "<ToggleItem>" },
{ "/Screen/Auto frame",		NULL, _(framerate), FRAME_AUTO, "<RadioItem>" },
{ "/Screen/60 FPS",		NULL, _(framerate), FRAME_FULL, "/Screen/Auto frame" },
{ "/Screen/30 FPS",		NULL, _(framerate), FRAME_30, "/Screen/60 FPS"},
{ "/Screen/20 FPS",		NULL, _(framerate), FRAME_20, "/Screen/30 FPS"},
{ "/Screen/15 FPS",		NULL, _(framerate), FRAME_15, "/Screen/20 FPS"},
{ "/Screen/sep3",		NULL, NULL, 0, "<Separator>" },
{ "/Screen/Screen Option",	NULL, _(create_screen_dialog), 0, NULL },
{ "/Device",			NULL, NULL, 0, "<Branch>" },
{ "/Device/Keyboard",		NULL, NULL, 0, "<Branch>" },
{ "/Device/Keyboard/Keyboard",	NULL, _(keyboard), 0, "<RadioItem>" },
{ "/Device/Keyboard/Joykey-1",	NULL, _(keyboard), 1, "/Device/Keyboard/Keyboard"},
{ "/Device/Keyboard/Joykey-2",	NULL, _(keyboard), 2, "/Device/Keyboard/Joykey-1"},
{ "/Device/Keyboard/MouseKey",	NULL, _(keyboard), 3, "/Device/Keyboard/Joykey-2"},
{ "/Device/Keyboard/sep1",	NULL, NULL, 0, "<Separator>" },
{ "/Device/Keyboard/mechanical SHIFT",	NULL, _(toggle), XSHIFT_SHIFT, "<ToggleItem>" },
{ "/Device/Keyboard/mechanical CTRL",	NULL, _(toggle), XSHIFT_CTRL, "<ToggleItem>" },
{ "/Device/Keyboard/mechanical GRPH",	NULL, _(toggle), XSHIFT_GRPH, "<ToggleItem>" },
{ "/Device/Keyboard/sep2",	NULL, NULL, 0, "<Separator>" },
{ "/Device/Keyboard/F12 = Mouse", NULL, _(f12), 0, "<RadioItem>" },
{ "/Device/Keyboard/F12 = Copy", NULL, _(f12),1,"/Device/Keyboard/F12 = Mouse"},
{ "/Device/Keyboard/F12 = Stop", NULL, _(f12),2,"/Device/Keyboard/F12 = Copy" },
{ "/Device/Sound",		NULL, NULL, 0, "<Branch>" },
{ "/Device/Sound/Beep off",	NULL, _(beepvol), BEEP_OFF, "<RadioItem>" },
{ "/Device/Sound/Beep low",	NULL, _(beepvol), BEEP_LOW, "/Device/Sound/Beep off" },
{ "/Device/Sound/Beep mid",	NULL, _(beepvol), BEEP_MID, "/Device/Sound/Beep low" },
{ "/Device/Sound/Beep high",	NULL, _(beepvol), BEEP_HIGH, "/Device/Sound/Beep mid" },
{ "/Device/Sound/sep1",		NULL, NULL, 0, "<Separator>" },
{ "/Device/Sound/Disable board", NULL, _(sound_board), NOSOUNDBOARD, "<RadioItem>" },
{ "/Device/Sound/PC-9801-14",	NULL, _(sound_board), PC9801_14, "/Device/Sound/Disable board" },
{ "/Device/Sound/PC-9801-26K",	NULL, _(sound_board), PC9801_26K, "/Device/Sound/PC-9801-14" },
{ "/Device/Sound/PC-9801-86",	NULL, _(sound_board), PC9801_86, "/Device/Sound/PC-9801-26K" },
{ "/Device/Sound/PC-9801-118",	NULL, _(sound_board), PC9801_118, "/Device/Sound/PC-9801-86" },
{ "/Device/Sound/Chibi-oto",	NULL, _(sound_board), CHIBI_OTO, "/Device/Sound/PC-9801-118" },
{ "/Device/Sound/Speak board",	NULL, _(sound_board), SPEAK_BOARD, "/Device/Sound/Chibi-oto" },
{ "/Device/Sound/Spark board",	NULL, _(sound_board), SPARK_BOARD, "/Device/Sound/Speak board" },
{ "/Device/Sound/sep2",		NULL, NULL, 0, "<Separator>" },
{ "/Device/Sound/Seek Sound",	NULL, NULL, 0, "<ToggleItem>" },
{ "/Device/Memory",		NULL, NULL, 0, "<Branch>" },
{ "/Device/Memory/640KB",	NULL, _(memory), 0, "<RadioItem>" },
{ "/Device/Memory/1.6MB",	NULL, _(memory), 1, "/Device/Memory/640KB" },
{ "/Device/Memory/3.6MB",	NULL, _(memory), 2, "/Device/Memory/1.6MB" },
{ "/Device/Memory/7.6MB",	NULL, _(memory), 3, "/Device/Memory/3.6MB" },
{ "/Device/Mouse",		NULL, NULL, 0, "<Branch>" },
{ "/Device/Mouse/Mouse mode",	NULL, _(toggle), MOUSE_MODE, "<ToggleItem>" },
{ "/Device/Mouse/sep",		NULL, NULL, 0, "<Separator>" },
{ "/Device/Mouse/x0.5",		NULL, _(mouse_ratio), MOUSE_RATIO_050, "<RadioItem>" },
{ "/Device/Mouse/x0.75",	NULL, _(mouse_ratio), MOUSE_RATIO_075, "/Device/Mouse/x0.5" },
{ "/Device/Mouse/x1.0",		NULL, _(mouse_ratio), MOUSE_RATIO_100, "/Device/Mouse/x0.75" },
{ "/Device/Mouse/x1.5",		NULL, _(mouse_ratio), MOUSE_RATIO_150, "/Device/Mouse/x1.0" },
{ "/Device/Mouse/x2.0",		NULL, _(mouse_ratio), MOUSE_RATIO_200, "/Device/Mouse/x1.5" },
{ "/Device/Mouse/x4.0",		NULL, _(mouse_ratio), MOUSE_RATIO_400, "/Device/Mouse/x2.0" },
{ "/Device/Mouse/x8.0",		NULL, _(mouse_ratio), MOUSE_RATIO_800, "/Device/Mouse/x4.0" },
{ "/Device/sep2",		NULL, NULL, 0, "<Separator>" },
{ "/Device/MIDI Option",	NULL, NULL, 0, NULL },
{ "/Device/MIDI Panic",		NULL, _(midi_panic), 0, NULL },
{ "/Device/sep3",		NULL, NULL, 0, "<Separator>" },
{ "/Device/Sound Option",	NULL, _(create_sound_dialog), 0, NULL },
{ "/Other",			NULL, NULL, 0, "<Branch>" },
{ "/Other/BMP Save",		NULL, NULL, 0, NULL },
{ "/Other/S98 logging",		NULL, NULL, 0, NULL },
{ "/Other/Clock Disp",		NULL, _(toggle), CLOCK_DISP, "<ToggleItem>" },
{ "/Other/Frame Disp",		NULL, _(toggle), FRAME_DISP, "<ToggleItem>" },
{ "/Other/Joy Reverse",		NULL, _(toggle), JOY_REVERSE, "<ToggleItem>" },
{ "/Other/Joy Rapid",		NULL, _(toggle), JOY_RAPID, "<ToggleItem>" },
{ "/Other/Mouse Rapid",		NULL, _(toggle), MOUSE_RAPID, "<ToggleItem>" },
{ "/Other/sep",			NULL, NULL, 0, "<Separator>" },
{ "/Other/About",		NULL, _(create_about_dialog), 0, NULL },
};
#undef	_
static gint nmenu_items = NELEMENTS(menu_items);

typedef struct toggle_tag {
	char *name;
	int stat;
} toggle_t;

static toggle_t toggle_items[] = {
	{ "/Screen/Disp Vsync",			0 },
	{ "/Screen/Real Palettes",		0 },
	{ "/Screen/No Wait",			0 },
	{ "/Device/Sound/Seek Sound",		0 },
	{ "/Device/Mouse/Mouse mode",		0 },
	{ "/Other/Clock Disp",			0 },
	{ "/Other/Frame Disp",			0 },
	{ "/Device/Keyboard/mechanical SHIFT",	0 },
	{ "/Device/Keyboard/mechanical CTRL",	0 },
	{ "/Device/Keyboard/mechanical GRPH",	0 },
	{ "/Other/Joy Reverse",			0 },
	{ "/Other/Joy Rapid",			0 },
	{ "/Other/Mouse Rapid",			0 },
};
static gint ntoggle_items = NELEMENTS(toggle_items);

static GtkItemFactory *item_factory;
static int inited;

/*
 * init
 */
GtkWidget *
create_menu(GtkWidget *w)
{
	GtkAccelGroup *accel_group;

	accel_group = gtk_accel_group_new();
	item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", accel_group);
	gtk_item_factory_create_items(item_factory, nmenu_items, menu_items, NULL);
	gtk_accel_group_attach(accel_group, GTK_OBJECT(w));

	disable_unused_items();

	xmenu_toggle_item(DISP_VSYNC, np2cfg.DISPSYNC, TRUE);
	xmenu_toggle_item(REAL_PALETTES, np2cfg.RASTER, TRUE);
	xmenu_toggle_item(NO_WAIT, np2cfg.NOWAIT, TRUE);
	xmenu_toggle_item(SEEK_SOUND, np2cfg.MOTOR, TRUE);
	xmenu_toggle_item(MOUSE_MODE, np2cfg.MOUSE_SW, TRUE);
	xmenu_toggle_item(CLOCK_DISP, np2cfg.DISPCLK & 1, TRUE);
	xmenu_toggle_item(FRAME_DISP, np2cfg.DISPCLK & 2, TRUE);
	xmenu_toggle_item(XSHIFT_SHIFT, np2cfg.XSHIFT & 1, TRUE);
	xmenu_toggle_item(XSHIFT_CTRL, np2cfg.XSHIFT & 2, TRUE);
	xmenu_toggle_item(XSHIFT_GRPH, np2cfg.XSHIFT & 4, TRUE);
	xmenu_toggle_item(JOY_REVERSE, np2cfg.BTN_MODE, TRUE);
	xmenu_toggle_item(JOY_RAPID, np2cfg.BTN_RAPID, TRUE);
	xmenu_toggle_item(MOUSE_RAPID, np2cfg.MOUSERAPID, TRUE);

	xmenu_select_framerate(np2cfg.DRAW_SKIP);
	xmenu_select_beepvol(np2cfg.BEEP_VOL);
	xmenu_select_soundboard(np2cfg.SOUND_SW);
	xmenu_select_extmem(np2cfg.EXTMEM);
	xmenu_select_mouse_move_ratio(np2cfg.mouse_move_ratio);

	inited = 1;

	return gtk_item_factory_get_widget(item_factory, "<main>");
}

static void
disable_unused_items(void)
{
	static char *items[] = {
		"/Screen/Fullscreen",
		"/Device/Sound/PC-9801-14",
		"/Device/Sound/PC-9801-26K",
		"/Device/Sound/PC-9801-118",
		"/Device/Sound/Chibi-oto",
		"/Device/Sound/Speak board",
		"/Device/Sound/Spark board",
		"/Device/Sound/Seek Sound",
		"/Other/BMP Save",
		"/Other/S98 logging",
	};
	int i;

	for (i = 0; i < NELEMENTS(items); i++)
		disable_item(items[i]);
}

/*
 * misc
 */
static void
disable_item(char *name)
{
	GtkWidget *menu_item;

	menu_item = gtk_item_factory_get_widget(item_factory, name);
	gtk_widget_set_sensitive(menu_item, FALSE);
}

static void
select_item(char *name)
{
	GtkWidget *menu_item;

	menu_item = gtk_item_factory_get_widget(item_factory, name);
	gtk_signal_emit_by_name(GTK_OBJECT(menu_item), "activate-item");
}

void
xmenu_toggle_item(int arg, int onoff, int emitp)
{

	if (arg >= ntoggle_items)
		return;

	if (onoff != toggle_items[arg].stat) {
		toggle_items[arg].stat = onoff;
		if (emitp)
			select_item(toggle_items[arg].name);

		if (inited && arg != MOUSE_MODE) {
			INIT_RENEWAL;
		}
	}
}

static void
xmenu_select_framerate(int kind)
{
	static char *name[NUM_FRAMERATE] = {
		"/Screen/Auto frame",
		"/Screen/60 FPS",
		"/Screen/30 FPS",
		"/Screen/20 FPS",
		"/Screen/15 FPS",
	};

	if (kind >= NUM_FRAMERATE)
		return;

	select_item(name[kind]);
}

static void
xmenu_select_beepvol(int kind)
{
	static char *name[NUM_BEEP] = {
		"/Device/Sound/Beep off",
		"/Device/Sound/Beep low",
		"/Device/Sound/Beep mid",
		"/Device/Sound/Beep high",
	};

	if (kind >= NUM_BEEP)
		return;

	select_item(name[kind]);
}

static void
xmenu_select_soundboard(int kind)
{
	static struct {
		gchar *name;
		gint num;
	} soundboard[] = {
		{ "/Device/Sound/Disable board", NOSOUNDBOARD },
		{ "/Device/Sound/PC-9801-14", PC9801_14 },
		{ "/Device/Sound/PC-9801-26K", PC9801_26K },
		{ "/Device/Sound/PC-9801-86", PC9801_86 },
		{ "/Device/Sound/PC-9801-118", PC9801_118 },
		{ "/Device/Sound/Chibi-oto", CHIBI_OTO },
		{ "/Device/Sound/Speak board", SPEAK_BOARD },
		{ "/Device/Sound/Spark board", SPARK_BOARD },
	};
	gint i;

	for (i = 0; i < NELEMENTS(soundboard); i++) {
		if (kind == soundboard[i].num) {
			select_item(soundboard[i].name);
			break;
		}
	}
}

static void
xmenu_select_extmem(int kind)
{
	static struct {
		gchar *name;
		gint num;
	} extmem[] = {
		{ "/Device/Memory/640KB", 0 },
		{ "/Device/Memory/1.6MB", 1 },
		{ "/Device/Memory/3.6MB", 3 },
		{ "/Device/Memory/7.6MB", 7 },
	};
	gint i;

	for (i = 0; i < NELEMENTS(extmem); i++) {
		if (kind == extmem[i].num) {
			select_item(extmem[i].name);
			return;
		}
	}
	select_item("/Device/Memory/1.6MB");
}

static void
xmenu_select_mouse_move_ratio(int kind)
{
	static struct {
		gchar *name;
		gint num;
	} moveratio[] = {
		{ "/Device/Mouse/x0.5", MOUSE_RATIO_050 },
		{ "/Device/Mouse/x0.75", MOUSE_RATIO_075 },
		{ "/Device/Mouse/x1.0", MOUSE_RATIO_100 },
		{ "/Device/Mouse/x1.5", MOUSE_RATIO_150 },
		{ "/Device/Mouse/x2.0", MOUSE_RATIO_200 },
		{ "/Device/Mouse/x4.0", MOUSE_RATIO_400 },
		{ "/Device/Mouse/x8.0", MOUSE_RATIO_800 },
	};
	gint i;

	for (i = 0; i < NELEMENTS(moveratio); i++) {
		if (kind == moveratio[i].num) {
			select_item(moveratio[i].name);
			return;
		}
	}
	select_item("/Device/Mouse/x1.0");
}

/*
 * item function
 */
static void
exit_from_menu(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(action);
	UNUSED(w);

	gtk_widget_destroy(GTK_WIDGET(window));
}

static void
reset(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(action);
	UNUSED(w);

	if (np2_getmemsw()) {
		INIT_RENEWAL;
	}
	np2os_reset();
	np2_reset();
}

static void
rotate(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(w);

	changescreen(drawarea, action);
}

static void
fddopen(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(w);

	create_file_selection(FILE_TYPE_FDD, "Open FDD image",
	    fddfolder, action);
}

static void
fddeject(gpointer data, guint action, GtkWidget *w)
{
	int i;

	UNUSED(data);
	UNUSED(w);

	if (action == (guint)~0) {
		for (i = 0; i < 4; i++)
			diskdrv_ejectfdd(i);
	} else {
		diskdrv_ejectfdd(action);
	}

}

static void
sasiopen(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(w);

	create_file_selection(FILE_TYPE_SASI, "Open SASI image",
	    hddfolder, action);
}

static void
sasiremove(gpointer data, guint action, GtkWidget *w)
{
	int i;

	UNUSED(data);
	UNUSED(w);

	if (action == (guint)~0) {
		for (i = 0; i < 2; i++)
			diskdrv_ejecthdd(i);
	} else {
		diskdrv_ejecthdd(action);
	}

}

static void
change_font(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(action);
	UNUSED(w);

	create_file_selection(FILE_TYPE_FONT, "Open font file",
	    (char *)fontfile, 0);
}

static void
new_disk(gpointer data, guint action, GtkWidget *w)
{
	char newdiskfolder[MAX_PATH + 8];
	int type;
	char *title;

	UNUSED(data);
	UNUSED(w);

	switch (action) {
	case 0:
		type = FILE_TYPE_NEW_FD_IMAGE;
		title = "Create new floppy disk image file";
		break;

	case 1:
		type = FILE_TYPE_NEW_HD_IMAGE;
		title = "Create new hard disk image file";
		break;

	default:
		return;
	}

	strcpy(newdiskfolder, fddfolder);
	cutFileName(newdiskfolder);
	plusyen(newdiskfolder, sizeof(newdiskfolder) - 1);
	strcat(newdiskfolder, "newdisk");
	create_file_selection(type, title, newdiskfolder, 0);
}

static void
framerate(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(w);

	if (action != np2cfg.DRAW_SKIP) {
		INIT_RENEWAL;
		np2cfg.DRAW_SKIP = action;
	}
}

static void
keyboard(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(w);

	if (action >= 4)
		action = 0;

	if (np2cfg.KEY_MODE != action) {
		INIT_RENEWAL;
		np2cfg.KEY_MODE = action;
		np2key_resetjoykey();
	}
}

static void
f12(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(w);

	if (action >= 3)
		action = 0;

	if (np2cfg.F12COPY != action) {
		INIT_RENEWAL;
		np2cfg.F12COPY = action;
		np2key_resetcopyhelp();
	}
}

static void
beepvol(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(w);

	if (action != np2cfg.BEEP_VOL) {
		INIT_RENEWAL;
		np2cfg.BEEP_VOL = action;
	}
}

static void
sound_board(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(w);

	switch (action) {
	case NOSOUNDBOARD:
	case PC9801_86:
		if (action != np2cfg.SOUND_SW) {
			INIT_RENEWAL;
			np2cfg.SOUND_SW = action;
		}
		break;

	case PC9801_14:
	case PC9801_26K:
	case PC9801_118:
	case CHIBI_OTO:
	case SPEAK_BOARD:
	case SPARK_BOARD:
		break;
	}
}

static void
memory(gpointer data, guint action, GtkWidget *w)
{
	static const BYTE extmem[] = { 0, 1, 3, 7 };

	UNUSED(data);
	UNUSED(w);

	if (action >= NELEMENTS(extmem))
		action = 1;

	if (np2cfg.EXTMEM != extmem[action]) {
		INIT_RENEWAL;
		np2cfg.EXTMEM = extmem[action];
	}
}

static void
mouse_ratio(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(w);

	switch (action) {
	case MOUSE_RATIO_050:
	case MOUSE_RATIO_075:
	case MOUSE_RATIO_100:
	case MOUSE_RATIO_150:
	case MOUSE_RATIO_200:
	case MOUSE_RATIO_400:
	case MOUSE_RATIO_800:
		if (action != np2cfg.mouse_move_ratio) {
			INIT_RENEWAL;
			np2cfg.mouse_move_ratio = action;
		}
		mouse_move_ratio = np2cfg.mouse_move_ratio;
		mouse_move_mul = (mouse_move_ratio >> 4) & 0xf;
		mouse_move_div = mouse_move_ratio & 0xf;
		break;
	}
}

static void
midi_panic(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(action);
	UNUSED(w);

	MIDI_AllToneOff();
}

static void
toggle(gpointer data, guint action, GtkWidget *w)
{

	UNUSED(data);
	UNUSED(w);

	if (!inited)
		return;

	switch (action) {
	case DISP_VSYNC:
		np2cfg.DISPSYNC = !np2cfg.DISPSYNC;
		xmenu_toggle_item(DISP_VSYNC, np2cfg.DISPSYNC, FALSE);
		break;

	case REAL_PALETTES:
		np2cfg.RASTER = !np2cfg.RASTER;
		xmenu_toggle_item(REAL_PALETTES, np2cfg.RASTER, FALSE);
		break;

	case NO_WAIT:
		np2cfg.NOWAIT = !np2cfg.NOWAIT;
		xmenu_toggle_item(NO_WAIT, np2cfg.NOWAIT, FALSE);
		break;

	case MOUSE_MODE:
		np2cfg.MOUSE_SW = !np2cfg.MOUSE_SW;
		xmenu_toggle_item(MOUSE_MODE, np2cfg.MOUSE_SW, FALSE);
		xmouse_show(window, MOUSE_TOGGLE);
		break;

	case CLOCK_DISP:
		np2cfg.DISPCLK ^= 1;
		xmenu_toggle_item(CLOCK_DISP, np2cfg.DISPCLK & 1, FALSE);
		caption_statreset();
		caption_disp(2);
		break;

	case FRAME_DISP:
		np2cfg.DISPCLK ^= 2;
		xmenu_toggle_item(FRAME_DISP, np2cfg.DISPCLK & 2, FALSE);
		caption_statreset();
		caption_disp(4);
		break;

	case XSHIFT_SHIFT:
		np2cfg.XSHIFT ^= 1;
		xmenu_toggle_item(XSHIFT_SHIFT, np2cfg.XSHIFT & 1, FALSE);
		np2key_forcerelease(0x70);
		break;

	case XSHIFT_CTRL:
		np2cfg.XSHIFT ^= 2;
		xmenu_toggle_item(XSHIFT_SHIFT, np2cfg.XSHIFT & 2, FALSE);
		np2key_forcerelease(0x74);
		break;

	case XSHIFT_GRPH:
		np2cfg.XSHIFT ^= 4;
		xmenu_toggle_item(XSHIFT_SHIFT, np2cfg.XSHIFT & 4, FALSE);
		np2key_forcerelease(0x73);
		break;

	case JOY_REVERSE:
		np2cfg.BTN_MODE = !np2cfg.BTN_MODE;
		xmenu_toggle_item(JOY_REVERSE, np2cfg.BTN_MODE, FALSE);
		break;

	case JOY_RAPID:
		np2cfg.BTN_RAPID = !np2cfg.BTN_RAPID;
		xmenu_toggle_item(JOY_RAPID, np2cfg.BTN_RAPID, FALSE);
		break;

	case MOUSE_RAPID:
		np2cfg.MOUSERAPID = !np2cfg.MOUSERAPID;
		xmenu_toggle_item(MOUSE_RAPID, np2cfg.MOUSERAPID, FALSE);
		break;

	case NUM_TOGGLE_ITEMS:
	default:
		break;
	}
}
