#include	"compiler.h"

#include	"np2.h"
#include	"np2pc98.h"
#include	"pc_core.h"
#include	"palettes.h"
#include	"layer.h"


//	16bitߡ
static void lay16(void) {

	BYTE	*d = layer.surface;
	int		x;

	do {
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[NP2PAL_TEXT2];
			}
		}
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
}


//	text or grph 1ץ졼 / 16bit
static void np2layer16_1(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s[x] + NP2PAL_GRPH];
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph / 16bit
static void np2layer16_tg(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s1[x] + s2[x] + NP2PAL_GRPH];
			}
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


//	text + (grph:skipline) / 16bit
static void np2layer16_t_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s[x] + NP2PAL_GRPH];
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[(s[x] >> 4) + NP2PAL_TEXT];
			}
		}
		s += SURFACE_WIDTH;										// 
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	grph + skipline / 16bit
static void np2layer16_g_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s[x] + NP2PAL_GRPH];
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[NP2PAL_TEXT];
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph + skipline / 16bit
static void np2layer16_tg_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s1[x] + s2[x] + NP2PAL_GRPH];
			}
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[(s2[x] >> 4) + NP2PAL_TEXT];
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


// --- with extend skipline

//	grph + extend skipline / 16bit
static void np2layer16_g_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s[x] + NP2PAL_GRPH];
			}
		}
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s[x] + NP2PAL_SKIP];
			}
		}
		s += (SURFACE_WIDTH * 2);
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

//	text + grph + extend skipline / 16bit
static void np2layer16_tg_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s1[x] + s2[x] + NP2PAL_GRPH];
			}
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				BYTE	c;
				c = s2[x] >> 4;
				if (!c) {
					c = s1[x] + NP2PALS_TXT;
				}
				((WORD *)d)[x] = np2_pal16[c + NP2PAL_TEXT];
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}



// ----------------------------------------- support normal display

#ifdef SUPPORT_NORMALDISP

//	16bitߡ
static void lay16n(void) {

	BYTE	*d = layer.surface;
	int		x;

	do {
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (x=0; x<=layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[NP2PAL_TEXT2];
			}
		}
		d += layer.width;
	} while(++layer.y < layer.maxy);
	layer.surface = d;
}

//	text / 16bit
static void np2layer16n_t(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s[x] + NP2PAL_GRPH];
			}
			((WORD *)d)[x] = np2_pal16[NP2PAL_GRPH];
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

//	grph / 16bit
static void np2layer16n_g(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((WORD *)d)[0] = 0;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x + 1] = np2_pal16[s[x] + NP2PAL_GRPH];
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph / 16bit
static void np2layer16n_tg(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((WORD *)d)[0] = np2_pal16[(s2[0] >> 4) + NP2PAL_TEXT3];
			for (x=1; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s1[x-1] + s2[x] + NP2PAL_GRPH];
			}
			((WORD *)d)[x] = np2_pal16[s1[x-1] + NP2PAL_GRPH];
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


//	text + (grph:skipline) / 16bit
static void np2layer16n_t_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((WORD *)d)[0] = np2_pal16[(s[0] >> 4) + NP2PAL_TEXT3];
			for (x=1; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s[x] + NP2PAL_GRPH];
			}
			((WORD *)d)[x] = np2_pal16[NP2PAL_GRPH];
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((WORD *)d)[0] = np2_pal16[(s[0] >> 4) + NP2PAL_TEXT3];
			for (x=1; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[(s[x] >> 4) + NP2PAL_TEXT];
			}
			((WORD *)d)[x] = np2_pal16[NP2PAL_TEXT];
		}
		s += SURFACE_WIDTH;										// 
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	grph + skipline / 16bit
static void np2layer16n_g_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((WORD *)d)[0] = 0;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x + 1] = np2_pal16[s[x] + NP2PAL_GRPH];
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((WORD *)d)[0] = np2_pal16[NP2PAL_TEXT];
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x + 1] = np2_pal16[NP2PAL_TEXT];
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph + skipline / 16bit
static void np2layer16n_tg_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((WORD *)d)[0] = np2_pal16[(s2[0] >> 4) + NP2PAL_TEXT3];
			for (x=1; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s1[x-1] + s2[x] + NP2PAL_GRPH];
			}
			((WORD *)d)[x] = np2_pal16[s1[x-1] + NP2PAL_GRPH];
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((WORD *)d)[0] = np2_pal16[(s2[0] >> 4) + NP2PAL_TEXT3];
			for (x=1; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[(s2[x] >> 4) + NP2PAL_TEXT];
			}
			((WORD *)d)[x] = np2_pal16[NP2PAL_TEXT];
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


// --- with extend skipline

//	grph + extend skipline / 16bit
static void np2layer16n_g_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			((WORD *)d)[0] = 0;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x+1] = np2_pal16[s[x] + NP2PAL_GRPH];
			}
		}
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((WORD *)d)[0] = 0;
			for (x=0; x<layer.x; x++) {
				((WORD *)d)[x+1] = np2_pal16[s[x] + NP2PAL_SKIP];
			}
		}
		s += (SURFACE_WIDTH * 2);
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

//	text + grph + extend skipline / 16bit
static void np2layer16n_tg_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			((WORD *)d)[0] = np2_pal16[(s2[0] >> 4) + NP2PAL_TEXT3];
			for (x=1; x<layer.x; x++) {
				((WORD *)d)[x] = np2_pal16[s1[x-1] + s2[x] + NP2PAL_GRPH];
			}
			((WORD *)d)[x] = np2_pal16[s1[x-1] + NP2PAL_GRPH];
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((WORD *)d)[0] =np2_pal16[(s2[0] >> 4) + NP2PAL_TEXT3];
			for (x=1; x<layer.x; x++) {
				BYTE	c;
				c = s2[x] >> 4;
				if (!c) {
					c = s1[x-1] + NP2PALS_TXT;
				}
				((WORD *)d)[x] = np2_pal16[c + NP2PAL_TEXT];
			}
			((WORD *)d)[x] = np2_pal16[s1[x-1] + NP2PAL_SKIP];
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}

#endif



// ----------------------------------------------------------------- turn left

static void lay16l(void) {

	BYTE	*d = layer.surface;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (r=d, x=layer.x; x--;) {
				*((WORD *)r) = np2_pal16[NP2PAL_TEXT2];
				r += layer.width;
			}
		}
		d += sizeof(WORD);
	} while(++layer.y < layer.maxy);
	layer.surface = d;
}


static void np2layer16l_1(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((WORD *)r) = np2_pal16[s[x] + NP2PAL_GRPH];
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(WORD);
	} while(++layer.y < layer.maxy);
	layer.surface = d;
	layer.src1scrn = s;
}

static void np2layer16l_tg(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((WORD *)r) = np2_pal16[s1[x] + s2[x] + NP2PAL_GRPH];
				r += layer.width;
			}
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
		d += sizeof(WORD);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}

static void np2layer16l_t_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((WORD *)r) = np2_pal16[s[x] + NP2PAL_GRPH];
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(WORD);
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((WORD *)r) = np2_pal16[(s[x] >> 4) + NP2PAL_TEXT];
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(WORD);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

static void np2layer16l_g_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((WORD *)r) = np2_pal16[s[x] + NP2PAL_GRPH];
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(WORD);
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((WORD *)r) = np2_pal16[NP2PAL_TEXT];
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(WORD);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

static void np2layer16l_tg_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((WORD *)r) = np2_pal16[s1[x] + s2[x] + NP2PAL_GRPH];
				r += layer.width;
			}
		}
		s2 += SURFACE_WIDTH;
		d += sizeof(WORD);
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((WORD *)r) = np2_pal16[(s2[x] >> 4) + NP2PAL_TEXT];
				r += layer.width;
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += sizeof(WORD);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}



// ---------------------------------------------------------------- turn right

static void lay16r(void) {

	BYTE	*d;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(WORD));
	do {
		d -= sizeof(WORD);
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (r=d, x=0; x<layer.x; x++) {
				*((WORD *)r) = np2_pal16[NP2PAL_TEXT2];
				r += layer.width;
			}
		}
	} while(++layer.y < layer.maxy);
}


static void np2layer16r_1(void) {

	BYTE	*d;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(WORD));
	do {
		d -= sizeof(WORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((WORD *)r) = np2_pal16[s[x] + NP2PAL_GRPH];
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s;
}

static void np2layer16r_tg(void) {

	BYTE	*d;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(WORD));
	do {
		d -= sizeof(WORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((WORD *)r) = np2_pal16[s1[x] + s2[x] + NP2PAL_GRPH];
				r += layer.width;
			}
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s1;
	layer.src2scrn = s2;
}

static void np2layer16r_t_skip(void) {

	BYTE	*d;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(WORD));
	do {
		d -= sizeof(WORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((WORD *)r) = np2_pal16[s[x] + NP2PAL_GRPH];
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		layer.y++;
		d -= sizeof(WORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((WORD *)r) = np2_pal16[(s[x] >> 4) + NP2PAL_TEXT];
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s;
}

static void np2layer16r_g_skip(void) {

	BYTE	*d;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(WORD));
	do {
		d -= sizeof(WORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((WORD *)r) = np2_pal16[s[x] + NP2PAL_GRPH];
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		layer.y++;
		d -= sizeof(WORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((WORD *)r) = np2_pal16[NP2PAL_TEXT];
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s;
}

static void np2layer16r_tg_skip(void) {

	BYTE	*d;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(WORD));
	do {
		d -= sizeof(WORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((WORD *)r) = np2_pal16[s1[x] + s2[x] + NP2PAL_GRPH];
				r += layer.width;
			}
		}
		s2 += SURFACE_WIDTH;
		layer.y++;
		d -= sizeof(WORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((WORD *)r) = np2_pal16[(s2[x] >> 4) + NP2PAL_TEXT];
				r += layer.width;
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s1;
	layer.src2scrn = s2;
}





void (*layer16tbl[])(void) = {

		// Ρޥ
		lay16,	np2layer16_1,		np2layer16_1,		np2layer16_tg,
		lay16,	np2layer16_t_skip,	np2layer16_g_skip,	np2layer16_tg_skip,
		lay16,	np2layer16_t_skip,	np2layer16_g_ext,	np2layer16_tg_ext,

		lay16n,	np2layer16n_t,		np2layer16n_g,		np2layer16n_tg,
		lay16n,	np2layer16n_t_skip,	np2layer16n_g_skip,	np2layer16n_tg_skip,
		lay16n,	np2layer16n_t_skip,	np2layer16n_g_ext,	np2layer16n_tg_ext,


		// 
		lay16l,	np2layer16l_1,		np2layer16l_1,		np2layer16l_tg,
		lay16l,	np2layer16l_t_skip,	np2layer16l_g_skip,	np2layer16l_tg_skip,
		lay16l,	np2layer16l_t_skip,	np2layer16l_g_skip,	np2layer16l_tg_skip,

		lay16l,	np2layer16l_1,		np2layer16l_1,		np2layer16l_tg,
		lay16l,	np2layer16l_t_skip,	np2layer16l_g_skip,	np2layer16l_tg_skip,
		lay16l,	np2layer16l_t_skip,	np2layer16l_g_skip,	np2layer16l_tg_skip,


		// 
		lay16r,	np2layer16r_1,		np2layer16r_1,		np2layer16r_tg,
		lay16r,	np2layer16r_t_skip,	np2layer16r_g_skip,	np2layer16r_tg_skip,
		lay16r,	np2layer16r_t_skip,	np2layer16r_g_skip,	np2layer16r_tg_skip,

		lay16r,	np2layer16r_1,		np2layer16r_1,		np2layer16r_tg,
		lay16r,	np2layer16r_t_skip,	np2layer16r_g_skip,	np2layer16r_tg_skip,
		lay16r,	np2layer16r_t_skip,	np2layer16r_g_skip,	np2layer16r_tg_skip,
};


