#include	"compiler.h"

#include	"np2.h"
#include	"np2pc98.h"
#include	"pc_core.h"
#include	"palettes.h"
#include	"layer.h"


//	32bitߡ
static void lay32(void) {

	BYTE	*d = layer.surface;
	int		x;

	do {
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[NP2PAL_TEXT2].d;
			}
		}
		d += layer.width;
	} while(++layer.y < layer.maxy);
	layer.surface = d;
}


//	text or grph 1ץ졼 / 32bit
static void np2layer32_1(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s[x] + NP2PAL_GRPH].d;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph / 32bit
static void np2layer32_tg(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s1[x] + s2[x] + NP2PAL_GRPH].d;
			}
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


//	text + (grph:skipline) / 32bit
static void np2layer32_t_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s[x] + NP2PAL_GRPH].d;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[(s[x] >> 4) + NP2PAL_TEXT].d;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	grph:skipline / 32bit
static void np2layer32_g_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s[x] + NP2PAL_GRPH].d;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[NP2PAL_TEXT].d;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph:skipline / 32bit
static void np2layer32_tg_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s1[x] + s2[x] + NP2PAL_GRPH].d;
			}
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[(s2[x] >> 4) + NP2PAL_TEXT].d;
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


// --- with extend skipline

//	grph + extend skipline / 32bit
static void np2layer32_g_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s[x] + NP2PAL_GRPH].d;
			}
		}
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s[x] + NP2PAL_SKIP].d;
			}
		}
		s += (SURFACE_WIDTH * 2);
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

//	text + grph + extend skipline / 32bit
static void np2layer32_tg_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s1[x] + s2[x] + NP2PAL_GRPH].d;
			}
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				BYTE	c;
				c = s2[x] >> 4;
				if (!c) {
					c = s1[x] + NP2PALS_TXT;
				}
				((DWORD *)d)[x] = np2_pal32[c + NP2PAL_TEXT].d;
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


// ----------------------------------------- support normal display

#ifdef SUPPORT_NORMALDISP

//	32bitߡ
static void lay32n(void) {

	BYTE	*d = layer.surface;
	int		x;

	do {
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (x=0; x<=layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[NP2PAL_TEXT2].d;
			}
		}
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
}

//	text / 32bit
static void np2layer32n_t(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s[x] + NP2PAL_GRPH].d;
			}
			((DWORD *)d)[x] = np2_pal32[NP2PAL_GRPH].d;
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

//	grph / 32bit
static void np2layer32n_g(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((DWORD *)d)[0] = 0;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x + 1] = np2_pal32[s[x] + NP2PAL_GRPH].d;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph / 32bit
static void np2layer32n_tg(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((DWORD *)d)[0] = np2_pal32[(s2[0] >> 4) + NP2PAL_TEXT3].d;
			for (x=1; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s1[x-1] + s2[x] + NP2PAL_GRPH].d;
			}
			((DWORD *)d)[x] = np2_pal32[s1[x-1] + NP2PAL_GRPH].d;
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


//	text + (grph:skipline) / 32bit
static void np2layer32n_t_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((DWORD *)d)[0] = np2_pal32[(s[0] >> 4) + NP2PAL_TEXT3].d;
			for (x=1; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s[x] + NP2PAL_GRPH].d;
			}
			((DWORD *)d)[x] = np2_pal32[NP2PAL_GRPH].d;
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((DWORD *)d)[0] = np2_pal32[(s[0] >> 4) + NP2PAL_TEXT3].d;
			for (x=1; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[(s[x] >> 4) + NP2PAL_TEXT].d;
			}
			((DWORD *)d)[x] = np2_pal32[NP2PAL_TEXT].d;
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	grph + skipline / 32bit
static void np2layer32n_g_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((DWORD *)d)[0] = 0;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x + 1] = np2_pal32[s[x] + NP2PAL_GRPH].d;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((DWORD *)d)[0] = np2_pal32[NP2PAL_TEXT].d;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x + 1] = np2_pal32[NP2PAL_TEXT].d;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph + skipline / 32bit
static void np2layer32n_tg_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((DWORD *)d)[0] = np2_pal32[(s2[0] >> 4) + NP2PAL_TEXT3].d;
			for (x=1; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s1[x-1] + s2[x] + NP2PAL_GRPH].d;
			}
			((DWORD *)d)[x] = np2_pal32[s1[x-1] + NP2PAL_GRPH].d;
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((DWORD *)d)[0] = np2_pal32[(s2[0] >> 4) + NP2PAL_TEXT3].d;
			for (x=1; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[(s2[x] >> 4) + NP2PAL_TEXT].d;
			}
			((DWORD *)d)[x] = np2_pal32[NP2PAL_TEXT].d;
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


// --- with extend skipline

//	grph + extend skipline / 32bit
static void np2layer32n_g_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			((DWORD *)d)[0] = 0;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x+1] = np2_pal32[s[x] + NP2PAL_GRPH].d;
			}
		}
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((DWORD *)d)[0] = 0;
			for (x=0; x<layer.x; x++) {
				((DWORD *)d)[x+1] = np2_pal32[s[x] + NP2PAL_SKIP].d;
			}
		}
		s += (SURFACE_WIDTH * 2);
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

//	text + grph + extend skipline / 32bit
static void np2layer32n_tg_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			((DWORD *)d)[0] = np2_pal32[(s2[0] >> 4) + NP2PAL_TEXT3].d;
			for (x=1; x<layer.x; x++) {
				((DWORD *)d)[x] = np2_pal32[s1[x-1] + s2[x] + NP2PAL_GRPH].d;
			}
			((DWORD *)d)[x] = np2_pal32[s1[x-1] + NP2PAL_GRPH].d;
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			((DWORD *)d)[0] =np2_pal32[(s2[0] >> 4) + NP2PAL_TEXT3].d;
			for (x=1; x<layer.x; x++) {
				BYTE	c;
				c = s2[x] >> 4;
				if (!c) {
					c = s1[x-1] + NP2PALS_TXT;
				}
				((DWORD *)d)[x] = np2_pal32[c + NP2PAL_TEXT].d;
			}
			((DWORD *)d)[x] = np2_pal32[s1[x-1] + NP2PAL_SKIP].d;
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}

#endif



// ----------------------------------------------------------------- turn left

static void lay32l(void) {

	BYTE	*d = layer.surface;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (r=d, x=layer.x; x--;) {
				*((DWORD *)r) = np2_pal32[NP2PAL_TEXT2].d;
				r += layer.width;
			}
		}
		d += sizeof(DWORD);
	} while(++layer.y < layer.maxy);
	layer.surface = d;
}

static void np2layer32l_1(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((DWORD *)r) = np2_pal32[s[x] + NP2PAL_GRPH].d;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(DWORD);
	} while(++layer.y < layer.maxy);
	layer.surface = d;
	layer.src1scrn = s;
}

static void np2layer32l_tg(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((DWORD *)r) = np2_pal32[s1[x] + s2[x] + NP2PAL_GRPH].d;
				r += layer.width;
			}
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
		d += sizeof(DWORD);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}

static void np2layer32l_t_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((DWORD *)r) = np2_pal32[s[x] + NP2PAL_GRPH].d;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(DWORD);
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((DWORD *)r) = np2_pal32[(s[x] >> 4) + NP2PAL_TEXT].d;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(DWORD);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

static void np2layer32l_g_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((DWORD *)r) = np2_pal32[s[x] + NP2PAL_GRPH].d;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(DWORD);
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((DWORD *)r) = np2_pal32[NP2PAL_TEXT].d;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(DWORD);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

static void np2layer32l_tg_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((DWORD *)r) = np2_pal32[s1[x] + s2[x] + NP2PAL_GRPH].d;
				r += layer.width;
			}
		}
		s2 += SURFACE_WIDTH;
		d += sizeof(DWORD);
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*((DWORD *)r) = np2_pal32[(s2[x] >> 4) + NP2PAL_TEXT].d;
				r += layer.width;
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += sizeof(DWORD);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}



// ---------------------------------------------------------------- turn right

static void lay32r(void) {

	BYTE	*d;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(DWORD));
	do {
		d -= sizeof(DWORD);
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (r=d, x=0; x<layer.x; x++) {
				*((DWORD *)r) = np2_pal32[NP2PAL_TEXT2].d;
				r += layer.width;
			}
		}
	} while(++layer.y < layer.maxy);
}

static void np2layer32r_1(void) {

	BYTE	*d;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(DWORD));
	do {
		d -= sizeof(DWORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((DWORD *)r) = np2_pal32[s[x] + NP2PAL_GRPH].d;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s;
}

static void np2layer32r_tg(void) {

	BYTE	*d;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(DWORD));
	do {
		d -= sizeof(DWORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((DWORD *)r) = np2_pal32[s1[x] + s2[x] + NP2PAL_GRPH].d;
				r += layer.width;
			}
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s1;
	layer.src2scrn = s2;
}

static void np2layer32r_t_skip(void) {

	BYTE	*d;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(DWORD));
	do {
		d -= sizeof(DWORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((DWORD *)r) = np2_pal32[s[x] + NP2PAL_GRPH].d;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		layer.y++;
		d -= sizeof(DWORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((DWORD *)r) = np2_pal32[(s[x] >> 4) + NP2PAL_TEXT].d;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s;
}

static void np2layer32r_g_skip(void) {

	BYTE	*d;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(DWORD));
	do {
		d -= sizeof(DWORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((DWORD *)r) = np2_pal32[s[x] + NP2PAL_GRPH].d;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		layer.y++;
		d -= sizeof(DWORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((DWORD *)r) = np2_pal32[NP2PAL_TEXT].d;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s;
}

static void np2layer32r_tg_skip(void) {

	BYTE	*d;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(DWORD));
	do {
		d -= sizeof(DWORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((DWORD *)r) = np2_pal32[s1[x] + s2[x] + NP2PAL_GRPH].d;
				r += layer.width;
			}
		}
		s2 += SURFACE_WIDTH;
		layer.y++;
		d -= sizeof(DWORD);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*((DWORD *)r) = np2_pal32[(s2[x] >> 4) + NP2PAL_TEXT].d;
				r += layer.width;
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s1;
	layer.src2scrn = s2;
}




void (*layer32tbl[])(void) = {

		// Ρޥ
		lay32,	np2layer32_1,		np2layer32_1,		np2layer32_tg,
		lay32,	np2layer32_t_skip,	np2layer32_g_skip,	np2layer32_tg_skip,
		lay32,	np2layer32_t_skip,	np2layer32_g_ext,	np2layer32_tg_ext,

		lay32n,	np2layer32n_t,		np2layer32n_g,		np2layer32n_tg,
		lay32n,	np2layer32n_t_skip,	np2layer32n_g_skip,	np2layer32n_tg_skip,
		lay32n,	np2layer32n_t_skip,	np2layer32n_g_ext,	np2layer32n_tg_ext,


		// 
		lay32l,	np2layer32l_1,		np2layer32l_1,		np2layer32l_tg,
		lay32l,	np2layer32l_t_skip,	np2layer32l_g_skip,	np2layer32l_tg_skip,
		lay32l,	np2layer32l_t_skip,	np2layer32l_g_skip,	np2layer32l_tg_skip,

		lay32l,	np2layer32l_1,		np2layer32l_1,		np2layer32l_tg,
		lay32l,	np2layer32l_t_skip,	np2layer32l_g_skip,	np2layer32l_tg_skip,
		lay32l,	np2layer32l_t_skip,	np2layer32l_g_skip,	np2layer32l_tg_skip,


		// 
		lay32r,	np2layer32r_1,		np2layer32r_1,		np2layer32r_tg,
		lay32r,	np2layer32r_t_skip,	np2layer32r_g_skip,	np2layer32r_tg_skip,
		lay32r,	np2layer32r_t_skip,	np2layer32r_g_skip,	np2layer32r_tg_skip,

		lay32r,	np2layer32r_1,		np2layer32r_1,		np2layer32r_tg,
		lay32r,	np2layer32r_t_skip,	np2layer32r_g_skip,	np2layer32r_tg_skip,
		lay32r,	np2layer32r_t_skip,	np2layer32r_g_skip,	np2layer32r_tg_skip,
};
