#include	"compiler.h"

#include	"np2.h"
#include	"np2pc98.h"
#include	"pc_core.h"
#include	"palettes.h"
#include	"layer.h"



//	8bitߡ
static void lay8(void) {

	BYTE	*d = layer.surface;
	int		x;

	do {
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (x=0; x<layer.x; x++) {
				d[x] = START_PAL + NP2PAL_TEXT2;
			}
		}
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
}



//	text or grph 1ץ졼 / 8bit
static void np2layer8_1(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				d[x] = s[x] + START_PAL + NP2PAL_GRPH;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph / 8bit
static void np2layer8_tg(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				d[x] = s1[x] + s2[x] + START_PAL + NP2PAL_GRPH;
			}
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


//	text + (grph:skipline) / 8bit
static void np2layer8_t_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				d[x] = s[x] + START_PAL + NP2PAL_GRPH;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				d[x] = (s[x] >> 4) + START_PAL + NP2PAL_TEXT;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	grph:skipline / 8bit
static void np2layer8_g_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				d[x] = s[x] + START_PAL + NP2PAL_GRPH;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				d[x] = START_PAL + NP2PAL_TEXT;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph:skipline / 8bit
static void np2layer8_tg_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				d[x] = s1[x] + s2[x] + START_PAL + NP2PAL_GRPH;
			}
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				d[x] = (s2[x] >> 4) + START_PAL + NP2PAL_TEXT;
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


// --- with extend skipline

//	grph + extend skipline / 8bit
static void np2layer8_g_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			for (x=0; x<layer.x; x++) {
				d[x] = s[x] + START_PAL + NP2PAL_GRPH;
			}
		}
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				d[x] = s[x] + START_PAL + NP2PAL_SKIP;
			}
		}
		s += (SURFACE_WIDTH * 2);
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

//	text + grph + extend skipline / 8bit
static void np2layer8_tg_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			for (x=0; x<layer.x; x++) {
				d[x] = s1[x] + s2[x] + START_PAL + NP2PAL_GRPH;
			}
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				BYTE	c;
				c = s2[x] >> 4;
				if (!c) {
					c = s1[x] + NP2PALS_TXT;
				}
				d[x] = c + START_PAL + NP2PAL_TEXT;
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


// ----------------------------------------- support normal display

#ifdef SUPPORT_NORMALDISP

//	8bitߡ
static void lay8n(void) {

	BYTE	*d = layer.surface;
	int		x;

	do {
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (x=0; x<=layer.x; x++) {
				d[x] = START_PAL + NP2PAL_TEXT2;
			}
		}
		d += layer.width;
	} while(++layer.y < layer.maxy);
	layer.surface = d;
}

//	text / 8bit
static void np2layer8n_t(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (x=0; x<layer.x; x++) {
				d[x] = s[x] + START_PAL + NP2PAL_GRPH;
			}
			d[x] = START_PAL + NP2PAL_GRPH;
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

//	grph / 8bit
static void np2layer8n_g(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			d[0] = START_PAL + NP2PAL_TEXT;
			for (x=0; x<layer.x; x++) {
				d[x+1] = s[x] + START_PAL + NP2PAL_GRPH;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph / 8bit
static void np2layer8n_tg(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			d[0] = (s2[0] >> 4) + START_PAL + NP2PAL_TEXT3;
			for (x=1; x<layer.x; x++) {
				d[x] = s1[x-1] + s2[x] + START_PAL + NP2PAL_GRPH;
			}
			d[x] = s1[x-1] + START_PAL + NP2PAL_GRPH;
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


//	text + (grph:skipline) / 32bit
static void np2layer8n_t_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			d[0] = (s[0] >> 4) + START_PAL + NP2PAL_TEXT3;
			for (x=1; x<layer.x; x++) {
				d[x] = s[x] + START_PAL + NP2PAL_GRPH;
			}
			d[x] = START_PAL + NP2PAL_GRPH;
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			d[0] = (s[0] >> 4) + START_PAL + NP2PAL_TEXT3;
			for (x=1; x<layer.x; x++) {
				d[x] = (s[x] >> 4) + START_PAL + NP2PAL_TEXT;
			}
			d[x] = START_PAL + NP2PAL_TEXT;
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	grph + skipline / 8bit
static void np2layer8n_g_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			d[0] = START_PAL;
			for (x=0; x<layer.x; x++) {
				d[x+1] = s[x] + START_PAL + NP2PAL_GRPH;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			d[0] = START_PAL + NP2PAL_TEXT;
			for (x=0; x<layer.x; x++) {
				d[x+1] = START_PAL + NP2PAL_TEXT;
			}
		}
		s += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}


//	text + grph + skipline / 8bit
static void np2layer8n_tg_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			d[0] = (s2[0] >> 4) + START_PAL + NP2PAL_TEXT3;
			for (x=1; x<layer.x; x++) {
				d[x] = s1[x-1] + s2[x] + START_PAL + NP2PAL_GRPH;
			}
			d[x] = s1[x-1] + START_PAL + NP2PAL_GRPH;
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			d[0] = (s2[0] >> 4) + START_PAL + NP2PAL_TEXT3;
			for (x=1; x<layer.x; x++) {
				d[x] = (s2[x] >> 4) + START_PAL + NP2PAL_TEXT;
			}
			d[x] = START_PAL + NP2PAL_TEXT;
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}


// --- with extend skipline

//	grph + extend skipline / 8bit
static void np2layer8n_g_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			d[0] = START_PAL;
			for (x=0; x<layer.x; x++) {
				d[x+1] = s[x] + START_PAL + NP2PAL_GRPH;
			}
		}
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			d[0] = START_PAL;
			for (x=0; x<layer.x; x++) {
				d[x+1] = s[x] + START_PAL + NP2PAL_SKIP;
			}
		}
		s += (SURFACE_WIDTH * 2);
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

//	text + grph + extend skipline / 8bit
static void np2layer8n_tg_ext(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			renewal_line[layer.y + 1] |= 0x80;				// next line draw
			d[0] = (s2[0] >> 4) + START_PAL + NP2PAL_TEXT3;
			for (x=1; x<layer.x; x++) {
				d[x] = s1[x-1] + s2[x] + START_PAL + NP2PAL_GRPH;
			}
			d[x] = s1[x-1] + START_PAL + NP2PAL_GRPH;
		}
		s2 += SURFACE_WIDTH;
		d += layer.width;
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			d[0] = (s2[0] >> 4) + START_PAL + NP2PAL_TEXT3;
			for (x=1; x<layer.x; x++) {
				BYTE	c;
				c = s2[x] >> 4;
				if (!c) {
					c = s1[x-1] + NP2PALS_TXT;
				}
				d[x] = c + START_PAL + NP2PAL_TEXT;
			}
			d[x] = s1[x-1] + START_PAL + NP2PAL_SKIP;
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += layer.width;
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}

#endif



// ----------------------------------------------------------------- turn left

static void lay8l(void) {

	BYTE	*d = layer.surface;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (r=d, x=layer.x; x--;) {
				*r = START_PAL + NP2PAL_TEXT2;
				r += layer.width;
			}
		}
		d += sizeof(BYTE);
	} while(++layer.y < layer.maxy);
	layer.surface = d;
}


static void np2layer8l_1(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*r = s[x] + START_PAL + NP2PAL_GRPH;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(BYTE);
	} while(++layer.y < layer.maxy);
	layer.surface = d;
	layer.src1scrn = s;
}

static void np2layer8l_tg(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*r = s1[x] + s2[x] + START_PAL + NP2PAL_GRPH;
				r += layer.width;
			}
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
		d += sizeof(BYTE);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}

static void np2layer8l_t_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*r = s[x] + START_PAL + NP2PAL_GRPH;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(BYTE);
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*r = (s[x] >> 4) + START_PAL + NP2PAL_TEXT;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(BYTE);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

static void np2layer8l_g_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*r = s[x] + START_PAL + NP2PAL_GRPH;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(BYTE);
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*r = START_PAL + NP2PAL_TEXT;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		d += sizeof(BYTE);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s;
}

static void np2layer8l_tg_skip(void) {

	BYTE	*d = layer.surface;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	do {
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*r = s1[x] + s2[x] + START_PAL + NP2PAL_GRPH;
				r += layer.width;
			}
		}
		s2 += SURFACE_WIDTH;
		d += sizeof(BYTE);
		layer.y++;
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=layer.x; x--;) {
				*r = (s2[x] >> 4) + START_PAL + NP2PAL_TEXT;
				r += layer.width;
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
		d += sizeof(BYTE);
	} while(++layer.y < layer.maxy);

	layer.surface = d;
	layer.src1scrn = s1;
	layer.src2scrn = s2;
}



// ---------------------------------------------------------------- turn right

static void lay8r(void) {

	BYTE	*d;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(BYTE));
	do {
		d -= sizeof(BYTE);
		if (renewal_line[layer.y] & 0x80) {
			renewal_line[layer.y] &= 0x7f;
			for (r=d, x=0; x<layer.x; x++) {
				*r = START_PAL + NP2PAL_TEXT2;
				r += layer.width;
			}
		}
	} while(++layer.y < layer.maxy);
}


static void np2layer8r_1(void) {

	BYTE	*d;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(BYTE));
	do {
		d -= sizeof(BYTE);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*r = s[x] + START_PAL + NP2PAL_GRPH;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s;
}

static void np2layer8r_tg(void) {

	BYTE	*d;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(BYTE));
	do {
		d -= sizeof(BYTE);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*r = s1[x] + s2[x] + START_PAL + NP2PAL_GRPH;
				r += layer.width;
			}
		}
		s1 += SURFACE_WIDTH;
		s2 += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s1;
	layer.src2scrn = s2;
}

static void np2layer8r_t_skip(void) {

	BYTE	*d;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(BYTE));
	do {
		d -= sizeof(BYTE);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*r = s[x] + START_PAL + NP2PAL_GRPH;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		layer.y++;
		d -= sizeof(BYTE);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*r = (s[x] >> 4) + START_PAL + NP2PAL_TEXT;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s;
}

static void np2layer8r_g_skip(void) {

	BYTE	*d;
	BYTE	*s = layer.src1scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(BYTE));
	do {
		d -= sizeof(BYTE);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*r = s[x] + START_PAL + NP2PAL_GRPH;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
		layer.y++;
		d -= sizeof(BYTE);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*r = START_PAL + NP2PAL_TEXT;
				r += layer.width;
			}
		}
		s += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s;
}

static void np2layer8r_tg_skip(void) {

	BYTE	*d;
	BYTE	*s1 = layer.src1scrn;
	BYTE	*s2 = layer.src2scrn;
	int		x;
	BYTE	*r;

	d = layer.surface + ((layer.maxy - layer.y) * sizeof(BYTE));
	do {
		d -= sizeof(BYTE);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*r = s1[x] + s2[x] + START_PAL + NP2PAL_GRPH;
				r += layer.width;
			}
		}
		s2 += SURFACE_WIDTH;
		layer.y++;
		d -= sizeof(BYTE);
		if (renewal_line[layer.y] & layer.bit) {
			renewal_line[layer.y] &= ~layer.bit;
			for (r=d, x=0; x<layer.x; x++) {
				*r = (s2[x] >> 4) + START_PAL + NP2PAL_TEXT;
				r += layer.width;
			}
		}
		s1 += (SURFACE_WIDTH * 2);
		s2 += SURFACE_WIDTH;
	} while(++layer.y < layer.maxy);

	layer.src1scrn = s1;
	layer.src2scrn = s2;
}





void (*layer8tbl[])(void) = {

		// Ρޥ
		lay8,	np2layer8_1,		np2layer8_1,		np2layer8_tg,
		lay8,	np2layer8_t_skip,	np2layer8_g_skip,	np2layer8_tg_skip,
		lay8,	np2layer8_t_skip,	np2layer8_g_ext,	np2layer8_tg_ext,

		lay8n,	np2layer8n_t,		np2layer8n_g,		np2layer8n_tg,
		lay8n,	np2layer8n_t_skip,	np2layer8n_g_skip,	np2layer8n_tg_skip,
		lay8n,	np2layer8n_t_skip,	np2layer8n_g_ext,	np2layer8n_tg_ext,


		// 
		lay8l,	np2layer8l_1,		np2layer8l_1,		np2layer8l_tg,
		lay8l,	np2layer8l_t_skip,	np2layer8l_g_skip,	np2layer8l_tg_skip,
		lay8l,	np2layer8l_t_skip,	np2layer8l_g_skip,	np2layer8l_tg_skip,

		lay8l,	np2layer8l_1,		np2layer8l_1,		np2layer8l_tg,
		lay8l,	np2layer8l_t_skip,	np2layer8l_g_skip,	np2layer8l_tg_skip,
		lay8l,	np2layer8l_t_skip,	np2layer8l_g_skip,	np2layer8l_tg_skip,


		// 
		lay8r,	np2layer8r_1,		np2layer8r_1,		np2layer8r_tg,
		lay8r,	np2layer8r_t_skip,	np2layer8r_g_skip,	np2layer8r_tg_skip,
		lay8r,	np2layer8r_t_skip,	np2layer8r_g_skip,	np2layer8r_tg_skip,

		lay8r,	np2layer8r_1,		np2layer8r_1,		np2layer8r_tg,
		lay8r,	np2layer8r_t_skip,	np2layer8r_g_skip,	np2layer8r_tg_skip,
		lay8r,	np2layer8r_t_skip,	np2layer8r_g_skip,	np2layer8r_tg_skip,
};
