#include	"compiler.h"

#include	"np2.h"
#include	"np2pc98.h"
#include	"pc_core.h"
#include	"i286.h"
#include	"memory.h"
#include	"palettes.h"
#include	"nevent.h"
#include	"iobridge.h"
#include	"pic.h"
#include	"gdc.h"
#include	"gdc_cmd.h"
#include	"gdc_sub.h"
#include	"crtc.h"
#include	"sndboard.h"
#include	"np2sound.h"
#include	"beep.h"
#include	"mouseif.h"
#include	"cgrom.h"
#include	"bios.h"
#include	"maketext.h"
#include	"makegrph.h"
#include	"dosio.h"
#include	"serial.h"
#include	"itimer.h"
#include	"sysport.h"
#include	"mpu98ii.h"
#include	"fddfile.h"
#include	"printif.h"
#include	"fdd_mtr.h"
#include	"dmac.h"
#include	"fdc.h"
#include	"calender.h"
#include	"font.h"
#include	"egc.h"
#include	"vram.h"
#include	"ini.h"
#include	"cpuio.h"
#include	"nmiio.h"
#include	"np2sysp.h"
#include	"dispsync.h"
#include	"extmem.h"
#include	"layer.h"
#include	"disk.h"
#include	"artic.h"												// ver0.28

#include	"trace.h"

	PCCORE_TABLE	pc = {	PCBASECLOCK25,
							4,
							4 * PCBASECLOCK25,
							4 * PCBASECLOCK25 * 50 / 3104,
							4 * PCBASECLOCK25 * 5 / 3104,
							4 * PCBASECLOCK25 / 120,
							4 * PCBASECLOCK25 / 200,
							4 * PCBASECLOCK25 / 100,
							4 * PCBASECLOCK25 / 1920,
							4 * PCBASECLOCK25 / 3125,
							(4 * PCBASECLOCK25 / 22050) + 1,
							(4 * PCBASECLOCK25 / 56400),
							100, 20,
							0};

									// on=0, off=1
	BYTE	dip_default[3] = {0x3e, 0x63, 0x7a};
	BYTE	msw_default[8] = {0x48, 0x05, 0x04, 0x00, 0x01, 0x00, 0x00, 0x6E};

	BYTE	screenupdate = 3;
	int		screendispflag = 1;
	DWORD	np2drawtime = 0;

	BOOL	np2scrnoutflg = 0;

// ---------------------------------------------------------------------------

static void setvsyncclock(void) {								// ver0.28

	DWORD	lf, vbp, vfp, vs;
	DWORD	disp, max;
	DWORD	cnt;

	vfp = gdc.m.para[GDC_SYNC + 5] & 0x3f;
	if (!vfp) {
		vfp = 1;
	}
	vbp = gdc.m.para[GDC_SYNC + 7] >> 2;
	if (!vbp) {
		vbp = 1;
	}
	lf = (GETWORD(gdc.m.para + GDC_SYNC + 6)) & 0x3ff;
	if (!lf) {
		lf = 1024;
	}
	disp = vfp + vbp + lf;
	vs = ((GETWORD(gdc.m.para + GDC_SYNC + 4)) >> 5) & 0x1f;
	if (!vs) {
		vs = 1;
	}
	max = disp + vs;
	cnt = (pc.realclock * 5) / 282;
	pc.raster = cnt / max;
	pc.hsync = (pc.raster * 4) / 5;
	pc.dispclock = pc.raster * disp;
	pc.vsyncclock = cnt - pc.dispclock;
}

static void setpcclock(DWORD base, DWORD multiple) {			// ver0.28

	if (base >= ((PCBASECLOCK25 + PCBASECLOCK20) / 2)) {
		pc.baseclock = PCBASECLOCK25;			// 2.5MHz
		pc.cpumode = 0;
	}
	else {
		pc.baseclock = PCBASECLOCK20;			// 2.0MHz
		pc.cpumode = CPUMODE_8MHz;
	}
	if (multiple == 0) {
		multiple = 1;
	}
	else if (multiple > 32) {
		multiple = 32;
	}
	pc.multiple = multiple;
	pc.realclock = pc.baseclock * multiple;
	pc.raster = pc.realclock / 24816;							// ver0.28
	pc.hsync = (pc.raster * 4) / 5;								// ver0.28
	pc.dispclock = pc.realclock * 50 / 3102;
	pc.vsyncclock = pc.realclock * 5 / 3102;
	pc.mouseclock = pc.realclock / 120;
	pc.dsoundclock = (pc.realclock + 100) / 200;				// ver0.28
	pc.dsoundclock2 = pc.realclock / 100;
	pc.keyboardclock = pc.realclock / 1920;
	pc.midiclock = pc.realclock / 3125;
	pc.frame1000 = pc.realclock / 56400;
	if (np2sound.rate) {
		pc.sampleclock = (pc.realclock / np2sound.rate) + 1;
	}
	else {
		pc.sampleclock = 0;
	}
}

// --------------------------------------------------------------------------

extern void _np2_init(void);

void np2_init(void) {

	_np2_init();
	layer_init();

	cpu_init();
	font_init();
	font_load(fontfile, TRUE);
	maketext_init();
	makegrph_init();
	gdcsub_init();
	bios0x1b_init();
	fddfile_init();
	disk_init();

	np2sysp_init();
	TRACE_START();
}

void np2_term(void) {

	TRACE_END();

	fdd_eject(0);
	fdd_eject(1);
	fdd_eject(2);
	fdd_eject(3);

	extmemmng_clear();											// ver0.28
	bios0x1b_allclose();
}


BOOL np2_getmemsw(void) {

	int		i;
	BOOL	ret = FALSE;

	for (i=0; i<8; i++) {
		if (np2cfg.memsw[i] != mem[0xa3fe2 + i*4]) {
			np2cfg.memsw[i] = mem[0xa3fe2 + i*4];
			ret = TRUE;
		}
	}
	return(ret);
}

void np2_reset(void) {

	int		i;

	dispsync_reset();											// ver0.26

#if 1															// ver0.28
	ZeroMemory(mem, 0x10fff0);
	ZeroMemory(mem + VRAM1_B, 0x18000);
	ZeroMemory(mem + VRAM1_E, 0x08000);
	ZeroMemory(mem + FONT_ADRS, 0x08000);
#else
	ZeroMemory(mem, sizeof(mem));
#endif
	cpu_reset();
	CPUTYPE = 0;
	if (np2cfg.dipsw[2] & 0x80) {
		CPUTYPE = CPUTYPE_V30;
	}

	//ꥹå
	for (i=0; i<8; i++) {
		mem[0xa3fe2 + i*4] = np2cfg.memsw[i];
	}

	fddfile_reset2dmode();
	bios0x18_16(0x20, 0xe1);


	sndboard_reset(np2cfg.SOUND_SW);

	setpcclock(np2cfg.baseclock, np2cfg.multiple);
	nevent_init();
	cgrom_init();
	gdc_init();
	itimer_init();
	pic_init();
	crtc_init();
	mouseif_init();
	keyboard_init();
	systemport_init();
	mpu98ii_init();
	printif_init();
	fddmtr_init();
	fdc_init();
	dmac_init();
	calender_init();
	egc_init();
	vram_init();
	cpuio_init();
	nmiio_init();
	rs232c_init();
	pal_change(1);

	bios_init();
	bios0x1b_hddreopen();

	np2stat_getstatus(NULL, NP2STAT_NONE, NP2STAT_VENEVT);

	sound_reset();
	sound_make();

	if (np2cfg.ITF_WORK) {
		CS_BASE = 0xf0000;
		I286_CS = 0xf000;
		I286_IP = 0xfff0;
	}
	else {
		for (i=0; i<8; i++) {
			MEM_SETB(0xa3fe2 + i*4, msw_default[i]);
		}
		CS_BASE = 0xfd800;
		I286_CS = 0xfd80;
		I286_IP = 0x0002;
	}
}


// ---------------------------------------------------------------------------

static void drawscreen(void) {

extern	WORD	framecnt;
		BYTE	timing;
		BYTE	flag;

	tramflag.timing++;
	timing = ((GETWORD(gdc.m.para + GDC_CSRFORM + 1)) >> 5) & 0x3e;
	if (!timing) {
		timing = 0x40;
	}
	if (tramflag.timing >= timing) {
		tramflag.timing = 0;
		tramflag.count++;
		tramflag.renewal |= (tramflag.count ^ 2) & 2;
		tramflag.renewal |= 1;
	}

	if ((gdcs.textdisp & GDCSCRN_EXT) ||						// ver0.28
		(gdcs.grphdisp & GDCSCRN_EXT)) {
		setvsyncclock();
	}

	if (np2scrnoutflg) {
		if ((gdcs.textdisp & GDCSCRN_EXT) ||					// ver0.26
			(gdcs.grphdisp & GDCSCRN_EXT)) {
			if (dispsync_set()) {
				gdcs.textdisp |= GDCSCRN_ALLDRAW2;
				gdcs.grphdisp |= GDCSCRN_ALLDRAW2;
				screenupdate |= 2;
			}
		}

		if (gdcs.textdisp & GDCSCRN_EXT) {
			gdcs.textdisp &= ~GDCSCRN_EXT;
			tramflag.renewal |= 1;
			flag = 0;
			if ((!(np2cfg.LCD_MODE & 1)) && (gdc.display < 3)) {
				flag |= 1;
			}
			if (np2os_setextflag(flag)) {
				screenupdate |= 2;
			}
		}
		if (gdcs.textdisp & GDCSCRN_ENABLE) {
			if (tramflag.renewal) {
				gdcs.textdisp |= maketext_curblink();
			}
			if ((cgwindow.writable & 0x80) && (tramflag.gaiji)) {
				gdcs.textdisp |= GDCSCRN_ALLDRAW;
			}
			cgwindow.writable &= ~0x80;
			if (gdcs.textdisp & GDCSCRN_MAKE) {
				if (!(gdc.mode1 & 0x4)) {
					maketext(gdcs.textdisp & GDCSCRN_ALLDRAW);
				}
				else {
					maketext40(gdcs.textdisp & GDCSCRN_ALLDRAW);
				}
				gdcs.textdisp &= ~GDCSCRN_MAKE;
				screenupdate |= 1;
			}
		}
		if (gdcs.palchange) {
			gdcs.palchange = 0;
			pal_change(0);
			screenupdate |= 1;
		}
		if (gdcs.grphdisp & GDCSCRN_EXT) {
			gdcs.grphdisp &= ~GDCSCRN_EXT;
			if (((gdc.clock & 0x80) && (gdc.clock != 0x83)) ||
				(gdc.clock == 0x03)) {
				gdc.clock ^= 0x80;
				gdcs.grphdisp |= GDCSCRN_ALLDRAW2;
			}
		}
		if (gdcs.grphdisp & GDCSCRN_ENABLE) {
			if (!gdcs.disp) {
				if (gdcs.grphdisp & GDCSCRN_MAKE) {
					makegrph(0, gdcs.grphdisp & GDCSCRN_ALLDRAW);
					gdcs.grphdisp &= ~GDCSCRN_MAKE;
					screenupdate |= 1;
				}
			}
			else {
				if (gdcs.grphdisp & (GDCSCRN_MAKE << 1)) {
					makegrph(1, gdcs.grphdisp & (GDCSCRN_ALLDRAW << 1));
					gdcs.grphdisp &= ~(GDCSCRN_MAKE << 1);
					screenupdate |= 1;
				}
			}
		}
		if (screenupdate) {
			np2drawtime++;
		}
		screenupdate = np2os_drawing(screenupdate);
	}
}


NEVENT_PROC screendisp(NEVENT_ITEM * item) {

	gdc_work(GDCWORK_SLAVE);
	gdc.vsync = 0;
	screendispflag = 0;
	if (!np2cfg.DISPSYNC) {
		drawscreen();
	}
	if (pic[0].irr & PIC_CRTV) {
		pic[0].irr &= ~PIC_CRTV;
		gdc.vsyncint = 1;
	}
}

NEVENT_PROC screenvsync(NEVENT_ITEM * item) {

	vramop.tramwait = np2cfg.wait[1];
	vramop.vramwait = np2cfg.wait[3];
	vramop.grcgwait = np2cfg.wait[5];
	gdc_work(GDCWORK_MASTER);
	gdc.vsync = 0x20;
	if (gdc.vsyncint) {
		gdc.vsyncint = 0;
		pic_setirq(2);
	}
	nevent_set(NEVENT_FLAMES, pc.vsyncclock, screendisp, NEVENT_RELATIVE);

	// drawscreen pc.vsyncclockѹǽޤ		// ver0.28
	if (np2cfg.DISPSYNC) {
		drawscreen();
	}
}

// ---------------------------------------------------------------------------

void np2main_exec(BOOL screenupdate) {

	sound_make();													// happy!

	np2scrnoutflg = screenupdate;
	pal_eventclear();
	disk_callbackfdd();
	mousekey_make();												// ver0.28
	mouseif_sync();													// ver0.28
	gdc.vsync = 0;
	screendispflag = 1;
	vramop.tramwait = np2cfg.wait[0];
	vramop.vramwait = np2cfg.wait[2];
	vramop.grcgwait = np2cfg.wait[4];
	nevent_set(NEVENT_FLAMES, pc.dispclock, screenvsync, NEVENT_RELATIVE);

//	nevent_get1stevent();

	while(screendispflag) {
		pic_irq();
		if (cpuio.reset_req) {
			cpuio.reset_req = 0;
			I286_CS = 0xf000;
			CS_BASE = 0xf0000;
			I286_IP = 0xfff0;
#ifdef CPU386
			I286_DX = 0x0300;
#endif
		}
#ifndef TRACE
		if (nevent.remainclock >= 0) {
			if (!I286_TRAP) {
				if (!dma.working) {
					CPU_EXEC();
				}
				else {
					CPU_WITHDMA();
				}
			}
			else {
				CPU_WITHTRAP();
			}
		}
#else
		while(nevent.remainclock >= 0) {
#if 1
			CPU_STEP();
#else
extern void i286c_step(void);
			BYTE	code;
			BYTE	dispatch;
//			TRACE1();
			code = i286_memoryread(CS_BASE + I286_IP);

			if ((code == 0x8e) || (code == 0x17) ||
				(code == 0xfb) ||
				((code & (~0x18)) == 0x26) ||
				((code & 0xfe) == 0xf2) ||
				(code >= 0xf7))
			{
				dispatch = 0;
			}
			else {
				dispatch = 1;
			}
#if 1
			if (code == 0xf7) {
				BYTE c2 = i286_memoryread(CS_BASE + ((I286_IP + 1) & 0xffff));
				if (((c2 >> 3) & 7) < 4) {
					dispatch = 1;
				}
			}
#endif
			if (!dispatch) {
				i286_step();
			}
			else {
				i286c_step();
			}
#endif
		}
#endif
		nevent_progress();
	}
	artic_callback();												// ver0.28
	calender_inc();
	sound_make();													// happy!
}

