#include	"compiler.h"

#include	"np2.h"
#include	"np2sound.h"
#include	"fm.h"
#include	"adpcm.h"

#define		ADTIMING_BIT	11
#define		ADTIMING		(1 << ADTIMING_BIT)
#define		ADPCM_SHIFT		3

static	DWORD	adpcmdeltatable[16] = {
		//	0.89,	0.89,	0.89,	0.89,	1.2,	1.6,	2.0,	2.4
#if 0
			1823,	1823,	1823,	1823,	2458,	3277,	4096,	4916,
			1823,	1823,	1823,	1823,	2458,	3277,	4096,	4916,
#endif
#if 0
			228,	228,	228,	228,	307,	409,	512,	614,
			228,	228,	228,	228,	307,	409,	512,	614,
#endif
#if 1
			228,	228,	228,	228,	308,	408,	512,	612,
			228,	228,	228,	228,	308,	408,	512,	612,
#endif
			};

	ADPCM_T		adpcm;

void adpcm_reset(void) {

	ZeroMemory(&adpcm, sizeof(adpcm));

	adpcm.mask = 0;					// (BYTE)~0x1c;
	adpcm.delta = 128;
	adpcm.reg.stop = 0x0002;
	adpcm.reg.limit = 0xffff;
	adpcm.stop = 0x000060;
	adpcm.limit = 0x200000;

	adpcm.base = ADTIMING * (OPNA_CLOCK/72) / np2sound.rate;
	adpcm.step = adpcm.base >> 8;
	adpcm.pertim = (1 << (ADTIMING_BIT*2)) / adpcm.step;
}


void adpcm_volset(void) {									// ver0.28

	adpcm.level = ((DWORD)adpcm.reg.level * np2cfg.vol_adpcm) >> 4;
}


LABEL static void __fastcall adpcm_datawrite(BYTE data) {

	__asm {
					push	ebx
					push	ecx
					mov		edx, adpcm.pos
					and		edx, 1fffffh
					mov		ebx, edx
					shr		ebx, 3
					test	adpcm.reg.ctrl2, 2
					jne		adpcm_bitsepmode
					and		ebx, 3ffffh
					mov		adpcm.buf[ebx], cl
					add		edx, 8
					jmp		adpcm_isposstop
adpcm_bitsepmode:	and		ebx, 07fffh
					add		ebx, offset (adpcm.buf[0])
					mov		ah, cl
					mov		ecx, edx
					and		ecx, 7
					mov		al, 1
					shl		al, cl
					rol		ah, cl
					not		ah
					mov		cl, 8
adpcm_bytelp:		or		[ebx], al
					mov		ch, al
					and		ch, ah
					xor		[ebx], ch
					ror		ah, 1
					add		ebx, 8000h
					dec		cl
					jne		adpcm_bytelp
					inc		edx
adpcm_isposstop:	cmp		edx, adpcm.stop
					jne		adpcm_isposlimit
					or		adpcm.status, 4
					and		edx, 1fffffh
adpcm_isposlimit:	cmp		edx, adpcm.limit
					sbb		ebx, ebx
					and		edx, ebx
					mov		adpcm.pos, edx
					pop		ecx
					pop		ebx
					ret
		}
}


void adpcm_setregister(BYTE reg, BYTE value) {

	*((BYTE *)(&adpcm.reg) + reg) = value;
	switch(reg) {
		case 0x00:								// control1
			if ((value & 0x80) && (!adpcm.play)) {
				adpcm.play = 0x20;
				adpcm.pos = adpcm.start;
				adpcm.samp = 0;
				adpcm.delta = 128;
				adpcm.remain = 0;
			}
			if (value & 1) {
				adpcm.play = 0x00;
			}
			break;

		case 0x01:								// control2
			break;

		case 0x02:	case 0x03:					// start address
			adpcm.pos = adpcm.start = (DWORD)adpcm.reg.start << 5;
			break;

		case 0x04:	case 0x05:					// stop address
			adpcm.stop = ((DWORD)adpcm.reg.stop + 1) << 5;
			break;

		case 0x08:								// data
			if ((adpcm.reg.ctrl1 & 0x60) == 0x60) {
				adpcm_datawrite(value);
			}
			break;

		case 0x09:	case 0x0a:					// delta
			adpcm.step = (adpcm.reg.delta * adpcm.base) >> 16;
			if (adpcm.step < 0x80) {
				adpcm.step = 0x80;
			}
			adpcm.pertim = (1 << (ADTIMING_BIT*2)) / adpcm.step;
			break;

		case 0x0b:								// level
			adpcm.level = ((DWORD)value * np2cfg.vol_adpcm) >> 4;	// ver0.28
			break;

		case 0x0c:	case 0x0d:					// limit address
			adpcm.limit = ((DWORD)adpcm.reg.limit + 1) << 5;
			break;

		case 0x10:								// flag
			if (value & 0x80) {
				adpcm.status = 0;
			}
			else {
				adpcm.mask = ~(value & 0x1f);
			}
			break;
	}
}

BYTE adpcm_status(void) {

	return(((adpcm.status | 8) & adpcm.mask) | adpcm.play);
}


LABEL BYTE adpcm_readsample(void) {

		__asm {
					mov		al, adpcm.reg.ctrl1
					and		al, 60h
					cmp		al, 20h
					mov		al, 0
					jne		datafifoin

					push	ebx
					mov		edx, adpcm.pos
					and		edx, 1fffffh
					mov		ebx, edx
					shr		ebx, 3
					test	adpcm.reg.ctrl2, 2
					jne		adpcm_bytemode
					add		edx, 8
					and		ebx, 3ffffh
					mov		al, adpcm.buf[ebx]
					jmp		adpcm_isposstop
adpcm_bytemode:		and		ebx, 07ffeh
					mov		ecx, edx
					inc		edx
					and		ecx, 15
					bt		word ptr (adpcm.buf[ebx + 38000h]), cx
					setc	al
					bt		word ptr (adpcm.buf[ebx + 30000h]), cx
					rcl		al, 1
					bt		word ptr (adpcm.buf[ebx + 28000h]), cx
					rcl		al, 1
					bt		word ptr (adpcm.buf[ebx + 20000h]), cx
					rcl		al, 1
					bt		word ptr (adpcm.buf[ebx + 18000h]), cx
					rcl		al, 1
					bt		word ptr (adpcm.buf[ebx + 10000h]), cx
					rcl		al, 1
					bt		word ptr (adpcm.buf[ebx + 08000h]), cx
					rcl		al, 1
					bt		word ptr (adpcm.buf[ebx + 00000h]), cx
					rcl		al, 1
adpcm_isposstop:	cmp		edx, adpcm.stop
					jne		adpcm_isposlimit
					or		adpcm.status, 4
					and		edx, 1fffffh
adpcm_isposlimit:	cmp		edx, adpcm.limit
					sbb		ebx, ebx
					and		edx, ebx
					mov		adpcm.pos, edx
					pop		ebx
datafifoin:			movzx	edx, adpcm.fifopos
					xor		adpcm.fifopos, 1
					xchg	al, adpcm.fifo[edx]
					ret
		}
}

// -------------------------------------------------------------------------

LABEL static void getadpcmdata(void) {

		__asm {
					push	ecx
					mov		ebx, adpcm.pos
					mov		edx, ebx
					shr		edx, 3
					test	adpcm.reg.ctrl2, 2
					jne		short adpcm_8sep
					and		edx, 3ffffh
					movzx	ecx, byte ptr adpcm.buf[edx]
					add		ebx, 80000004h
					js		short bit1hbs
					and		ecx, 0fh
					jmp		short adpcm_decode
bit1hbs:			shr		ecx, 4
					jmp		short adpcm_decode
adpcm_8sep:			xor		ecx, ecx
					and		edx, 7ffch
					mov		eax, ebx
					and		eax, 31
					add		ebx, 80000000h
					js		short bit8hbs
					inc		ebx
					bt		dword ptr adpcm.buf[edx + 18000h], ax
					rcl		ecx, 1
					bt		dword ptr adpcm.buf[edx + 10000h], ax
					rcl		ecx, 1
					bt		dword ptr adpcm.buf[edx + 08000h], ax
					rcl		ecx, 1
					bt		dword ptr adpcm.buf[edx + 00000h], ax
					rcl		ecx, 1
					jmp		short adpcm_decode
bit8hbs:			bt		dword ptr adpcm.buf[edx + 38000h], ax
					rcl		ecx, 1
					bt		dword ptr adpcm.buf[edx + 30000h], ax
					rcl		ecx, 1
					bt		dword ptr adpcm.buf[edx + 28000h], ax
					rcl		ecx, 1
					bt		dword ptr adpcm.buf[edx + 20000h], ax
					rcl		ecx, 1
adpcm_decode:		mov		eax, adpcmdeltatable[ecx*4]
					mul		adpcm.delta
					sub		eax, 12				// add 1024
					shr		eax, 8
					cmp		eax, 126			// 128
					ja		short deltabolow
					mov		eax, 126			// 128
					jmp		short deltaout
deltabolow:			cmp		eax, 24000			// 28672
					jb		short deltaout
					mov		eax, 24000			// 28672
deltaout:			xchg	adpcm.delta, eax
					btr		ecx, 3
					jc		short adpcmvalc
					add		ecx, ecx
					inc		ecx
					mul		ecx
					shr		eax, ADPCM_SHIFT
					add		eax, adpcm.samp
					cmp		eax, 000008000h
					jl		short adpcmvalout
					mov		eax, 000007fffh
					jmp		short adpcmvalout
adpcmvalc:			adc		ecx, ecx
					mul		ecx
					shr		eax, ADPCM_SHIFT
					neg		eax
					add		eax, adpcm.samp
					cmp		eax, 0ffff8000h
					jg		short adpcmvalout
					mov		eax, 0ffff8001h
adpcmvalout:		mov		adpcm.samp, eax
					cmp		ebx, 0
					js		short adpcmrestorepos
					cmp		ebx, adpcm.stop
					jne		short nonstop
					test	adpcm.reg.ctrl1, 10h
					je		short adpcmbreak
					mov		ebx, adpcm.start
					mov		adpcm.samp, 0
					mov		adpcm.delta, 128
					jmp		short adpcmrestorepos
adpcmbreak:			and		ebx, 1fffffh
					or		adpcm.status, 4
					mov		adpcm.play, 0
					jmp		short adpcmrestorepos
nonstop:			cmp		ebx, adpcm.limit
					sbb		edx, edx
					and		ebx, edx
adpcmrestorepos:	mov		adpcm.pos, ebx
					pop		ecx
					imul	adpcm.level
					sar		eax, (10+1)
					mov		ebx, adpcm.out1
					mov		adpcm.out0, ebx
					mov		edx, eax
					sar		edx, 1
					xchg	edx, adpcm.fb
					add		edx, eax
					mov		adpcm.out1, edx
					ret
			}
}

LABEL void __fastcall adpcm_update(void *buf, int buflength) {

		__asm {
					cmp		adpcm.play, 0
					je		adpcmnomake
					cmp		edx, 0
					je		adpcmnomake
					pushad
					mov		ebp, adpcm.remain
					mov		edi, adpcm.step
					cmp		edi, ADTIMING
					ja		makeadpcm3
					mov		ebx, adpcm.out0
					mov		edx, adpcm.out1
makeadpcm1_lp:		cmp		ebp, 0
					jns		makeadpcm1_calc
					add		ebp, ADTIMING
					call	getadpcmdata
					cmp		adpcm.play, 0
					je		makeadpcm2_lp
makeadpcm1_calc:	mov		esi, ebx
					imul	esi, ebp
					mov		eax, ADTIMING
					sub		eax, ebp
					imul	eax, edx
					add		eax, esi
					sar		eax, ADTIMING_BIT
					test	adpcm.reg.ctrl2, 80h
					je		adpcmout_l1
					add		[ecx], eax
adpcmout_l1:		test	adpcm.reg.ctrl2, 40h
					je		adpcmout_r1
					add		[ecx+4], eax
adpcmout_r1:		add		ecx, 8
					sub		ebp, adpcm.step
					dec		dword ptr [esp+20]				// (edx)
					jne		makeadpcm1_lp
					jmp		adpcm_makeend
makeadpcm2_lp:		cmp		ebp, 0
					js		makeadpcm_stop
					mov		eax, ebp
					imul	ebx
					sar		eax, ADTIMING_BIT
					test	adpcm.reg.ctrl2, 80h
					je		adpcmout_l2
					add		[ecx], eax
adpcmout_l2:		test	adpcm.reg.ctrl2, 40h
					je		adpcmout_r2
					add		[ecx+4], eax
adpcmout_r2:		add		ecx, 8
					sub		ebp, adpcm.step
					dec		dword ptr [esp+20]				// (edx)
					jne		makeadpcm2_lp
makeadpcm_stop:		xor		eax, eax
					mov		adpcm.out0, eax
					mov		adpcm.out1, eax
					mov		adpcm.fb, eax
					xor		ebp, ebp
					jmp		adpcm_makeend
makeadpcm3:			mov		esi, adpcm.pertim
makeadpcm3_lp:		mov		eax, ADTIMING
					sub		eax, ebp
					mov		ebx, adpcm.out0
					imul	ebx
					cmp		ebp, 0
					jle		makeadpcm3_calced
makeadpcm3_calclp:	push	eax
					call	getadpcmdata
					pop		eax
					cmp		adpcm.play, 0
					je		short makeadpcm_stop
					mov		edx, ebp
					cmp		edx, esi
					jb		makeadpcm3_less
					mov		edx, esi
makeadpcm3_less:	imul	ebx, edx
					add		eax, ebx
					sub		ebp, esi
					jg		makeadpcm3_calclp
makeadpcm3_calced:	add		ebp, ADTIMING
					sar		eax, ADTIMING_BIT
					test	adpcm.reg.ctrl2, 80h
					je		adpcmout_l3
					add		[ecx], eax
adpcmout_l3:		test	adpcm.reg.ctrl2, 40h
					je		adpcmout_r3
					add		[ecx+4], eax
adpcmout_r3:		add		ecx, 8
					dec		dword ptr [esp+20]				// (edx)
					jne		makeadpcm3_lp
adpcm_makeend:		mov		adpcm.remain, ebp
					popad
adpcmnomake:		ret
			}
}
