#include	"compiler.h"

#include	"pc_core.h"
#include	"iobridge.h"
#include	"dosio.h"
#include	"sndboard.h"
#include	"np2sound.h"
#include	"cs4231.h"
#include	"dmac.h"
#include	"pic.h"
#include	"nevent.h"
#include	"memory.h"

	CS4231_TABLE	cs4231;

static WORD samprate[] = {
					 8000,	 5510,	16000,	11025,
					27420,	18900,	32000,	22050,
					54840,	37800,	64000,	44100,
					48000,	33075,	 9600,	 6620};

static BYTE dmach[] =  {0xff, 0x00, 0x01, 0x03, 0xff, 0xff, 0xff, 0xff};
static BYTE dmairq[] = {0xff, 0x03, 0x06, 0x0a, 0x0c, 0xff, 0xff, 0xff};

extern BYTE (SOUND_CALL * cs4231enc[])(void);

void cs4231_dma(NEVENT_ITEM *item) {

	BYTE	ret;
	DWORD	cnt;

	if (item->flag & NEVENT_SETEVENT) {
		if (dma.dmach[0].leng.w) {
			sound_make();
			ret = cs4231.proc();
			if ((ret) && (cs4231.reg.pinctrl & 2)) {
				dma.dmach[0].leng.w = 0;
				cs4231.intflag = 1;
				pic_setirq(0x0c);
			}
		}
		cnt = pc.realclock * 16 / np2sound.rate;
		nevent_set(NEVENT_CS4231, cnt, cs4231_dma, NEVENT_RELATIVE);
	}
}


DMAC_EXT cs4231dmafunc(BYTE func) {

	DWORD	cnt;

	switch(func) {
		case DMAEXT_START:
			cnt = pc.realclock * 16 / np2sound.rate;
			nevent_set(NEVENT_CS4231, cnt, cs4231_dma, NEVENT_ABSOLUTE);
			break;

		case DMAEXT_BREAK:
			nevent_reset(NEVENT_CS4231);
			break;
	}
	return(0);
}


static BYTE SOUND_CALL cs4231nomake(void) {

	return(0);
}

void cs4231_disable(void) {

	ZeroMemory(&cs4231, sizeof(cs4231));
	cs4231.proc = cs4231nomake;
	cs4231.port = 0xffff;
	cs4231.dmach = 0xff;
	cs4231.dmairq = 0xff;
}

void cs4231_reset(void) {

	cs4231.enable = 1;
	cs4231.port = 0xf40;
	cs4231.adrs = 0x21;
	cs4231.dmach = 0;
	cs4231.dmairq = 0x0c;
	cs4231.step = 22050;
}

static void cs4231control(int index, BYTE value) {

	*(((BYTE *)(&cs4231.reg)) + index) = value;
	switch(index) {
		case 8:					// playback data format
			cs4231.proc = cs4231enc[value >> 4];
			cs4231.step = samprate[value & 0x0f];
			break;

		case 9:
			if ((value & 0x5) == 0x5) {
				dma.dmach[0].ready = 1;
			}
			else {
				dma.dmach[0].ready = 0;
			}
			dmac_check();
			break;
	}
}



I286_OUT cs4231p_w(WORD port, BYTE value) {

	if (cs4231.enable) {
		switch(port & 0x0f) {
			case 0x0004:
				cs4231.portctrl = value;
				break;
			case 0x000B:
				if ((cs4231.portctrl & 0x3f) == 0x30) {
					*(BYTE *)&cs4231.port = value & 0xf0;
				}
				break;
			case 0x000D:
				if ((cs4231.portctrl & 0x3f) == 0x30) {
					*(((BYTE *)&cs4231.port) + 1) = value;
				}
				break;
		}
	}
}

I286_INB cs4231p_r(WORD port) {

	if (cs4231.enable) {
		switch(port & 0x0f) {
			case 0x0004:
				return(0x80 | cs4231.portctrl);
				break;
			case 0x000B:
				switch(cs4231.portctrl & 0x1f) {
					case 0x10:
						return(cs4231.port & 0xff);
					case 0x11:
						return(0x60);
					case 0x14:
						return(0x88);
				}
				break;
			case 0x000D:
				switch(cs4231.portctrl & 0x1f) {
					case 0x10:
						return(cs4231.port >> 8);
					case 0x11:
						return(0xa4);
					case 0x14:
						return(0x01);
				}
				break;
		}
	}
	return(0xff);
}

I286_OUT cs4231_w(WORD port, BYTE value) {

	switch(port & 0x0f) {
		case 0:
#if 0
			cs4231.adrs = value;
			cs4231.dmairq = dmairq[(value >> 3) & 3];
			cs4231.dmach = dmach[value & 7];
#endif
			break;

		case 4:
			cs4231.index = value;
			break;

		case 5:
			cs4231control(cs4231.index & 0x1f, value);
			break;

		case 6:
			cs4231.intflag = 0;
			break;
	}
}

I286_INB cs4231_r(WORD port) {

	BYTE	ret = 0;

	switch(port & 0x0f) {
		case 0:
			return(cs4231.adrs);
			break;

		case 3:
			return(0x04);

		case 4:
			return(cs4231.index & 0x7f);

		case 5:
			return(*(((BYTE *)(&cs4231.reg)) + (cs4231.index & 0x1f)));

		case 6:
			return(cs4231.intflag);

	}
	return(ret);
}

// -------------------------------------------------------------------------

#ifdef SOUND_CS4231_ASM

static LABEL BYTE __fastcall cs4231_pcm8s(void) {

	__asm {
				pushad
				movzx	ebx, dma.dmach[0].leng.w
				cmp		ebx, 2
				jc		cs4231dma_notrans

				mov		edi, cs4231.bufsize
				cmp		edi, 64
				jnc		cs4231dma_notrans

				mov		ecx, dma.dmach[0].adrs.d
				mov		ebp, cs4231.curtime
				mov		esi, cs4231.writepos
cs4231dma_checklp:
				cmp		edi, 64
				jnc		cs4231dma_full
cs4231dma_checkcnt:
				cmp		ebp, cs4231.step
				jnc		cs4231dma_outdata
				add		ebp, np2sound.rate
				sub		ebx, 2
				jc		cs4231dma_empty
				call	i286_memoryread
				inc		ecx
				sub		al, 80h
				mov		ah, al
				shl		eax, 16
				call	i286_memoryread
				inc		ecx
				sub		al, 80h
				mov		ah, al
				mov		cs4231.pcmdata, eax
				jmp		cs4231dma_checkcnt
cs4231dma_outdata:
				sub		ebp, cs4231.step
				mov		eax, cs4231.pcmdata
				mov		dword ptr (cs4231.buffer[esi]), eax
				add		esi, 4
				and		esi, 255
				inc		edi
				jmp		cs4231dma_checklp
cs4231dma_empty:
				xor		ebx, ebx
cs4231dma_full:
				mov		cs4231.bufsize, edi
				mov		dma.dmach[0].leng.w, bx
				mov		dma.dmach[0].adrs.d, ecx
				mov		cs4231.curtime, ebp
				mov		cs4231.writepos, esi
cs4231dma_notrans:
				cmp		ebx, 2
				popad
				setc	al
				ret
	}
}

static LABEL BYTE __fastcall cs4231_pcm8m(void) {

	__asm {
				pushad
				movzx	ebx, dma.dmach[0].leng.w
				cmp		ebx, 1
				jc		cs4231dma_notrans

				mov		edi, cs4231.bufsize
				cmp		edi, 64
				jnc		cs4231dma_notrans

				mov		ecx, dma.dmach[0].adrs.d
				mov		ebp, cs4231.curtime
				mov		esi, cs4231.writepos
cs4231dma_checklp:
				cmp		edi, 64
				jnc		cs4231dma_full
cs4231dma_checkcnt:
				cmp		ebp, cs4231.step
				jnc		cs4231dma_outdata
				add		ebp, np2sound.rate
				sub		ebx, 1
				jc		cs4231dma_empty
				call	i286_memoryread
				inc		ecx
				sub		al, 80h
				mov		ah, al
				mov		word ptr (cs4231.pcmdata), ax
				mov		word ptr (cs4231.pcmdata+2), ax
				jmp		cs4231dma_checkcnt
cs4231dma_outdata:
				sub		ebp, cs4231.step
				mov		eax, cs4231.pcmdata
				mov		dword ptr (cs4231.buffer[esi]), eax
				add		esi, 4
				and		esi, 255
				inc		edi
				jmp		cs4231dma_checklp
cs4231dma_empty:
				xor		ebx, ebx
cs4231dma_full:
				mov		cs4231.bufsize, edi
				mov		dma.dmach[0].leng.w, bx
				mov		dma.dmach[0].adrs.d, ecx
				mov		cs4231.curtime, ebp
				mov		cs4231.writepos, esi
cs4231dma_notrans:
				cmp		ebx, 2
				popad
				setc	al
				ret
	}
}

static LABEL BYTE __fastcall cs4231_pcm16s(void) {

	__asm {
				pushad
				movzx	ebx, dma.dmach[0].leng.w
				cmp		ebx, 4
				jc		cs4231dma_notrans

				mov		edi, cs4231.bufsize
				cmp		edi, 64
				jnc		cs4231dma_notrans

				mov		ecx, dma.dmach[0].adrs.d
				mov		ebp, cs4231.curtime
				mov		esi, cs4231.writepos
cs4231dma_checklp:
				cmp		edi, 64
				jnc		cs4231dma_full
cs4231dma_checkcnt:
				cmp		ebp, cs4231.step
				jnc		cs4231dma_outdata
				add		ebp, np2sound.rate
				sub		ebx, 4
				jc		cs4231dma_empty
				call	i286_memoryread_w
				add		ecx, 2
				xchg	al, ah
				shl		eax, 16
				call	i286_memoryread_w
				add		ecx, 2
				xchg	al, ah
				mov		cs4231.pcmdata, eax
				jmp		cs4231dma_checkcnt
cs4231dma_outdata:
				sub		ebp, cs4231.step
				mov		eax, cs4231.pcmdata
				mov		dword ptr (cs4231.buffer[esi]), eax
				add		esi, 4
				and		esi, 255
				inc		edi
				jmp		cs4231dma_checklp
cs4231dma_empty:
				xor		ebx, ebx
cs4231dma_full:
				mov		cs4231.bufsize, edi
				mov		dma.dmach[0].leng.w, bx
				mov		dma.dmach[0].adrs.d, ecx
				mov		cs4231.curtime, ebp
				mov		cs4231.writepos, esi
cs4231dma_notrans:
				cmp		ebx, 4
				popad
				setc	al
				ret
	}
}

static LABEL BYTE __fastcall cs4231_pcm16m(void) {

	__asm {
				pushad
				movzx	ebx, dma.dmach[0].leng.w
				cmp		ebx, 2
				jc		cs4231dma_notrans

				mov		edi, cs4231.bufsize
				cmp		edi, 64
				jnc		cs4231dma_notrans

				mov		ecx, dma.dmach[0].adrs.d
				mov		ebp, cs4231.curtime
				mov		esi, cs4231.writepos
cs4231dma_checklp:
				cmp		edi, 64
				jnc		cs4231dma_full
cs4231dma_checkcnt:
				cmp		ebp, cs4231.step
				jnc		cs4231dma_outdata
				add		ebp, np2sound.rate
				sub		ebx, 2
				jc		cs4231dma_empty
				call	i286_memoryread_w
				add		ecx, 2
				xchg	al, ah
				mov		word ptr (cs4231.pcmdata), ax
				mov		word ptr (cs4231.pcmdata+2), ax
				jmp		cs4231dma_checkcnt
cs4231dma_outdata:
				sub		ebp, cs4231.step
				mov		eax, cs4231.pcmdata
				mov		dword ptr (cs4231.buffer[esi]), eax
				add		esi, 4
				and		esi, 255
				inc		edi
				jmp		cs4231dma_checklp
cs4231dma_empty:
				xor		ebx, ebx
cs4231dma_full:
				mov		cs4231.bufsize, edi
				mov		dma.dmach[0].leng.w, bx
				mov		dma.dmach[0].adrs.d, ecx
				mov		cs4231.curtime, ebp
				mov		cs4231.writepos, esi
cs4231dma_notrans:
				cmp		ebx, 2
				popad
				setc	al
				ret
	}
}

static BYTE (__fastcall * cs4231enc[])(void) = {
			cs4231_pcm8m,		// 0: 8bit PCM
			cs4231_pcm8s,
			cs4231nomake,		// 1: u-Law
			cs4231nomake,
			cs4231nomake,		// 2:
			cs4231nomake,
			cs4231nomake,		// 3: A-law
			cs4231nomake,
			cs4231nomake,		// 4:
			cs4231nomake,
			cs4231nomake,		// 5: ADPCM
			cs4231nomake,
			cs4231_pcm16m,		// 6: 16bit PCM
			cs4231_pcm16s,
			cs4231nomake,		// 7: ADPCM
			cs4231nomake};

// --------------------------------------------------------------------------

LABEL void __fastcall cs4231_update(void *buf, int buflength) {

	__asm {
				cmp		edx, cs4231.bufsize
				jbe		short cs4231mk_smp
				mov		edx, cs4231.bufsize
cs4231mk_smp:	cmp		edx, 0
				je		cs4231mk_end

				sub		cs4231.bufsize, edx
				push	ecx
				push	esi
				mov		esi, cs4231.readpos
cs4231mk_lp:	movsx	eax, word ptr (cs4231.buffer[esi])
				sar		eax, 1
				add		[ecx], eax
				movsx	eax, word ptr (cs4231.buffer[esi+2])
				sar		eax, 1
				add		[ecx+4], eax
				add		ecx, 8
				add		esi, 4
				and		esi, 255
				dec		edx
				jne		cs4231mk_lp
				mov		cs4231.readpos, esi
				pop		esi
				pop		ecx
cs4231mk_end:	ret
	}
}

#else

static BYTE (SOUND_CALL * cs4231enc[])(void) = {

			cs4231nomake,		// 0: 8bit PCM
			cs4231nomake,
			cs4231nomake,		// 1: u-Law
			cs4231nomake,
			cs4231nomake,		// 2:
			cs4231nomake,
			cs4231nomake,		// 3: A-law
			cs4231nomake,
			cs4231nomake,		// 4:
			cs4231nomake,
			cs4231nomake,		// 5: ADPCM
			cs4231nomake,
			cs4231nomake,		// 6: 16bit PCM
			cs4231nomake,
			cs4231nomake,		// 7: ADPCM
			cs4231nomake};


void SOUND_CALL cs4231_update(void * buf, int buflength) { }

#endif
