#include	"compiler.h"

#include	"np2.h"
#include	"pc_core.h"
#include	"iobridge.h"
#include	"sndboard.h"
#include	"np2sound.h"
#include	"pic.h"
#include	"nevent.h"
#include	"fmtimer.h"
#include	"pcm86.h"

	FMTIMER_T	fmtimer;

static BYTE irqtable[4] = {0x03, 0x0d, 0x0a, 0x0c};

static void set_fmtimeraevent(BOOL absolute);
static void set_fmtimerbevent(BOOL absolute);

static void fmport_a(NEVENT_ITEM *item) {

	BOOL	intreq = FALSE;

	if (item->flag & NEVENT_SETEVENT) {
		if (fmtimer.reg & 0x01) {
			fmtimer.status |= 0x01;
		}
		if (fmtimer.reg & 0x04) {
			intreq = TRUE;
		}
		else if (pcm86.fifo & 0x20) {
			sound_make();
			if (pcm86.bufsize < pcm86.fifosize) {
				intreq = TRUE;
			}
		}
		if (intreq) {
			pcm86.write = 1;
			pic_setirq(fmtimer.irq);
		}
		if (fmtimer.reg & 0x10) {
			set_fmtimeraevent(NEVENT_RELATIVE);
		}
	}
}

static void fmport_b(NEVENT_ITEM *item) {

	BOOL	intreq = FALSE;

	if (item->flag & NEVENT_SETEVENT) {
		if (fmtimer.reg & 0x02) {
			fmtimer.status |= 0x02;
		}
		if (fmtimer.reg & 0x08) {
			intreq = TRUE;
		}
		else if (pcm86.fifo & 0x20) {
			sound_make();
			if (pcm86.bufsize < pcm86.fifosize) {
				intreq = TRUE;
			}
		}
		if (intreq) {
			pcm86.write = 1;
			pic_setirq(fmtimer.irq);
		}
		if (fmtimer.reg & 0x20) {
			set_fmtimerbevent(NEVENT_RELATIVE);
		}
	}
}

static void set_fmtimeraevent(BOOL absolute) {

	long l = 18 * (1024 - fmtimer.timera);

	if (pc.cpumode & CPUMODE_8MHz) {		// 4MHz
		l = (l * 1248 / 625) * pc.multiple;
	}
	else {									// 5MHz
		l = (l * 1536 / 625) * pc.multiple;
	}
	nevent_set(NEVENT_FMTIMERA, l, fmport_a, absolute);
}

static void set_fmtimerbevent(BOOL absolute) {

	long l = 288 * (256 - fmtimer.timerb);

	if (pc.cpumode & CPUMODE_8MHz) {		// 4MHz
		l = (l * 1248 / 625) * pc.multiple;
	}
	else {									// 5MHz
		l = (l * 1536 / 625) * pc.multiple;
	}
	nevent_set(NEVENT_FMTIMERB, l, fmport_b, absolute);
}


// -------------------------------------------------------------------------

void fmtimer_reset(BYTE irq) {

	ZeroMemory(&fmtimer, sizeof(fmtimer));
	fmtimer.intr = irq & 0xc0;
	fmtimer.intdisabel = irq & 0x10;
	fmtimer.irq = irqtable[irq >> 6];
}

void fmtimer_setreg(BYTE reg, BYTE value) {

	switch(reg) {
		case 0x24:
			fmtimer.timera = (value << 2) + (fmtimer.timera & 3);
			break;

		case 0x25:
			fmtimer.timera = (fmtimer.timera & 0x3fc) + (value & 3);
			break;

		case 0x26:
			fmtimer.timerb = value;
			break;

		case 0x27:
			fmtimer.reg = value;
			fmtimer.status &= ~((value & 0x30) >> 4);
			if ((value & 0x10) && (!nevent_iswork(NEVENT_FMTIMERA))) {
				set_fmtimeraevent(NEVENT_ABSOLUTE);
			}
			if ((value & 0x20) && (!nevent_iswork(NEVENT_FMTIMERB))) {
				set_fmtimerbevent(NEVENT_ABSOLUTE);
			}
			break;
	}
}

