#include	"compiler.h"

#include	"np2.h"
#include	"i286.h"
#include	"pc_core.h"
#include	"nevent.h"
#include	"iobridge.h"
#include	"pic.h"
#include	"mpu98ii.h"
#include	"trace.h"


#define	MIDIIN_AVAIL	0x80
#define	MIDIOUT_BUSY	0x40

#define	MPU_INT			0xfd
#define MPU_ACK         0xfe

#define	MIDITIMEOUTCLOCK	3000
#define	MIDITIMEOUTCLOCK2	300


#define	MIDI_EXCLUSIVE		0xf0
#define MIDI_TIMECODE		0xf1
#define MIDI_SONGPOS		0xf2
#define MIDI_SONGSELECT		0xf3
#define	MIDI_TUNEREQUEST	0xf6
#define	MIDI_EOX			0xf7
#define	MIDI_TIMING			0xf8
#define MIDI_START			0xfa
#define MIDI_CONTINUE		0xfb
#define	MIDI_STOP			0xfc
#define	MIDI_ACTIVESENSE	0xfe
#define	MIDI_SYSTEMRESET	0xff



	MPU98_TABLE		mpu98;

static	int		midiready = 0;
static	BYTE	mpuirqnum[4] = {3, 5, 6, 12};



NEVENT_PROC miditimeout(NEVENT_ITEM *item);

static void makeintclock(void) {

	DWORD	l;

	l = mpu98.tempo * 2 * mpu98.tempos / 0x40;
	if (l < 5*2) {
		l = 5*2;
	}
	l *= mpu98.timebase;						//	*12
	mpu98.clock = (pc.realclock * 5 / l);		//	/12
}

static void setrecvdata(BYTE data) {

	if (mpu98.cnt < MPU98_RECVBUFS) {
		mpu98.buf[(mpu98.pos + mpu98.cnt) & (MPU98_RECVBUFS - 1)] = data;
		mpu98.cnt++;
	}
}

static void ch_step(void) {

	int		i;
	BYTE	bit;

	for (i=0, bit=1; bit; bit<<=1, i++) {
		if (mpu98.intch & bit) {
			if (mpu98.ch[i].step) {
				mpu98.ch[i].step--;
			}
		}
	}
}

static BOOL ch_nextsearch(void) {

	int		i, j;
	BYTE	bit;

	i = mpu98.intreq & 7;
	bit = 1 << i;
	for (; bit; bit<<=1, i++) {
		if ((mpu98.intch & bit) && (!mpu98.ch[i].step)) {
			if ((mpu98.ch[i].datas) && (mpu98.ch[i].remain == 0)) {
				if (mpu98.ch[i].data[0] == MIDI_STOP) {
					mpu98.ch[i].datas = 0;
					mpu98.intreq = 0;
					setrecvdata(0xfc);
					pic_setirq(mpu98.irqnum);					// ver0.28
					return(TRUE);
				}
				for (j=0; j<mpu98.ch[i].datas; j++) {
					np2os_midimessage(mpu98.ch[i].data[j]);
				}
				mpu98.ch[i].datas = 0;
			}
			nevent_set(NEVENT_MIDITIMEOUT, MIDITIMEOUTCLOCK,
											miditimeout, NEVENT_ABSOLUTE);

			mpu98.intreq = (BYTE)(0xf0 + i);
			setrecvdata(mpu98.intreq);
			mpu98.recvevent |= 1;
TRACE_("midi_int!", mpu98.intreq);
			pic_setirq(mpu98.irqnum);							// ver0.28
			return(TRUE);
		}
	}
	mpu98.intreq = 0;
#if 1
	if (!(++mpu98.timing)) {
		setrecvdata(0xf9);
		pic_setirq(mpu98.irqnum);								// ver0.28
	}
#endif
	return(FALSE);
}

NEVENT_PROC midiint(NEVENT_ITEM * item) {

TRACE_("midi int!", 0);
	if (!mpu98.intch) {
		setrecvdata(MPU_INT);
		pic_setirq(mpu98.irqnum);								// ver0.28
	}
	else {
		if (!nevent_iswork(NEVENT_MIDITIMEOUT)) {
			ch_step();
			mpu98.intreq = 0xf0;
			ch_nextsearch();
		}
		else {
			mpu98.remainstep++;
		}
	}
	nevent_set(NEVENT_MIDIINT, mpu98.clock, midiint, NEVENT_RELATIVE);
}

NEVENT_PROC midiwaitout(NEVENT_ITEM * item) {

	mpu98.status &= ~MIDIOUT_BUSY;
}

NEVENT_PROC miditimeout(NEVENT_ITEM * item) {

	if ((pic[0].isr & 0x40) || isI286DI) {
		nevent_set(NEVENT_MIDITIMEOUT, MIDITIMEOUTCLOCK2,
											miditimeout, NEVENT_ABSOLUTE);
	}
	else {
		while(1) {
			ch_nextsearch();
			if ((mpu98.intreq) || (!mpu98.remainstep)) {
				break;
			}
			mpu98.remainstep--;
			ch_step();
			mpu98.intreq = 0xf0;
		}
	}
}

static void midiwait(long waitclock) {

	if (!nevent_iswork(NEVENT_MIDIWAIT)) {
		mpu98.status |= MIDIOUT_BUSY;
		nevent_set(NEVENT_MIDIWAIT, waitclock, midiwaitout, NEVENT_ABSOLUTE);
	}
}

static BOOL sendcmd(BYTE cmd) {

	BYTE	work;

	mpu98.cmd = cmd;
	switch(cmd & 0xf0) {
		case 0xe0:				// send 2nddata
			mpu98.recvevent |= 2;
			return(TRUE);
		case 0xa0:				// recv data
			return(TRUE);

		case 0xc0:				// timebase
			work = cmd & 0x0f;
			if ((!work) || (work > MPU98_MAXTIMEBASE)) {
				return(FALSE);
			}
			mpu98.timebase = work;
			makeintclock();
			return(TRUE);
	}

	switch(cmd) {
		case 0xd0:				// send short
		case 0xdf:				// send long
			break;

		case 0xff:				// reset
			np2os_midireset();
			mpu98.intch = 0;
			mpu98.recvevent = 0;
			mpu98.intreq = 0;
			ZeroMemory(mpu98.ch, sizeof(mpu98.ch));
			nevent_reset(NEVENT_MIDIINT);
			nevent_reset(NEVENT_MIDITIMEOUT);
			break;

		case 0x3f:				// uart mode on
			mpu98.mode = 1;
			np2os_midireset();								// PR2
			break;

		case 0x94:				// disable clock to host
		case 0x05:
			mpu98.recvevent = 0;
			mpu98.intreq = 0;
			ZeroMemory(mpu98.ch, sizeof(mpu98.ch));
			nevent_reset(NEVENT_MIDIINT);
			nevent_reset(NEVENT_MIDITIMEOUT);
			break;

		case 0x95:				// enable clock to host
		case 0x0a:
			if (!nevent_iswork(NEVENT_MIDIINT)) {
				nevent_set(NEVENT_MIDIINT, mpu98.clock,
											midiint, NEVENT_ABSOLUTE);
			}
			break;

#if 0							// ݤʤΤǥޥɥåʤ
		case 0x01:				// send MIDI stop
		case 0x02:				// send MIDI start
		case 0x03:				// send MIDI continue

		case 0x15:				// stop recording, playback and MIDI

		case 0x32:				// ?
		case 0x34:				// return timing bytes in stop mode
		case 0x35:				// enable mode messages to PC
		case 0x38:				// enable sys common messages to PC
		case 0x39:				// Enable real time messages to PC
		case 0x3c:				// use CLS sync
		case 0x3d:				// use SMPTE sync

		case 0x80:				// use MIDI sync
		case 0x81:				// use FSK sync
		case 0x82:				// use MIDI sync
		case 0x83:				// enable metronome without accents
		case 0x84:				// disable metronome
		case 0x8a:				// disable data in stopped mode
		case 0x8b:				// enable data in stop mode
		case 0x8c:				// disable measure end messages to host

		case 0x91:				// enable ext MIDI ctrl
		case 0x97:				// enable system exclusive messages to PC
			break;

		default:
			return(FALSE);
#endif
	}
	return(TRUE);
}

static void senddat(BYTE data) {

	if (mpu98.recvevent & 2) {
		mpu98.recvevent &= ~2;
		switch(mpu98.cmd) {
			case 0xe0:				// tempo
				mpu98.tempo = data;
				mpu98.tempos = 0x40;
				makeintclock();
				break;

			case 0xe1:				// ? Хƥݤäݤɡ
				mpu98.tempos = data;
				makeintclock();
				break;

			case 0xe2:				// ?
				break;

			case 0xe4:				// clocks/click
				break;

			case 0xe6:				// beats/measure
				TRACE_("beat/measure:", data);
				break;

			case 0xe7:				// send all clocks to host
				break;

			case 0xec:				// channel mask?
				mpu98.intch = data;
				break;
		}
	}
	else if (mpu98.intreq) {
		MPU_CH *ch = &mpu98.ch[mpu98.intreq & 7];
	 	if (mpu98.recvevent & 1) {
			mpu98.recvevent &= ~1;
			if (data < 0xf0) {
				ch->step = data;
				ch->remain = 0xff;
			}
			else {
				ch->step = 0;
				ch->remain = 0;
				ch->data[0] = data;
				ch->datas = 1;
				ch->remain = 0;
			}
		}
		else if (ch->remain) {
			if (ch->remain == 0xff) {
				ch->datas = 0;
				switch(data & 0xf0) {
						case 0xc0:
						case 0xd0:
						ch->remain = 2;
						break;
					case 0x80:
					case 0x90:
					case 0xa0:
					case 0xb0:
					case 0xe0:
						ch->remain = 3;
						break;
					case 0xf0:
						ch->remain = 1;
						break;
					default:
						ch->remain = 0;
						return;
				}
			}
			ch->data[ch->datas] = data;
			(ch->datas)++;
			(ch->remain)--;
		}
		if (ch->remain) {
			nevent_set(NEVENT_MIDITIMEOUT, MIDITIMEOUTCLOCK,
											miditimeout, NEVENT_ABSOLUTE);
		}
		else {
			nevent_set(NEVENT_MIDITIMEOUT, MIDITIMEOUTCLOCK2,
											miditimeout, NEVENT_ABSOLUTE);
		}
	}
}


// --------------------------------------------------------------------------

void mpu98ii_init(void) {

	ZeroMemory(&mpu98, sizeof(mpu98));
	mpu98.data = MPU_ACK;
	mpu98.tempo = 120;
	mpu98.tempos = 0x40;
	mpu98.timebase = 2;
	mpu98.port = 0xc0d0 | ((np2cfg.mpuopt & 0xf0) << 6);		// ver0.28
	mpu98.irqnum = mpuirqnum[np2cfg.mpuopt & 3];				// ver0.28
	makeintclock();
	midiready = np2os_ismidiready();
}

I286_OUT mpu98ii_w(WORD port, BYTE value) {

	long	l;

	TRACEOUT(port, value);

	if (midiready) {
		switch(port & 0x03) {
			case 0x00:
				l = 1;
				if (mpu98.mode) {
					l = np2os_midimessage(value);
				}
				else {
					if ((mpu98.cmd == 0xd0) || (mpu98.cmd == 0xdf)) {
						l = np2os_midimessage(value);
					}
					else {
						senddat(value);
					}
				}
				if (l) {
					midiwait(pc.midiclock * l);
				}
				break;

			case 0x02:
				if (!mpu98.mode) {
					if (sendcmd(value)) {
						setrecvdata(MPU_ACK);
						switch(value) {
							case 0xac:			// get mpu major version?
								setrecvdata(1);
								break;
							case 0xad:			// get mpu minor version?
								setrecvdata(0);
								break;
						}
					}
				}
				else {
					if (value == 0xff) {
						mpu98.mode = 0;
						setrecvdata(MPU_ACK);
					}
				}
				midiwait(pc.realclock / 10000);
				break;
		}
	}
}

I286_INB mpu98ii_r(WORD port) {

	BYTE	ret = 0xff;

	if (midiready) {
		switch(port & 0x03) {
			case 0x00:
				if (mpu98.cnt) {
					mpu98.cnt--;
					mpu98.data = mpu98.buf[mpu98.pos];
					mpu98.pos = (mpu98.pos + 1) & (MPU98_RECVBUFS - 1);
				}
				ret = mpu98.data;
				break;

			case 0x02:
				ret = mpu98.status;
				if (!mpu98.cnt) {
					ret |= MIDIIN_AVAIL;
				}
				break;
		}
	}
	return(ret);
}

