#include	"compiler.h"

#include	"np2pc98.h"
#include	"iobridge.h"
#include	"sndboard.h"
#include	"np2sound.h"
#include	"pc_core.h"
#include	"nevent.h"
#include	"beep.h"
#include	"fm.h"
#include	"psg.h"
#include	"rhythm.h"
#include	"adpcm.h"


#ifndef NP2SOUND_BUFSIZE
#define NP2SOUND_BUFSIZE	44100
#endif

static	long	buffer[NP2SOUND_BUFSIZE*2];
static	void	noncb(void) { }


		NP2SOUND_T	np2sound = {22050, 110, 0, 0, 0, 0, noncb, 2};

void sound_reset(void) {

	np2sound.exist = 0;
	np2sound.rp = 0;
	np2sound.wp = 0;
	np2sound.lastclock = nevent.clock;
	beep_eventreset();
}


static void sound_bufferwrite(int length) {

	if (length + np2sound.exist > NP2SOUND_BUFSIZE) {
		length = NP2SOUND_BUFSIZE - np2sound.exist;
	}
	np2sound.exist += length;
	if (length + np2sound.wp >= NP2SOUND_BUFSIZE) {
		int len = NP2SOUND_BUFSIZE - np2sound.wp;
		if (len) {
			sndproc.outbuffer(buffer + (np2sound.wp * 2), len);
			length -= len;
		}
		np2sound.wp = 0;
	}
	if (length) {
		sndproc.outbuffer(buffer + (np2sound.wp * 2), length);
		np2sound.wp += length;
	}
}


void sound_make(void) {

	DWORD	len;

	len = (nevent.clock + nevent.baseclock - nevent.remainclock
								- np2sound.lastclock) * np2sound.rate200;
	if (len >= pc.dsoundclock2) {
		len /= pc.dsoundclock;
		sound_bufferwrite(len);
		np2sound.lastclock += (len * pc.dsoundclock / np2sound.rate200) + 1;
		beep_eventreset();
	}
	np2sound.callback();
}

// --------------------------------------------------------------------------

void np2sound_init(DWORD rate, void (*callback)(void)) {

	np2sound.rate = rate;
	np2sound.rate200 = (rate + 199) / 200;
	if (callback) {
		np2sound.callback = callback;
	}
	else {
		np2sound.callback = noncb;
	}

	// ץ󥰥졼Ȥη
	if (rate < AVE(11025, 22050)) {
		np2sound.pcm86rate = 4;
	}
	else if (rate < AVE(22050, 44100)) {
		np2sound.pcm86rate = 2;
	}
	else {
		np2sound.pcm86rate = 1;
	}

	fm_tableinit();
	psg_tableinit();
	rhythm_tableinit();
	sound_reset();
}

void np2sound_term(void) {

	rhythm_tableterm();
}


void np2sound_getsounddata(int leng, NP2SNDBUF_T *sndbuf) {

	NP2SNDBUF_T ret;

	if (leng > np2sound.exist) {
		sound_bufferwrite(leng - np2sound.exist);
		np2sound.lastclock = nevent.clock;
		beep_eventreset();
	}
	np2sound.exist -= leng;

	ret.buf1 = buffer + (np2sound.rp * 2);
	if (leng + np2sound.rp >= NP2SOUND_BUFSIZE) {
		ret.len1 = NP2SOUND_BUFSIZE - np2sound.rp;
		ret.buf2 = buffer;
		ret.len2 = leng - ret.len1;
		np2sound.rp = ret.len2;
	}
	else {
		ret.len1 = leng;
		ret.buf2 = buffer;
		ret.len2 = 0;
		np2sound.rp += leng;
	}

	if (sndbuf) {
		*sndbuf = ret;
	}
}


void np2sound_clrsounddata(void) {

	sound_reset();
}


void np2sound_volume(void) {

	fm_volume();
	psg_volume();
	rhythm_volume();
//	adpcm_volume();
}
