//
//  PC-98 Sound logging
//    for S98amp S98 Input plugin for Winamp Version 1.3.1+ by Mamiya
//

#include	"compiler.h"

#include	"dosio.h"
#include	"s98.h"
#include	"nevent.h"
#include	"pc_core.h"
#include	"iobridge.h"

#define S98LOG_BUFSIZE (32*1024)

static struct {
	FILEH	fh;
	DWORD	intcount;
	DWORD	clock;
	DWORD	p;
	BYTE	buf[S98LOG_BUFSIZE];
} s98log = { (FILEH)-1, 0, 0, 0, {0} };

void s98timer(NEVENT_ITEM *item);

static void sets98event(BOOL absolute) {

	s98log.intcount++;
	nevent_set(NEVENT_S98TIMER, s98log.clock, s98timer, NEVENT_RELATIVE);
}

void s98timer(NEVENT_ITEM * item) {

	if (s98log.fh != (FILEH)-1) {
		sets98event(NEVENT_RELATIVE);
	}
}

static void S98_flush(void) {

	if (s98log.fh != (FILEH)-1) {
		if (s98log.p) {
			file_lwrite(s98log.fh, s98log.buf, s98log.p);
			s98log.p = 0;
		}
	}
}

static void S98_putc(BYTE data) {

	if (s98log.fh != (FILEH)-1) {
		s98log.buf[s98log.p++] = data;
		if (s98log.p == S98LOG_BUFSIZE) {
			S98_flush();
		}
	}
}

static void S98_putint(void) {

	if (s98log.intcount) {
		if (s98log.intcount == 1) {
			S98_putc(0xFF);					/* SYNC(1) */
		}
		else if (s98log.intcount == 2) {
			S98_putc(0xFF);					/* SYNC(1) */
			S98_putc(0xFF);					/* SYNC(1) */
		}
		else {
			S98_putc(0xFE);					/* SYNC(n) */
			s98log.intcount -= 2;
			while (s98log.intcount > 0x7f) {
				S98_putc((BYTE)(0x80 | (s98log.intcount & 0x7f)));
				s98log.intcount >>= 7;
			}
			S98_putc((BYTE)(s98log.intcount & 0x7f));
		}
		s98log.intcount = 0;
	}
}

void S98_put(BYTE ms, BYTE adr, BYTE data) {

	if (s98log.fh != (FILEH)-1) {
		S98_putint();
		S98_putc(ms);
		S98_putc(adr);
		S98_putc(data);
	}
}

BOOL S98_open(char *filename) {

	unsigned int			i;
	S98_TABLE	s98 = {{'S','9','8'}, '1', 1, 0, 0, 0x40, 0x80, 0, {0}, {0}};

	// եΥץ
	if ((s98log.fh = file_create(filename, FTYPE_S98)) == (FILEH)-1) {
		return(FALSE);
	}

	// 
	s98log.clock = pc.realclock / 1000;
	s98log.p = 0;

	// إå¸
	for (i=0; i<sizeof(S98_TABLE); i++) {
		S98_putc(*(((BYTE *)&s98) + i));
	}


#if 0
	// FM
	for (i=0x30; i<0xb6; i++) {
		if ((i & 3) != 3) {
			S98_putc(NORMAL2608);
			S98_putc((BYTE)i);
			S98_putc(opna.reg.b[i]);
			S98_putc(EXTEND2608);
			S98_putc((BYTE)i);
			S98_putc(opna.reg.b[i+0x100]);
		}
	}
	// PSG
	for (i=0x00; i<0x0e; i++) {
		S98_putc(NORMAL2608);
		S98_putc((BYTE)i);
		S98_putc(opna.reg.b[i]);
	}
#endif

	// ѥǥ
	s98log.intcount = 10;

	sets98event(NEVENT_ABSOLUTE);
	return(TRUE);
}

void S98_close(void) {

//	if (s98log.fh != (FILEH)-1) {

		S98_putint();
		S98_putc(0xFD);				/* END MARK */
		S98_flush();
		nevent_reset(NEVENT_S98TIMER);
		file_close(s98log.fh);
		s98log.fh = (FILEH)-1;
//	}
}
