#include	"compiler.h"

#include	"np2.h"
#include	"iobridge.h"
#include	"sndboard.h"
#include	"np2sound.h"
#include	"fmtimer.h"
#include	"beep.h"
#include	"fm.h"
#include	"psg.h"
#include	"rhythm.h"
#include	"pcm86.h"
#include	"s98.h"

static void SOUND_CALL makesounds(void *buf, int length) {

	fm_update(buf, length);
	psg_update(buf, 0, length);
	psg_update(buf, 1, length);
	rhythm_update(buf, length);
	pcm86_update(buf, length);
	beep_update(buf, length);
}

void sndbx2_init(void) {

	fmtimer_reset((np2cfg.snd86opt & 0x10) |
					((np2cfg.snd86opt & 0x4) << 5) |
					((np2cfg.snd86opt & 0x8) << 3));
	fm_reset();
	psg_reset(0);
	psg_reset(1);
	rhythm_reset();
	pcm86_reset();
	fm_channel(6, FM_MONORAL | 0x3f);
	if (np2cfg.snd86opt & 2) {
		sndbios_load(0xcc000, "86");
	}
	sndreg.base = ((np2cfg.snd86opt & 0x01)?0x000:0x100);
}

static I286_OUT sndbx2_w(WORD port, BYTE value) {

	port -= sndreg.base;
	switch(port & 0x707) {
		case 0x000:
			sndreg.opn2reg = value;
			break;

		case 0x002:
			sndreg.reg[sndreg.opn2reg + 0x200] = value;
			if (sndreg.opn2reg < 0x10) {
				if (sndreg.opn2reg != 0x0e) {
					sound_make();							// ver0.28
					psg_setreg(0, sndreg.opn2reg, value);
				}
			}
			else {
				if (sndreg.opn2reg < 0x30) {
					if (sndreg.opn2reg == 0x28) {
						if ((value & 0x0f) < 3) {
							sound_make();
							fm_keyon(value & 0x0f, value);
						}
					}
				}
				else if (sndreg.opn2reg < 0xc0) {
					sound_make();
					fm_setreg(sndreg.opn2reg, value, 0);
				}
			}
			break;

		case 0x100:
			sndreg.opnreg = value;
			break;

		case 0x102:
			S98_put(NORMAL2608, sndreg.opnreg, value);
			sndreg.reg[sndreg.opnreg] = value;
			if (sndreg.opnreg < 0x10) {
				if (sndreg.opnreg != 0x0e) {
					sound_make();							// ver0.28
					psg_setreg(1, sndreg.opnreg, value);
				}
			}
			else {
				if (sndreg.opnreg < 0x20) {
					if (sndreg.extend) {
						rhythm_setreg(sndreg.opnreg, value);
					}
				}
				else if (sndreg.opnreg < 0x30) {
					if (sndreg.opnreg == 0x28) {
						if ((value & 0x0f) < 3) {
							sound_make();
							fm_keyon((value & 0x0f)+3, value);
						}
						else if (((value & 0x0f) != 3) &&
								((value & 0x0f) < 7)) {
							sound_make();
							fm_keyon((value & 0x0f)+2, value);
						}
					}
					else {
						fmtimer_setreg(sndreg.opnreg, value);
						if (sndreg.opnreg == 0x27) {
							fm_extchannel(5, value & 0xc0);
						}
					}
				}
				else if (sndreg.opnreg < 0xc0) {
					sound_make();
					fm_setreg(sndreg.opnreg, value, 3);
				}
			}
			break;

		case 0x104:
			sndreg.extreg = value;
			break;

		case 0x106:
			S98_put(EXTEND2608, sndreg.extreg, value);
			sndreg.reg[sndreg.extreg + 0x100] = value;
			if (sndreg.extreg >= 0x30) {
				sound_make();
				fm_setreg(sndreg.extreg, value, 6);
			}
			else {
				if (sndreg.extreg == 0x10) {
					if (!(value & 0x80)) {
						sndreg.adpcmmask = ~(value & 0x1c);
					}
				}
			}
			break;
	}
}

static I286_INB sndbx2_r(WORD port) {

	port -= sndreg.base;
	switch(port & 0x707) {
		case 0x000:
			return(fmtimer.status);

		case 0x002:
			if (sndreg.opn2reg == 0x0e) {
				return(0xff);
			}
			if (sndreg.opn2reg < 0x10) {
				return(sndreg.reg[sndreg.opn2reg + 0x200]);
			}
			break;

		case 0x100:
			return(fmtimer.status);

		case 0x102:
			if (sndreg.opnreg == 0x0e) {
				return(psg_getjoyport(1));
			}
			return(sndreg.reg[sndreg.opnreg]);

		case 0x104:
			if (sndreg.extend) {
				return((fmtimer.status & 3) | (sndreg.adpcmmask & 8));
			}
			break;

		case 0x106:
			if (sndreg.extend) {
				return(sndreg.reg[sndreg.opnreg]);
			}
			break;
	}
	return(0xff);
}

static void extendchannel(BYTE enable) {

	sndreg.extend = enable;
	if (enable) {
		fm_channel(9, FM_STEREO | 0x1f8);
	}
	else {
		fm_channel(6, FM_MONORAL | 0x38);
		rhythm_setreg(0x10, 0xff);
	}
}

SNDPROC_T sndproc_bx2 = {
				sndbx2_w,
				sndbx2_r,
				pcm86_w,
				pcm86_r,
				makesounds,
				extendchannel};

