/*	$Id: sdl.c,v 1.4 2003/06/05 18:11:19 nonaka Exp $	*/

/*
 * Copyright (c) 2002-2003 NONAKA Kimihiro
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#if defined(USE_SDLAUDIO)

#include "compiler.h"

#include "audio.h"

#include "SDL.h"

static int sdlaudio_samples;

static int sdlaudio_init(DWORD, DWORD);
static BOOL sdlaudio_term(void);
static void sdlaudio_play(void);
static void sdlaudio_stop(void);
static void sdlaudio_callback(void *, unsigned char *, int);

int
sdlaudio_setup(void)
{

	audiodrv.init = sdlaudio_init;
	audiodrv.term = sdlaudio_term;
	audiodrv.play = sdlaudio_play;
	audiodrv.stop = sdlaudio_stop;

	return SUCCESS;
}

static int
sdlaudio_init(DWORD rate, DWORD samples)
{
	SDL_AudioSpec fmt;
	int rv;

	fmt.freq = rate;
	fmt.format = AUDIO_S16SYS;
	fmt.channels = 2;
	fmt.samples = samples;
	fmt.callback = sdlaudio_callback;
	fmt.userdata = NULL;

	sdlaudio_samples = samples;

	rv = SDL_Init(SDL_INIT_AUDIO);
	if (rv < 0) {
		fprintf(stderr, "sdlaudio_init: SDL_Init\n");
		return FAILURE;
	}

	audio_fd = SDL_OpenAudio(&fmt, NULL);
	if (audio_fd < 0) {
		fprintf(stderr, "sdlaudio_init: SDL_OpenAudio\n");
		SDL_Quit();
		return FAILURE;
	}

	return SUCCESS;
}

static BOOL
sdlaudio_term(void)
{

	SDL_CloseAudio();
	SDL_Quit();

	return SUCCESS;
}

static void
sdlaudio_play(void)
{

	SDL_PauseAudio(0);
}

static void
sdlaudio_stop(void)
{

	SDL_PauseAudio(1);
}

static void
sdlaudio_callback(void *userdata, unsigned char *stream, int len)
{

	UNUSED(userdata);
	UNUSED(len);

	if (audio_event != NULL)
		memset(audio_event, 0, sdlaudio_samples * 4);

	SDL_MixAudio(stream, audio_buffer[audio_nextbuf], sdlaudio_samples * 4,
	    SDL_MIX_MAXVOLUME);

	audio_nextbuf++;
	if (audio_nextbuf >= NSOUNDBUFFER)
		audio_nextbuf = 0;
	audio_event = audio_buffer[audio_nextbuf];
}

#endif	/* USE_SDLAUDIO */
